package be.iminds.ilabt.jfed.ui.javafx.userlogin;

import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.*;
import be.iminds.ilabt.jfed.ui.javafx.am_list_gui.AMList;
import be.iminds.ilabt.jfed.ui.javafx.choosers.AuthorityChooser;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.x509_certificate_gui.X509CertificatePanel;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.JFedCommonDialogs;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.scene.Parent;
import javafx.scene.control.*;
import javafx.scene.layout.HBox;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Provider;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

/**
 * UserLoginController
 */
public class UserLoginController {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(UserLoginController.class);

    private final UserLoginModelManager userLoginModel;

    private boolean requireAllUserInfo = true;
    private Stage stage;

    public void setStage(Stage stage) {
        this.stage = stage;
    }

    @Inject
    public UserLoginController(UserLoginModelManager userLoginModel) {
        this.userLoginModel = userLoginModel;
    }

    @FXML
    private void initialize() {
        modelToGUI();

        defaultLoginPanelPasswordBox.managedProperty().bind(defaultLoginPanelPasswordBox.visibleProperty());
        loginButton.managedProperty().bind(loginButton.visibleProperty());
        pemKeyCertInputFromFileBox.managedProperty().bind(pemKeyCertInputFromFileBox.visibleProperty());
        errorLogLabel.managedProperty().bind(errorLogLabel.visibleProperty());
        errorLog.managedProperty().bind(errorLog.visibleProperty());
        derivedCertificateDetails.managedProperty().bind(derivedCertificateDetails.visibleProperty());
        manualCertificateDetails.managedProperty().bind(manualCertificateDetails.visibleProperty());
        missingInfoWarningLabel.managedProperty().bind(missingInfoWarningLabel.visibleProperty());
        missingInfoErrorLabel.managedProperty().bind(missingInfoErrorLabel.visibleProperty());
        planetLabLoginPanel.managedProperty().bind(planetLabLoginPanel.visibleProperty());
        defaultLoginPanel.managedProperty().bind(defaultLoginPanel.visibleProperty());

        planetlabPasswordBox.managedProperty().bind(planetlabPasswordBox.visibleProperty());
        planetlabCreateCertificateButton.managedProperty().bind(planetlabCreateCertificateButton.visibleProperty());
        planetlabCertificateFeedbackLabel.managedProperty().bind(planetlabCertificateFeedbackLabel.visibleProperty());

        errorLog.setMaxHeight(Control.USE_PREF_SIZE);

        defaultLoginPanelPemField.textProperty().addListener((observableValue, s, s1) -> {
            if (blockModelToGUI_eventhelper) return;

            onPemEdit();
        });

        defaultLoginPanelSliceAuthNameChooser.selectedAuthorityProperty().addListener((observableValue, oldAuthInfo, newAuthInfo) -> {
            if (blockModelToGUI_eventhelper) return;

            userLoginModel.getKeyCertWithManualInfoUserLoginModel().setAuthority(newAuthInfo.getSfaAuthority());

            if (!userLoginModel.getKeyCertWithManualInfoUserLoginModel().isValidUserUrn()) {
                GeniUrn authUrn = GeniUrn.parse(newAuthInfo.getSfaAuthority().getUrnString());
                if (authUrn != null) {
                    GeniUrn templateUrn = GeniUrn.createGeniUrnFromEncodedParts(authUrn.getEncodedTopLevelAuthority(), "user", "<username>");
                    defaultLoginPanelUserUrnField.setText(templateUrn.toString());
                }
            }
        });

        defaultLoginPanelUserUrnField.textProperty().addListener((observableValue, oldUrn, newUrn) -> {
            if (blockModelToGUI_eventhelper) return;

            userLoginModel.getKeyCertWithManualInfoUserLoginModel().setUserUrn(newUrn);
            if (userLoginModel.getKeyCertWithManualInfoUserLoginModel().deriveAuthorityFromUrn()) {
                SfaAuthority auth = userLoginModel.getKeyCertWithManualInfoUserLoginModel().getUserAuthority();
                if (auth != null) {
                    boolean blockModelToGUI_eventhelper_origState = blockModelToGUI_eventhelper;
                    blockModelToGUI_eventhelper = true;
                    defaultLoginPanelSliceAuthNameChooser.select(auth);
                    blockModelToGUI_eventhelper = blockModelToGUI_eventhelper_origState;
                }
            }
        });


        planetlabAuthChooser.selectedAuthorityProperty().addListener((observableValue, oldAuthInfo, newAuthInfo) -> {
            if (blockModelToGUI_eventhelper) return;
            PlanetlabUserLoginModel m = userLoginModel.getPlanetlabUserLoginModel();

            m.setAuthority(newAuthInfo.getSfaAuthority());
            modelToGUI_derived();
        });

        planetlabSfaHrnField.textProperty().addListener((observableValue, oldHrn, newHrn) -> {
            if (blockModelToGUI_eventhelper) return;

            PlanetlabUserLoginModel m = userLoginModel.getPlanetlabUserLoginModel();
            assert m != null;
            m.setPlanetlabSfaHrn(newHrn);

            //might have updated user authority
//                    assert planetlabAuthChooser.getSelectedAuthority() != null;
//                    assert planetlabAuthChooser.getSelectedAuthority().getSfaAuthority() != null;
            if (m.getUserAuthority() != null &&
                    (planetlabAuthChooser.getSelectedAuthority() == null ||
                            m.getUserAuthority() != planetlabAuthChooser.getSelectedAuthority().getSfaAuthority()
                    )
                    ) {
                boolean blockModelToGUI_eventhelper_origState = blockModelToGUI_eventhelper;
                blockModelToGUI_eventhelper = true;
                planetlabAuthChooser.select(m.getUserAuthority());
                blockModelToGUI_eventhelper = blockModelToGUI_eventhelper_origState;
            }

            modelToGUI_derived();
        });
    }

    public void setRequireAllUserInfo(boolean requireAllUserInfo) {
        this.requireAllUserInfo = requireAllUserInfo;
    }

    @FXML
    private Parent defaultLoginPanel;
    @FXML
    private Parent planetLabLoginPanel;

    @FXML
    private RadioButton pemKeyCertInputFromFileRadioButton;
    @FXML
    private RadioButton pemKeyCertInputFromStringRadioButton;
    @FXML
    private Button pemKeyCertInputFromFileLoadButton;
    @FXML
    private HBox defaultLoginPanelPasswordBox;
    @FXML
    private Button defaultLoginPanelPasswordButton;
    @FXML
    private TextArea defaultLoginPanelPemField;
    @FXML
    private TextField defaultLoginPanelPassword;
    @FXML
    private TextField pemKeyCertInputFromFileField;
    @FXML
    private HBox pemKeyCertInputFromFileBox;

    @FXML
    private Label defaultLoginPanelSliceAuthName;
    @FXML
    private Label defaultLoginPanelSliceAuthUrn;
    @FXML
    private Label defaultLoginPanelUsername;
    @FXML
    private Label defaultLoginPanelUserUrn;

    @FXML
    private Button loginButton;
    @FXML
    private Button logoutButton;
    @FXML
    private Label errorLogLabel;
    @FXML
    private TextArea errorLog;


    @FXML
    private Label missingInfoWarningLabel;
    @FXML
    private Label missingInfoErrorLabel;


    @FXML
    private Parent derivedCertificateDetails;
    @FXML
    private Parent manualCertificateDetails;

    @FXML
    private AuthorityChooser defaultLoginPanelSliceAuthNameChooser;
    @FXML
    private TextField defaultLoginPanelUserUrnField;

    @FXML
    private RadioButton loginTypeKeyCertIntRadioButton;
    @FXML
    private RadioButton loginTypeKeyCertExtRadioButton;
    @FXML
    private RadioButton loginTypePlanetlabRadioButton;


    @FXML
    private AuthorityChooser planetlabAuthChooser;
    @FXML
    private Label planetlabAuthFeedbackLabel;
    @FXML
    private TextField planetlabSfaHrnField;
    @FXML
    private TextField planetlabFileField;

    @FXML
    private Parent planetlabPasswordBox;
    @FXML
    private PasswordField planetlabPasswordField;

    @FXML
    private Label planetlabCertificateFeedbackLabel;
    @FXML
    private Button planetlabCreateCertificateButton;

    @Inject
    private Provider<AMList> amListProvider;


    @FXML
    private void selectLoginType() {
        if (blockModelToGUI_eventhelper) return;

        UserLoginModelManager.UserLoginModelType origType = userLoginModel.getUserLoginModelType();

        KeyCertUserLoginModel mInt = userLoginModel.getKeyCertUserLoginModel();
        KeyCertWithManualInfoUserLoginModel mExt = userLoginModel.getKeyCertWithManualInfoUserLoginModel();

        if (loginTypeKeyCertIntRadioButton.isSelected()) {
            userLoginModel.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);

            if (origType == UserLoginModelManager.UserLoginModelType.KEY_CERT_EXTERNAL_INFO) {
                //convert
                try {
                    if (mExt.isFromFile())
                        mInt.setKeyCertPemFile(mExt.getKeyCertFile());
                    else
                        mInt.setKeyCertPemString(mExt.getKeyCertContent());
                } catch (InvalidLoginException ex) {
                    showError(ex);
                }
            }
        }
        if (loginTypeKeyCertExtRadioButton.isSelected()) {
            userLoginModel.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_EXTERNAL_INFO);

            if (origType == UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO) {
                //convert
                try {
                    if (mInt.isFromFile())
                        mExt.setKeyCertPemFile(mInt.getKeyCertFile());
                    else
                        mExt.setKeyCertPemString(mInt.getKeyCertContent());

                    clearError();
                } catch (InvalidLoginException ex) {
                    showError(ex);
                }
                mExt.setAuthority(mInt.getUserAuthority());
                mExt.setUserUrn(mInt.getUserUrnString());
            }
        }
        if (loginTypePlanetlabRadioButton.isSelected())
            userLoginModel.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);

        modelToGUI();
    }

    private void clearError() {
        errorLogLabel.setVisible(false);
        errorLog.setVisible(false);

        errorLog.setText("");
    }

    private void showLoginType(UserLoginModelManager.UserLoginModelType type) {
        switch (type) {
            case KEY_CERT_INTERNAL_INFO: {
                defaultLoginPanel.setVisible(true);
                planetLabLoginPanel.setVisible(false);
                derivedCertificateDetails.setVisible(true);
                manualCertificateDetails.setVisible(false);
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                defaultLoginPanel.setVisible(true);
                planetLabLoginPanel.setVisible(false);
                derivedCertificateDetails.setVisible(false);
                manualCertificateDetails.setVisible(true);
                break;
            }
            case PLANETLAB: {
                defaultLoginPanel.setVisible(false);
                planetLabLoginPanel.setVisible(true);
                break;
            }
        }
    }

    /**
     * Copy settings from userLoginModel to GUI
     */
    private boolean blockModelToGUI_eventhelper = false;

    protected void modelToGUI() {
        if (blockModelToGUI_eventhelper) return;
        boolean blockModelToGUI_eventhelper_origState = blockModelToGUI_eventhelper;
        blockModelToGUI_eventhelper = true; //don't let it trigger modelToGUI() again, as we're doing it already

        showLoginType(userLoginModel.getUserLoginModelType());
        modelToGUI_input();
        modelToGUI_derived();
        blockModelToGUI_eventhelper = blockModelToGUI_eventhelper_origState;
    }

    protected void modelToGUI_input() {
        if (userLoginModel == null) return;

        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                loginTypeKeyCertIntRadioButton.setSelected(true);
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                loginTypeKeyCertExtRadioButton.setSelected(true);
                break;
            }
            case PLANETLAB: {
                loginTypePlanetlabRadioButton.setSelected(true);
                break;
            }
        }

        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_EXTERNAL_INFO: //fall-through
            case KEY_CERT_INTERNAL_INFO: {
                KeyCertUserLoginModel m;
                switch (userLoginModel.getUserLoginModelType()) {
                    case KEY_CERT_INTERNAL_INFO: {
                        m = userLoginModel.getKeyCertUserLoginModel();
                        break;
                    }
                    case KEY_CERT_EXTERNAL_INFO: {
                        m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();
                        break;
                    }
                    default:
                        return;
                }
                assert userLoginModel.getCurrentUserLoginModel() == m :
                        "userLoginModel.getCurrentUserLoginModel() class="
                                + userLoginModel.getCurrentUserLoginModel().getClass().getName() + " m class=" + m.getClass().getName();

                if (m.isFromFile()) {
                    String pemkeycertfilename = m.getKeyCertFile() == null ? null : m.getKeyCertFile().getPath();
                    pemKeyCertInputFromFileField.setText(pemkeycertfilename);
                    pemKeyCertInputFromFileRadioButton.setSelected(true);
                    defaultLoginPanelPemField.setEditable(false);
                } else {
                    pemKeyCertInputFromFileField.setText("");
                    pemKeyCertInputFromStringRadioButton.setSelected(true);
                    defaultLoginPanelPemField.setEditable(true);
                }

                pemKeyCertInputFromFileBox.setVisible(m.isFromFile());

                if (m.getKeyCertContent() != null)
                    defaultLoginPanelPemField.setText(m.getKeyCertContent());
                else
                    defaultLoginPanelPemField.setText("");
                break;
            }
            case PLANETLAB: {
                PlanetlabUserLoginModel m = userLoginModel.getPlanetlabUserLoginModel();
                if (m.getUserAuthority() != null) {
                    boolean blockModelToGUI_eventhelper_origState = blockModelToGUI_eventhelper;
                    blockModelToGUI_eventhelper = true;
                    planetlabAuthChooser.select(m.getUserAuthority());
                    blockModelToGUI_eventhelper = blockModelToGUI_eventhelper_origState;
                }
//                else
//                    planetlabAuthChooser.select(null); //not allowed: must always select something
                planetlabFileField.setText(m.getSshPrivateKeyFile() != null ? m.getSshPrivateKeyFile().getPath() : "");
                boolean blockModelToGUI_eventhelper_origState = blockModelToGUI_eventhelper;
                blockModelToGUI_eventhelper = true;
                planetlabSfaHrnField.setText(m.getSfaHrn() != null ? m.getSfaHrn() : "");
                blockModelToGUI_eventhelper = blockModelToGUI_eventhelper_origState;
                break;
            }
        }

    }

    private static enum ValidationStyle {ERROR, WARNING, OK, EXPLICIT_OK}

    private static void setValidationStyle(Parent p, ValidationStyle validationStyle) {
        String styleName;
        switch (validationStyle) {
            case ERROR: {
                styleName = "validation_error";
                break;
            }
            case WARNING: {
                styleName = "validation_warning";
                break;
            }
            case OK: {
                styleName = null;
                break;
            }
            case EXPLICIT_OK: {
                styleName = "validation_explicit_ok";
                break;
            }
            default:
                throw new RuntimeException("Unsupported ValidationStyle");
        }

        List<String> otherStyles = new ArrayList<>();
        otherStyles.add("validation_error");
        otherStyles.add("validation_warning");
        otherStyles.add("validation_explicit_ok");

        if (styleName != null)
            otherStyles.remove(styleName);

        p.getStyleClass().removeAll(otherStyles);
        if (styleName != null)
            if (!p.getStyleClass().contains(styleName))
                p.getStyleClass().add(styleName);
    }

    protected void modelToGUI_derived() {
        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                KeyCertUserLoginModel m = userLoginModel.getKeyCertUserLoginModel();

                defaultLoginPanelUserUrn.setText(m.getUserUrnString());
                GeniUrn userUrn = GeniUrn.parse(m.getUserUrnString());
                if (userUrn != null)
                    defaultLoginPanelUsername.setText(userUrn.getEncodedResourceName());
                else
                    defaultLoginPanelUsername.setText("");
                defaultLoginPanelSliceAuthUrn.setText(m.getUserAuthorityUrn());
                if (m.getUserAuthority() != null)
                    defaultLoginPanelSliceAuthName.setText(m.getUserAuthority().getName());
                else {
                    defaultLoginPanelSliceAuthName.setText("Authority not known");
                }
                defaultLoginPanelPasswordBox.setVisible(m.isPasswordRequired());
                loginButton.setVisible(!m.isPasswordRequired());
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                KeyCertWithManualInfoUserLoginModel m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();

                defaultLoginPanelUserUrnField.setText(m.getUserUrnString());

                defaultLoginPanelSliceAuthNameChooser.select(m.getUserAuthority());

                defaultLoginPanelPasswordBox.setVisible(m.isPasswordRequired());
                loginButton.setVisible(!m.isPasswordRequired());
                break;
            }
            case PLANETLAB: {
                loginButton.setVisible(true);
                PlanetlabUserLoginModel m = userLoginModel.getPlanetlabUserLoginModel();

                //set key correctness, password label, hrn correctness, authority correctness and details and show planetlabCreateCertificateButton
                if (m.correctPlanetlabSfaUrn())
                    setValidationStyle(planetlabSfaHrnField, ValidationStyle.EXPLICIT_OK);
                else
                    setValidationStyle(planetlabSfaHrnField, ValidationStyle.ERROR);

                boolean privatekeyReadable = m.getPrivateKeyContent() != null;
                boolean privatekeyUnlocked = m.getPrivateKey() != null;
                if (privatekeyReadable) {
                    if (privatekeyUnlocked) {
                        setValidationStyle(planetlabFileField, ValidationStyle.EXPLICIT_OK);
                        planetlabPasswordBox.setVisible(false);
                    } else {
                        setValidationStyle(planetlabFileField, ValidationStyle.WARNING);
                        planetlabPasswordBox.setVisible(true);
                    }
                } else {
                    setValidationStyle(planetlabFileField, ValidationStyle.ERROR);
                    planetlabPasswordBox.setVisible(false);
                }

                planetlabCreateCertificateButton.setVisible(true);
                if (m.isReadyToFetchCertificate()) {
                    planetlabCreateCertificateButton.setDisable(false);
//                    planetlabCreateCertificateButton.setVisible(true);
                } else {
                    planetlabCreateCertificateButton.setDisable(true);
//                    planetlabCreateCertificateButton.setVisible(false);
                }

                if (m.getClientCertificateChain() == null) {
                    if (m.isReadyToFetchCertificate()) {
                        planetlabCertificateFeedbackLabel.setText("No certificate - All data available to fetch certificate");
                        setValidationStyle(planetlabCertificateFeedbackLabel, ValidationStyle.WARNING);
                    } else {
                        planetlabCertificateFeedbackLabel.setText("No certificate - Missing data needed to fetch certificate");
                        setValidationStyle(planetlabCertificateFeedbackLabel, ValidationStyle.ERROR);
                    }
                } else {
                    planetlabCertificateFeedbackLabel.setText("Certificate present");
                    setValidationStyle(planetlabCertificateFeedbackLabel, ValidationStyle.EXPLICIT_OK);
                }

                boolean correctAuth = m.correctAuthority();
                if (correctAuth) {
                    planetlabAuthFeedbackLabel.setText("OK");
                    setValidationStyle(planetlabAuthFeedbackLabel, ValidationStyle.EXPLICIT_OK);
                } else {
                    setValidationStyle(planetlabAuthFeedbackLabel, ValidationStyle.ERROR);
                    if (m.getUserAuthority() == null)
                        planetlabAuthFeedbackLabel.setText("no selection");
                    else
                        planetlabAuthFeedbackLabel.setText("Authority is missing URL for Planetlab Sfa Registry");
                }

                break;
            }
        }

        UserLoginModel m = userLoginModel.getCurrentUserLoginModel();
        boolean missingInfo = m.getUserAuthority() == null || m.getUserUrnString() == null;
        if (requireAllUserInfo) {
            missingInfoWarningLabel.setVisible(false);
            missingInfoErrorLabel.setVisible(m.isUserLoggedIn() && missingInfo);
        } else {
            missingInfoWarningLabel.setVisible(m.isUserLoggedIn() && missingInfo);
            missingInfoErrorLabel.setVisible(false);
        }

        updateErrorLog();
        updateLoginButtonEnabled();
    }

    public void updateFileOrString() {
        LOG.trace("updateFileOrString blockModelToGUI_eventhelper=" + blockModelToGUI_eventhelper + "  modelType=" + userLoginModel.getUserLoginModelType());
        if (blockModelToGUI_eventhelper) return;

        KeyCertUserLoginModel m;
        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                m = userLoginModel.getKeyCertUserLoginModel();
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();
                break;
            }
            default:
                return;
        }
        assert userLoginModel.getCurrentUserLoginModel() == m;

        assert pemKeyCertInputFromFileRadioButton.isSelected() != pemKeyCertInputFromStringRadioButton.isSelected();

        boolean newIsfromFile = pemKeyCertInputFromFileRadioButton.isSelected();

        if (newIsfromFile) {
            if (m.isFromString()) {
                try {
                    m.setKeyCertPemFile(null);
                } catch (InvalidLoginException e) {
                    showError(e);
                }
            }
        } else {
            if (m.isFromFile()) {
                try {
                    m.setKeyCertPemString(defaultLoginPanelPemField.getText());
                } catch (InvalidLoginException e) {
                    showError(e);
                }
            }
        }

        modelToGUI();
    }

    private void showError(Exception e) {
        showError(e.getMessage());

    }

    public void onPemEdit() {
        if (blockModelToGUI_eventhelper) return;

        KeyCertUserLoginModel m;
        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                m = userLoginModel.getKeyCertUserLoginModel();
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();
                break;
            }
            default:
                return;
        }
        assert userLoginModel.getCurrentUserLoginModel() == m;

        if (m.isFromString()) {
            String text = defaultLoginPanelPemField.getText();
            assert text != null;
            try {
                m.setKeyCertPemString(text);
            } catch (InvalidLoginException e) {
                showError(e);
            }
            modelToGUI();
        } else
            LOG.warn("assumption bug in onPemEdit()");
    }

    public void showCertificateDetailsWindow() {
        KeyCertUserLoginModel m;
        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                m = userLoginModel.getKeyCertUserLoginModel();
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();
                break;
            }
            default:
                return;
        }
        assert userLoginModel.getCurrentUserLoginModel() == m;

        String keyCertContent = m.getKeyCertContent();
        if (keyCertContent != null) {
            X509CertificatePanel.showX509Certificate(keyCertContent);
        }
    }

    public void unlockAndLogin() {
        boolean unlocked = unlock();
        if (unlocked && loginAllowed())
            login();
        else {
            UserLoginModel m = userLoginModel.getCurrentUserLoginModel();
            LOG.debug("userLoginModel.unlockAndLogin() not successful unlocked={}\n" +
                            "\tuserLoginModel.getCurrentUserLoginModel().getClass().getName()={}\n" +
                            "\tm.isUserLoggedIn()={}\n" +
                            "\tm.getUserAuthority() == null={}\n" +
                            "\tm.getUserUrn() == null={}\n" +
                            "\trequireAllUserInfo={}", unlocked, userLoginModel.getCurrentUserLoginModel().getClass().getName()
                    , m.isUserLoggedIn(), (m.getUserAuthority() == null), (m.getUserUrnString() == null),
                    requireAllUserInfo);
        }
    }

    public boolean unlock() {
        final char[] password = defaultLoginPanelPassword.getText().toCharArray();

        KeyCertUserLoginModel m;
        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                m = userLoginModel.getKeyCertUserLoginModel();
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();
                break;
            }
            default:
                return false;
        }
        assert userLoginModel.getCurrentUserLoginModel() == m;

        boolean success = m.unlock(password);
        updateLoginButtonEnabled();
        updateErrorLog();
        return success;
    }

    public void planetlabFetchCertificate() {
        PlanetlabUserLoginModel m = userLoginModel.getPlanetlabUserLoginModel();

        //recreate certificate if exists   if (m.getCertificate() != null) return;

        if (m.isPasswordRequired() && m.getPrivateKey() == null && m.getPrivateKeyContent() != null) {
            if (!planetLabUnlock()) return;
        }

        if (!m.isReadyToFetchCertificate()) return;

        try {
            m.fetchCertificate();
        } catch (PlanetlabUserLoginModel.CertificateFetchException e) {
            showError(e);
        }

        modelToGUI_derived();
    }

//    public void planetLabUnlockAndLogin() {
//        if (planetLabUnlock() && loginAllowed())
//            login();
//    }

    public boolean planetLabUnlock() {
        final char[] password = planetlabPasswordField.getText().toCharArray();
        boolean success = userLoginModel.getPlanetlabUserLoginModel().unlock(password);
        modelToGUI_derived();
        return success;
    }

    private boolean loginAllowed() {
        UserLoginModel m = userLoginModel.getCurrentUserLoginModel();

        if (!m.isUserLoggedIn()) return false; // is bare minimum supplied?

        boolean missingInfo = m.getUserAuthority() == null || m.getUserUrnString() == null;
        return !missingInfo || !requireAllUserInfo;
    }

    public void updateLoginButtonEnabled() {
        loginButton.setDisable(!loginAllowed());
    }

    public void updateErrorLog() {
        KeyCertUserLoginModel m;

        switch (userLoginModel.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                m = userLoginModel.getKeyCertUserLoginModel();
                break;
            }
            case KEY_CERT_EXTERNAL_INFO: {
                m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();
                break;
            }
            default:
                return;
        }
        assert userLoginModel.getCurrentUserLoginModel() == m;

        assert m != null;
    }

    private void showError(String error) {
        if (error != null) {
            errorLogLabel.setVisible(true);
            errorLog.setVisible(true);

            errorLog.setText(error);
        } else {
            clearError();
        }
    }

    public void login() {
        LOG.debug("Advanced User Login window \"login\" button pressed: log in");

        if (!loginAllowed()) {

            UserLoginModel m = userLoginModel.getCurrentUserLoginModel();
            LOG.debug("userLoginModel.login() not allowed\n" +
                            "\tuserLoginModel.getCurrentUserLoginModel().getClass().getName()={}\n" +
                            "\tm.isUserLoggedIn()={}\n" +
                            "\tm.getUserAuthority() == null={}\n" +
                            "\tm.getUserUrn() == null={}\n" +
                            "\trequireAllUserInfo={}", userLoginModel.getCurrentUserLoginModel().getClass().getName()
                    , m.isUserLoggedIn(), (m.getUserAuthority() == null), (m.getUserUrnString() == null),
                    requireAllUserInfo);
            return;
        }
        boolean success = userLoginModel.login();
        LOG.trace("success=" + success);
        if (success) {
            if (userLoginModel.hasChanged()) {
                JFDialogs.create()
                        .message("Save login details now? (password is never saved)")
                        .buttonTypes(ButtonType.YES, ButtonType.NO)
                        .showWarning()
                        .filter(response -> response == ButtonType.YES)
                        .ifPresent(response -> userLoginModel.save());
            }

            String problem;
            try {
                for (X509Certificate cert : userLoginModel.getCurrentUserLoginModel().getClientCertificateChain()) {
                    cert.checkValidity();
                }
                problem = null;
            } catch (CertificateExpiredException e) {
                problem = "certificate has expired.";
            } catch (CertificateNotYetValidException e) {
                problem = "certificate is not yet valid";
            }
            if (problem != null) {
                Optional<ButtonType> reply = JFDialogs.create()
                        .message("There is a problem with your certificate, which will likely cause all calls to fail.\n" +
                                "Since this is an advanced login dialog, you can continue if you really wish.\n" +
                                "The problem is: " + problem + "\n" +
                                "\n" +
                                "Do you really want to continue login?")
                        .masthead("Warning: "+problem)
                        .title("Certificate problem")
                        .showConfirm();
                if (!reply.isPresent() && reply.get() != ButtonType.YES) {
                    success = true;
                } else {
                    success = false;
                }
            }

            if (success)
                Platform.runLater(stage::close);
        } else {

            UserLoginModel m = userLoginModel.getCurrentUserLoginModel();
            LOG.debug("userLoginModel.login() not successful\n" +
                            "\tuserLoginModel.getCurrentUserLoginModel().getClass().getName()={}\n" +
                            "\tm.isUserLoggedIn()={}\n" +
                            "\tm.getUserAuthority() == null={}\n" +
                            "\tm.getUserUrn() == null={}\n" +
                            "\trequireAllUserInfo={}", userLoginModel.getCurrentUserLoginModel().getClass().getName()
                    , m.isUserLoggedIn(), (m.getUserAuthority() == null), (m.getUserUrnString() == null),
                    requireAllUserInfo);
        }

        LOG.debug("Advanced User Login window \"login\" button handled: log in success=" + success);
        fireUserLoginControllerListeners(success);
    }

    public void nologin() {
        LOG.debug("Advanced User Login window \"skip login\" button pressed: skipping login");
        userLoginModel.logout();
        Platform.runLater(stage::close);
        fireUserLoginControllerListeners(false);
    }


    public void saveConfig() {
        //Note: password is never saved
        userLoginModel.save();
    }

    public void loadConfig() {
        userLoginModel.load();
        modelToGUI();
    }

    public void resetConfig() {
        userLoginModel.reset();
        modelToGUI();
    }

    @FXML
    private void loadCertificateFile() {
        FileChooser fileChooser = new FileChooser();
        File file = fileChooser.showOpenDialog(null);

        if (file != null) {
            KeyCertUserLoginModel m;
            switch (userLoginModel.getUserLoginModelType()) {
                case KEY_CERT_INTERNAL_INFO: {
                    m = userLoginModel.getKeyCertUserLoginModel();
                    break;
                }
                case KEY_CERT_EXTERNAL_INFO: {
                    m = userLoginModel.getKeyCertWithManualInfoUserLoginModel();
                    break;
                }
                default:
                    return;
            }
            assert userLoginModel.getCurrentUserLoginModel() == m;

            try {
                m.setKeyCertPemFile(file);
                clearError();
            } catch (InvalidLoginException e) {
                showError(e);
            }
        }

        modelToGUI();
    }

    @FXML
    private void loadPlanetlabFile() {
        FileChooser fileChooser = new FileChooser();
        File file = fileChooser.showOpenDialog(null);

        if (file != null) {
            try {
                userLoginModel.getPlanetlabUserLoginModel().setSshPrivateKeyFile(file);
                clearError();
            } catch (InvalidLoginException e) {
                showError(e);
            }
        }

        modelToGUI();
    }


    @FXML
    private void showAbout() {
        //TODO either remove about button or add about panel
//        AboutBoxController.showAbout();
    }

    @FXML
    private void editTestbeds() {
        amListProvider.get().show();
    }

    @FXML
    private void clearJfedDir() {
        JFedCommonDialogs.requestEraseJFedConfiguration(loginButton.getScene().getWindow());
    }


    private final List<UserLoginControllerListener> loginControllerListenerList = new ArrayList<>();

    public void addListener(UserLoginControllerListener l) {
        loginControllerListenerList.add(l);
    }

    public void removeListener(UserLoginControllerListener l) {
        loginControllerListenerList.remove(l);
    }

    void fireUserLoginControllerListeners(boolean successfulLogin) {
        for (UserLoginControllerListener l : loginControllerListenerList)
            l.onDone(successfulLogin);
    }

    public static interface UserLoginControllerListener {
        public void onDone(boolean successfulLogin);
    }
}
