package be.iminds.ilabt.jfed.ui.cli;

import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetailsWriter;
import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.ResultListener;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv2Wrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv3Wrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.impl.BasicExecuteAnsiblePlaybook;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.RspecXmlConstants;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecMerger;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.*;
import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.joran.JoranConfigurator;
import ch.qos.logback.core.joran.spi.JoranException;
import ch.qos.logback.core.util.StatusPrinter;
import com.google.inject.Injector;
import javanet.staxutils.IndentingXMLEventWriter;
import org.apache.commons.cli.*;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.FileConfiguration;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.slf4j.LoggerFactory;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.Charset;
import java.security.PublicKey;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import static java.time.format.DateTimeFormatter.ISO_OFFSET_DATE_TIME;

/**
 * ExperimenterCli
 */
@SuppressWarnings("AccessStaticViaInstance")
public class ExperimenterCli {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ExperimenterCli.class);

    private static final String KEY_EXPIRATION_TIME = "expiration-time";
    private static final String KEY_REQUEST_RSPEC = "request-rspec";

    private static final String CHOOSE_PROJECT_AUTOMATICALLY_OPTION = "CHOOSE_AUTOMATICALLY";

    public static void commandHelp(final PrintStream err, Options options, String command) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(200, "jfed-experimenter-cli <command> [command_options ... ]", "Available commands: create,delete,status,renew,manifest,userinfo,sliceinfo,createslice\n" + command + " options:", options,
                "\n\nTemplate Context Properties File:\n  username = <username>\n" +
                        "  passwordFilename = <filename of file containing password>\n" +
                        "  pemKeyAndCertFilename = <filename of file containing user certificate and private key in PEM format>\n" +
                        "  userAuthorityUrn = <URN of test user's authority>\n"
        );
    }

    public static void showVersion(PrintStream out) {
        out.println("jFed Experimenter CLI -- http://jfed.iminds.be/");
        out.println("Send bugreports to: jfed-bugreports@atlantis.ugent.be");

        final String VERSION_PROPERTIES_FILE = "/version.properties";
        final String PROP_VERSION = "version";
        final String PROP_GIT_COMMIT = "build.git_commit";
        final String PROP_GIT_BRANCH = "build.git_branch";
        final String PROP_BUILD_JENKINS = "build.jenkins_build";

        Properties prop = new Properties();
        //now load the version-properties file
        try {
            prop.load(ExperimenterCli.class.getResourceAsStream(VERSION_PROPERTIES_FILE));
        } catch (IOException e) {
            throw new RuntimeException("Something went wrong while retrieving the version info: " + e.getMessage(), e);
        }

        StringBuilder version = new StringBuilder();
        version.append(prop.getProperty(PROP_VERSION, "Unknown"));
        final String jenkinsBuildNr = prop.getProperty(PROP_BUILD_JENKINS);
        if (jenkinsBuildNr != null && !jenkinsBuildNr.isEmpty())
            version.append("  - build #").append(jenkinsBuildNr);

        final String gitCommit = prop.getProperty(PROP_GIT_COMMIT);
        if (gitCommit != null && !gitCommit.isEmpty())
            version.append("  - git commit #").append(gitCommit);

        final String gitBranch = prop.getProperty(PROP_GIT_BRANCH);
        if (gitBranch != null && !gitBranch.isEmpty())
            version.append(" on ").append(gitBranch);
        out.println("Version: " + version.toString());

        out.println("Environment: "
                + System.getProperty("os.name") + " " + System.getProperty("os.version") + " " + System.getProperty("os.arch")
                + " - Java " + System.getProperty("java.version") + " (" + System.getProperty("java.vendor") + ")");
    }

    public static void generalHelp(final PrintStream err) {
        err.println("Syntax: jfed-experimenter-cli <command> [command_options ... ]\n" +
                "Available Commands: create,delete,status,renew,userinfo,sliceinfo,createslice\n");
    }

    public static void main(String[] args) throws IOException, ClassNotFoundException {
        int retval = main_helper(args, System.out, System.err, System.in, new BaseCli(System.err), true);
        System.exit(retval);
    }

    public static int main_helper(String[] args,
                                  final PrintStream out, final PrintStream err, final InputStream in,
                                  BaseCli baseCli,
                                  boolean temporarilyRedirectSystemStreams) throws IOException, ClassNotFoundException {
        final PrintStream orig_out = System.out;
        final PrintStream orig_err = System.err;
        final InputStream orig_in = System.in;
        if (temporarilyRedirectSystemStreams) {
            System.setOut(out);
            System.setErr(err);
            System.setIn(in);
        }

        int retval = -1;
        try {
            retval = main_internal(args, out, err, in, baseCli);
        } finally {
            if (temporarilyRedirectSystemStreams) {
                System.setOut(orig_out);
                System.setErr(orig_err);
                System.setIn(orig_in);
            }
        }
        return retval;
    }

    public static int main_internal(String[] args,
                                    final PrintStream out, final PrintStream err, final InputStream in,
                                    BaseCli baseCli) throws IOException, ClassNotFoundException {
        String command;
        boolean isCreate = false;
        boolean isRenew = false;
        boolean isDelete = false;
        boolean isPoa = false;
        boolean isStatus = false;
        boolean isManifest = false;
        boolean isShowSliceInfo = false;
        boolean isCreateSliceOnly = false;
        boolean isListUserInfo = false;

        LinkedList<String> argList = new LinkedList<>(Arrays.asList(args));
        if (argList.isEmpty()) {
            generalHelp(err);
            return 3;
        } else {
            if (argList.get(0).equals("version") || argList.get(0).equals("--version") || argList.get(0).equals("-version") || argList.get(0).equals("-v")) {
                showVersion(out);
                return 0;
            } else {
                command = argList.removeFirst();
                if (command.equalsIgnoreCase("create")) isCreate = true;
                if (command.equalsIgnoreCase("createsliver")) isCreate = true;
                if (command.equalsIgnoreCase("createslivers")) isCreate = true;
                if (command.equalsIgnoreCase("renew")) isRenew = true;
                if (command.equalsIgnoreCase("delete")) isDelete = true;
                if (command.equalsIgnoreCase("status")) isStatus = true;
                if (command.equalsIgnoreCase("manifest")) isManifest = true;
                if (command.equalsIgnoreCase("userinfo")) isListUserInfo = true;
                if (command.equalsIgnoreCase("poa")) isPoa = true;

                Pattern isShowSliceCredentialPattern = Pattern.compile("(show(-)?)?(slice(-)?)?((cred(ential)?)|(info(rmation)?))");
                if (isShowSliceCredentialPattern.matcher(command).matches()) isShowSliceInfo = true;

                Pattern isCreateSlicePattern = Pattern.compile("(create(-)?)?slice((-)?only)?");
                if (isCreateSlicePattern.matcher(command).matches()) isCreateSliceOnly = true;

                if (command.equalsIgnoreCase("performOperationalAction")) isPoa = true;
                if (command.equalsIgnoreCase("perform-operational-action")) isPoa = true;
                if (command.equalsIgnoreCase("perform_operational_action")) isPoa = true;

                if (!isCreate && !isRenew && !isStatus && !isManifest && !isDelete && !isListUserInfo && !isPoa && !isCreateSliceOnly && !isShowSliceInfo) {
                    err.println("ERROR: unknown command \"" + command + "\"");
                    generalHelp(err);
                    return 3;
                }
            }
        }

        CommandLineParser parser = new BasicParser();

        boolean requiresSlice = isCreate || isDelete || isRenew || isStatus || isManifest || isPoa || isShowSliceInfo || isCreateSliceOnly;

        List<String> actionOptions = Arrays.asList("ReloadOS", "Restart", "ConsoleUrl");
        String actionOptionsString = "";
        for (String a : actionOptions) {
            if (!actionOptionsString.isEmpty())
                actionOptionsString += ", ";
            actionOptionsString += a;
        }

        // create the Options
        Options options = new Options();
//        options.addOption(OptionBuilder.withLongOpt("context-file")
//                .withDescription("The properties file containing context details (only the login info in the context file is used by this tool)")
//                .hasArg()
//                .withArgName("CONTEXT PROPERTIES FILE")
//                .create('c'));
//
//        options.addOption(OptionBuilder.withLongOpt("cert-and-key-file")
//                .withDescription("The file containing the user certificate and private key (both in PEM format)")
//                .hasArg()
//                .withArgName("PEM FILE")
//                .create('p'));
//        options.addOption(OptionBuilder.withLongOpt("private-key-password")
//                .withDescription("The password of the private key. Only used if private key is password protected. Default: interactively ask password when needed")
//                .hasArg()
//                .withArgName("CLEARTEXT PASSWORD")
//                .create('P'));
//        //TODO add options to specify username and user authority if not in certificate
        BaseCli.addUserConfig(options, "The properties file containing context details (only the login info in the context file is used by this tool)", false);

        BaseCli.addAuthoritiesConfig(options);
        if (requiresSlice)
            options.addOption(OptionBuilder.withLongOpt("slice")
                    .withDescription("The URN or name of the slice to use. (auto detected)")
                    .hasArg()
                    .withArgName("SLICE URN OR NAME")
                    .isRequired()
                    .create('s'));
        options.addOption(OptionBuilder.withLongOpt("speaks-for")
                .withDescription("The speaksFor certificate (in a file), enabling you make calls on behalf of another user.")
                .hasArg()
                .withArgName("CERTIFICATE FILENAME")
                .create());
        if (requiresSlice)
            options.addOption(OptionBuilder.withLongOpt("project-name")
                    .withDescription("The name of the project (= sub authority) of the slice. This is an optional argument (however some authorities might require a project!). " +
                            "You can also use \"" + CHOOSE_PROJECT_AUTOMATICALLY_OPTION + "\" as project name, in that case, the last (determined by your SA) project you are a member of will be used.")
                    .hasArg()
                    .withArgName("PROJECT NAME")
                    .create('S'));
        if (isCreate || isRenew || isCreateSliceOnly) {
            options.addOption(OptionBuilder.withLongOpt("expiration-date")
                    .withDescription("The date and time at which this slice expires. In RFC3339 format. This is an optional argument. default: 2 hours in the future")
                    .hasArg()
                    .withArgName("RFC3339 DATE")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("expiration-hours")
                    .withDescription("The number of hours after which this slice expires. This is an optional argument. default: 2 hours in the future")
                    .hasArg()
                    .withArgName("INTEGER")
                    .create('e'));
        }
        if (isCreate || isCreateSliceOnly) {
            options.addOption(OptionBuilder.withLongOpt("share-slice")
                    .withDescription("List of users to share slice with. Either use the URN of each user, or the short username. " +
                            "You can also specify the special value \"PROJECTUSERS\", which will automatically share the slice with all users in the project. " +
                            "Multiple users can be specified by separating them with a comma.")
                    .hasArg()
                    .withArgName("USERNAME(S)")
                    .create());
        }
        if (isListUserInfo || isShowSliceInfo || isCreateSliceOnly || isStatus)
            options.addOption(OptionBuilder.withLongOpt("output-format")
                    .withDescription("The output format to use (for user information). Choices: \"text\" or \"json\". Default:json")
                    .hasArg()
                    .withArgName("FORMAT")
                    .create());
        if (isPoa) {
            options.addOption(OptionBuilder.withLongOpt("action")
                    .withDescription("The operational action to perform. (Options: "+actionOptionsString+")")
                    .hasArg()
                    .withArgName("ACTION")
                    .isRequired()
                    .create('a'));
            options.addOption(OptionBuilder.withLongOpt("target-sliver")
                    .withDescription("The URN of the sliver to perform the action on.")
                    .hasArg()
                    .withArgName("SLIVER URN")
                    .isRequired()
                    .create('t'));
        }
        if (isCreate)
            options.addOption(OptionBuilder.withLongOpt("rspec")
                    .withDescription("The rspec file to use for creating a sliver")
                    .hasArg()
                    .withArgName("RSPEC XML FILE")
                    .isRequired()
                    .create('r'));
        if (isStatus || isRenew || isPoa)
            options.addOption(OptionBuilder.withLongOpt("rspec")
                    .withDescription("The rspec file to get the needed authority information from (manifest or request)  (not mandatory, but might not work without)")
                    .hasArg()
                    .withArgName("RSPEC XML FILE")
//                    .isRequired()
                    .create('r'));
        if (isCreate || isManifest)
            options.addOption(OptionBuilder.withLongOpt("ansible-dir")
                    .withDescription("The dir in which to save ansible config files. Will be created if it doesn't exist. Will overwrite any existing files.")
                    .hasArg()
                    .withArgName("DIR")
                    .create());
        if (isCreate || isManifest) {
            options.addOption(OptionBuilder.withLongOpt("ansible-execute-rspec-playbooks")
                    .withDescription("Execute all ansible playbooks found within request Rspec " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() +
                            " elements. (default: ignore them)")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("ansible-add-playbook")
                    .withDescription("Add an ansible playbook to execute. The argument specifies the source, " +
                            "and optionally (after a space) the target (this option can be specified multiple times)")
                    .hasArgs(2)
                    .withArgName(RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart() + " [" +
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + "]")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("ansible-allow-playbook-inputfile")
                    .withDescription("Allow local files in " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + " " +
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart() + ". (default: only allow URLs as " +
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart() + ")")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("ansible-allow-any-playbook-outputfile")
                    .withDescription("Allow any local file as " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + " " +
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + ". (default: only use the basename of the specified " +
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + ", and use the --ansible-dir as directory)")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("ansible-playbook-exe")
                    .withDescription("Executable for the ansible playbook. Default is \"ansible-playbook\" and is usually fine. " +
                            "On some systems might need to be \"/usr/bin/python2 /usr/bin/ansible-playbook\"")
                    .hasArg()
                    .create());
            options.addOption(OptionBuilder.withLongOpt("ansible-debug")
                    .withDescription("Set debug flags when calling ansible playbook.")
                    .create());
        }
        options.addOption(OptionBuilder.withLongOpt("manifest")
                .withDescription("The file in which the manifest must be stored. (default: manifest-<SLICENAME>.rspec)")
                .hasArg()
                .withArgName("FILE")
                .create());
        if (isDelete)
            options.addOption(OptionBuilder.withLongOpt("rspec")
                    .withDescription("The rspec file to get the needed authority information from (manifest or request) (not mandatory, but might not work without)")
                    .hasArg()
                    .withArgName("RSPEC XML FILE")
                    .create('r'));
        if (isCreate) {
            options.addOption(OptionBuilder.withLongOpt("delete-on-create-failure")
                    .withDescription("If there is a failure in a call made to create the sliver(s), delete all resources everywhere before exiting.")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("delete-on-become-ready-failure")
                    .withDescription("If there is a failure in a call while waiting for the sliver(s) to become ready, delete all resources everywhere before exiting.")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("abort-if-slivers-exist")
                    .withDescription("Do not create any slivers if any sliver already exists on any of the authorities.")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("create-slice")
                    .withDescription("Create the slice if it doesn't exist. (if it exists, this option is ignored)")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("rewrite-rspec")
                    .withDescription("Parse the provided RSpec, and reconstruct it again, before sending it to the server. This can help prevent some server sides bug caused by valid but atypical RSpecs.")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("bind-rspec")
                    .withDescription("Bind any unbound nodes in the RSpec to a specified authority. This leaves bound nodes as they are. (this imples --rewrite-rspec)")
                    .hasArg()
                    .withArgName("COMPONENT MANAGER URN")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("ssh-keys")
                    .withDescription("Specify which ssh keys to add. The argument is a (comma separated) list of options.\n" +
                            "Available options: usercert,userkeys,rspec,shareduserallkeys\n" +
                            "   usercert: add the user making the calls, with the ssh key from the certificate in the login PEM file.\n" +
                            "   userkeys: add the user making the calls, with the ssh keys that are stored on the user authority MA server. (usercert and userkeys can be combined to add all keys)\n" +
                            "   shareduserallkeys: add the ssh keys that are stored on the user authority MA server for the users the slice is shared with AND the ssh keys from that user's login certificate.\n" +
                            "   rspec: add the users and keys specified in the RSpec itself.\n" +
                            "This option is optional, default: usercert,rspec,shareduserallkeys")
                    .hasArg()
                    .withArgName("OPTION LIST")
                    .create('k'));
            options.addOption(OptionBuilder.withLongOpt("stitching")
                    .withDescription("If the RSpec requires stitching, allow it. By default, RSpecs that require stitching cause an abort")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("slice-recover-info")
                    .withDescription("A file into which slice recover info will be stored. This can be used by the jFed GUI (if copied to the correct directory).")
                    .hasArg()
                    .withArgName("FILE")
                    .create());
            options.addOption(OptionBuilder.withLongOpt("nowait")
                    .withDescription("Do not wait until sliver is ready (default: wait until ready)")
                    .create());
        }
        options.addOption(OptionBuilder.withLongOpt("call-log")
                .withDescription("A file into which all call details will be stored")
                .hasArg()
                .withArgName("FILE")
                .create());
        options.addOption(OptionBuilder.withLongOpt("print-calls")
                .withDescription("Print all calls to stdout")
                .create());
        options.addOption(OptionBuilder.withLongOpt("am-api")
                .withDescription("The AM Api version to use (\"2\" and \"3\" supported, default: choose automatically)")
                .hasArg()
                .withArgName("AM API VERSION")
                .create("am"));
        options.addOption(OptionBuilder.withLongOpt("fake")
                .withDescription("Do everything, expect actually making the slice and sliver calls. Useful for debugging syntax. (will make calls relating to retrieving user data)")
                .create());
        if (!isCreateSliceOnly && !isShowSliceInfo)
            options.addOption(OptionBuilder.withLongOpt("fake-sliver")
                    .withDescription("Do everything, expect actually making the sliver calls. (will make calls relating to retrieving user data and creating or retrieving slice)")
                    .create());
        options.addOption("q", "quiet", false, "less output");
        options.addOption(null, "silent", false, "less output (same as --quiet)");
        options.addOption("d", "debug", false, "extra debugging output");
        options.addOption("l", "logging", false, "activate logback logging output");
        options.addOption("v", "version", false, "show version and exit");

        CommandLine line;
        try {
            // parse the command line arguments
            String[] leftoverArgs = new String[argList.size()];
            int i = 0;
            for (String arg : argList)
                leftoverArgs[i++] = arg;
            line = parser.parse(options, leftoverArgs);
        } catch (ParseException exp) {
            err.println("Command line argument Syntax error: " + exp.getMessage());
            commandHelp(err, options, command);
            return 3;
        }

        if (line.hasOption("version")) {
            showVersion(out);
            return 0;
        }

        final String sliceIdArg = line.getOptionValue("slice");
        String projectNameArg = line.getOptionValue("project-name");
        final String sliceExpirationDateArg = line.getOptionValue("expiration-date");
        final String sliceExpirationHoursArg = line.getOptionValue("expiration-hours");
        final String rspecArg = line.getOptionValue("rspec");
        final String bindRspecArg = line.getOptionValue("bind-rspec");
        final String shareSliceUsersArg = line.getOptionValue("share-slice");
        final String outputFormatArg = line.getOptionValue("output-format") == null ? "json" : line.getOptionValue("output-format");
        final String amApiVersionArg = line.getOptionValue("am-api");
//        final String testContextFilenameArg = line.getOptionValue("context-file");
//        final String pemFilenameArg = line.getOptionValue("cert-and-key-file");
//        final String privateKeyPassArg = line.getOptionValue("private-key-password");
//        final String testAuthoritiesFilenameArg = line.getOptionValue("authorities-file");
        final String sshKeysArg = line.getOptionValue("ssh-keys");
        String manifestArg = line.getOptionValue("manifest");
        String callLogArg = line.getOptionValue("call-log");
        String sliceRecoverInfoArg = line.getOptionValue("slice-recover-info");
        final boolean createSliceArg = line.hasOption("create-slice");
        final boolean rewriteRspecArg = line.hasOption("rewrite-rspec") || bindRspecArg != null;
        final boolean stitchingArg = line.hasOption("stitching");
        final boolean debug = line.hasOption("debug");
        final boolean printCalls = line.hasOption("print-calls");
        final boolean fake = line.hasOption("fake");
        final boolean logbackLogging = line.hasOption("logging");
        final boolean fakeSliver = line.hasOption("fake-sliver");
        final boolean waitUntilReady = !line.hasOption("nowait");
        final boolean silent = !debug && (line.hasOption("quiet") || line.hasOption("silent"));
        final boolean abortIfSliversExist = line.hasOption("abort-if-slivers-exist");
        final boolean deleteOnCreateFailure = line.hasOption("--delete-on-create-failure");
        final boolean deleteOnBecomeReadyFailure = line.hasOption("--delete-on-become-ready-failure");
        final String[] speaksForArgs = line.getOptionValues("speaks-for");
        final String poaActionArg = line.getOptionValue("action");
        final String poaTargetSliverUrnArg = line.getOptionValue("target-sliver");
        final String ansibleDirName = line.getOptionValue("ansible-dir");

        final String ansiblePlayBookExe = line.getOptionValue("ansible-playbook-exe", "ansible-playbook");
        final boolean ansibleExecuteRspecPlaybooks = line.hasOption("ansible-execute-rspec-playbooks");
        final String[] ansibleExtraPlaybooks = line.getOptionValues("ansible-add-playbook");
        final boolean noPlaybookToExecute = !ansibleExecuteRspecPlaybooks && (ansibleExtraPlaybooks == null || ansibleExtraPlaybooks.length == 0);
        final boolean ansibleAllowPlaybookInputfile = line.hasOption("ansible-allow-playbook-inputfile");
        final boolean ansibleAllowAnyPlaybookOutputFile = line.hasOption("ansible-allow-any-playbook-outputfile");
        final boolean ansibleDebug = line.hasOption("ansible-debug");

        GeniUrn poaTargetSliverUrn = null;

        if (isPoa && poaActionArg != null) {
            boolean validAction = false;
            for (String o : actionOptions) {
                if (o.equalsIgnoreCase(poaActionArg.trim()))
                    validAction = true;
            }
            if (!validAction) {
                err.println("Unsupported action option \""+poaActionArg+"\" specified. " +
                        "Valid options are: "+actionOptionsString);
                commandHelp(err, options, command);
                return 3;
            }
        }
        if (isPoa && poaTargetSliverUrnArg != null) {
            poaTargetSliverUrn = GeniUrn.parse(poaTargetSliverUrnArg);
            if (poaTargetSliverUrn == null) {
                err.println("Invalid URN specified for option target sliver: \""+poaTargetSliverUrnArg+"\"");
                commandHelp(err, options, command);
                return 3;
            }
            if (!poaTargetSliverUrn.getResourceType().equals("sliver")) {
                err.println("\n\n**WARNING** URN specified for option target sliver is not a sliver URN: \""+poaTargetSliverUrnArg+"\"\n" +
                        "            This will be ignored, but will likely not cause the desired results.\n\n");
            }
        }

        File ansibleDir;
        if (ansibleDirName != null) {
            ansibleDir = new File(ansibleDirName);
            if (!ansibleDir.exists() && !ansibleDir.getParentFile().exists()) {
                err.println("Specified ansible-dir does not exist, and neither does parent dir: \""+ansibleDirName+"\"");
                commandHelp(err, options, command);
                return 3;
            }
        } else
            ansibleDir = null;

        if ((!noPlaybookToExecute) && ansibleDirName == null) {
            err.println("the --ansible-execute-rspec-playbooks and --ansible-add-playbook options requires that the "+
                    "ansible-dir option is specified.");
            return 2;
        }

        if (debug) {
            out.println("Showing version because debug is enabled:");
            showVersion(out);
        }

        if (logbackLogging) {
            LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
            loggerContext.reset();
            JoranConfigurator configurator = new JoranConfigurator();
            InputStream configStream = ExperimenterCli.class.getResourceAsStream("/logback_debug.xml");
            configurator.setContext(loggerContext);
            try {
                configurator.doConfigure(configStream); // loads logback file
            } catch (JoranException e) {
                LOG.error("Failed to configure logback", e);
                err.println("Failed to configure logback! "+e.getMessage());
                e.printStackTrace();
            }
            StatusPrinter.printInCaseOfErrorsOrWarnings(loggerContext);

            if (configStream != null)
                configStream.close();
        }

        if (requiresSlice) {
            //should not be required to check this here, as it is a required option, but let's do it anyway to prevent any issues.
            if (sliceIdArg == null) {
                err.println("The slice option is mandatory");
                commandHelp(err, options, command);
                return 3;
            }
        }

        boolean isJson = false;
        boolean isTextOutput = false;
        if (isListUserInfo || isShowSliceInfo || isCreateSliceOnly || isStatus) {
            boolean knownFormat = false;
            if (outputFormatArg.trim().equalsIgnoreCase("json")) {
                isJson = true;
                knownFormat = true;
            }
            if (outputFormatArg.trim().equalsIgnoreCase("text")) {
                isTextOutput = true;
                knownFormat = true;
            }
            if (!knownFormat) {
                err.println("Output format \"" + outputFormatArg + "\" is unknown");
                commandHelp(err, options, command);
                return 3;
            }
        }

        ExtraInfoCallback.setBonfireLoginCallback(new ExtraInfoCallback.BonfireLoginCallback() {
            private String cachedGroup = null;

            @Override
            public String getBonfireGroups() {
                LOG.info("Bonfire groups callback called.");
                if (cachedGroup == null) {
                    out.println("\n");
                    out.println("==============================================================");
                    cachedGroup = IOUtils.askCommandLineInput("Please enter \"Bonfire Group for experiment\":", in);
                    out.println("==============================================================\n");
                }
                return cachedGroup;
            }
        });
        ExtraInfoCallback.setAnyLoginCallback(new ExtraInfoCallback.AnyLoginCallback() {
            private JFedPasswordManager.LoginInfo cachedLogin = null;

            @Override
            public JFedPasswordManager.LoginInfo getLogin(String serviceName, String friendlyName) {
                LOG.info("Login callback called.");
                if (cachedLogin == null) {
                    out.println("\n");
                    out.println("==============================================================");
                    String username = IOUtils.askCommandLineInput("Please enter "+friendlyName+" username:", in);
                    String password = new String(IOUtils.askCommandLinePassword("Please enter "+friendlyName+" password:", in));
                    cachedLogin = new JFedPasswordManager.LoginInfo(serviceName, username, password);
                    out.println("==============================================================\n");
                }
                return cachedLogin;
            }
        });

        if (debug)
            for (Option o : line.getOptions())
                out.println("ARG: " + o.getOpt() + " " + o.getLongOpt() + " " + o.getValue() + " " + o.getValuesList() + " " + o.getValuesList().size());


        boolean sshKeysUserCert;
        boolean sshKeysUserServerKeys;
        boolean sshKeysSharedUserKeys;
        boolean sshKeysRspec;
        if (sshKeysArg != null) {
            String[] opts = sshKeysArg.split(",");

            sshKeysUserCert = false;
            sshKeysUserServerKeys = false;
            sshKeysSharedUserKeys = false;
            sshKeysRspec = false;

            for (String opt : opts) {
                boolean validOpt = false;

                if (opt.equalsIgnoreCase("usercert")) {
                    sshKeysUserCert = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("userkeys")) {
                    sshKeysUserServerKeys = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("shareduserallkeys")) {
                    sshKeysSharedUserKeys = true;
                    validOpt = true;
                }
                if (opt.equalsIgnoreCase("rspec")) {
                    sshKeysRspec = true;
                    validOpt = true;
                }

                if (!validOpt) {
                    err.println("The ssh-keys option you have specified (" + opt + ") is not valid");
                    commandHelp(err, options, command);
                    return 3;
                }
            }
        } else {
            //default: usercert,rspec
            sshKeysUserCert = true;
            sshKeysUserServerKeys = false;
            sshKeysSharedUserKeys = true;
            sshKeysRspec = true;
        }

        if (requiresSlice && sliceIdArg.startsWith("urn:")) {
            GeniUrn sliceUrn = GeniUrn.parse(sliceIdArg);
            if (sliceUrn == null) {
                err.println("Error parsing provided slice URN: " + sliceIdArg);
                return 2;
            }
            String sliceUrnprojectName = sliceUrn.getEncodedSubAuthName();
            if (((projectNameArg != null) != (sliceUrnprojectName != null)) ||
                    (projectNameArg != null && sliceUrnprojectName != null &&
                            !projectNameArg.equals(sliceUrnprojectName))) {
                if (projectNameArg != null) {
                    if (projectNameArg.equals(CHOOSE_PROJECT_AUTOMATICALLY_OPTION)) {
                        projectNameArg = sliceUrnprojectName;
                    } else {
                        err.println("The project-name you have specified (" + projectNameArg + ") does not match the sub authority in the slice URN you have specified (" + sliceIdArg + ")");
                        return 3;
                    }
                } else
                    projectNameArg = sliceUrnprojectName;
            }
        }
        final String sliceName = !requiresSlice ? null :
                (sliceIdArg.startsWith("urn:") ? GeniUrn.parse(sliceIdArg).getEncodedResourceName() : sliceIdArg);

        if (manifestArg == null)
            manifestArg = "manifest-" + sliceName + ".rspec";

        if (sliceExpirationDateArg != null && sliceExpirationHoursArg != null) {
            err.println("You may not specify both expiration-date and expiration-hours. You need to specify only one.");
            return 3;
        }

        final Date requestedExpirationDate;
        if (sliceExpirationDateArg == null) {
            if (sliceExpirationHoursArg == null)
                requestedExpirationDate = new Date(System.currentTimeMillis() + (2 * 60 * 60 * 1000l)); //2 hours
            else {
                try {
                    int hours = Integer.parseInt(sliceExpirationHoursArg);
                    requestedExpirationDate = new Date(System.currentTimeMillis() + (hours * 60 * 60 * 1000l));
                } catch (NumberFormatException e) {
                    err.println("Specified slice expiration hours (\"" + sliceExpirationDateArg + "\") is not a valid number.");
                    return 3;
                }
            }
        } else {
            try {
                requestedExpirationDate = RFC3339Util.rfc3339StringToDate(sliceExpirationDateArg);
            } catch (java.text.ParseException e) {
                err.println("Specified slice expiration date (\"" + sliceExpirationDateArg + "\") is not in RFC3339 format. The current date in RFC3339 format is: " +
                        RFC3339Util.dateToRFC3339String(new Date()));
                return 3;
            }
        }


        baseCli.initAuthoritiesModule(line);
        try {
            baseCli.initUserModule(line, out, err, in);
        } catch (BaseCli.CliArgumentException e) {
            LOG.debug("Got CliArgumentException", e);
            commandHelp(err, options, command);
            return 3;
        }
        final Injector injector = baseCli.getInjector(err);
        final AuthorityListModel authorityListModel = injector.getInstance(AuthorityListModel.class);
        final AuthorityFinder authorityFinder = injector.getInstance(AuthorityFinder.class);
        final GeniUser user = injector.getInstance(GeniUser.class);
        final JFedPreferences jFedPreferences = injector.getInstance(JFedPreferences.class);
        final Logger logger = injector.getInstance(Logger.class);

        GeniUrn bindRspecUrn = null;
        SfaAuthority bindRspecAuth = null;
        if (bindRspecArg != null) {
            bindRspecUrn = GeniUrn.parse(bindRspecArg);
            if (bindRspecUrn == null) {
                err.println("Error parsing provided component_manager_id URN: " + bindRspecArg);
                return 2;
            }
            bindRspecAuth = authorityFinder.findByUrn(bindRspecUrn, AuthorityFinder.Purpose.REQUEST_RSPEC);
            if (bindRspecAuth == null) {
                err.println("Error: could not find authority matching component_manager_id URN provided by --bind-rspec: " + bindRspecArg);
                return 2;
            }
        }


        final CallLogOutput callLogOutput = new CallLogOutput(err);
        if (callLogArg != null) {
            callLogOutput.fos = new FileOutputStream(callLogArg);

            try {
                XMLOutputFactory xof = XMLOutputFactory.newInstance();
                callLogOutput.xtw = new IndentingXMLEventWriter(xof.createXMLEventWriter(callLogOutput.fos, "UTF-8"));
                callLogOutput.eventFactory = XMLEventFactory.newInstance();

                callLogOutput.xtw.add(callLogOutput.eventFactory.createStartDocument());
                callLogOutput.xtw.add(callLogOutput.eventFactory.createStartElement(new QName("calls"), null, null));
            } catch (XMLStreamException e) {
                callLogOutput.onFailure();
                LOG.error("Exception generating XML for call output", e);
                err.println("Exception generating XML for call output: " + e.getMessage());
            }

            logger.addResultListener(result -> {
                if (callLogOutput.isDisabled())
                    return;

                SerializableApiCallDetails serializableApiCallDetails = new SerializableApiCallDetails(result);

                try {
                    if (result.getAuthority() != null)
                        callLogOutput.xtw.add(callLogOutput.eventFactory.createComment("" + result.getGeniMethodName() + " on " + result.getAuthority().getUrnString() + ""));

                    SerializableApiCallDetailsWriter.writeApiCallDetails(serializableApiCallDetails, callLogOutput.xtw, callLogOutput.eventFactory);

                    //add 2 newlines
                    callLogOutput.xtw.add(callLogOutput.eventFactory.createCharacters(System.getProperty("line.separator") + System.getProperty("line.separator")));

                    callLogOutput.xtw.flush();
                    callLogOutput.fos.flush();
                } catch (IOException e) {
                    err.println("Error flushing: " + e.getMessage());
                    e.printStackTrace();
                } catch (XMLStreamException e) {
                    err.println("Error writing call to stream: " + e.getMessage());
                    e.printStackTrace();

                    callLogOutput.onFailure();
                }
            });
        } else {
            callLogOutput.fos = null;
            callLogOutput.xtw = null;
            callLogOutput.eventFactory = null;
        }
        if (printCalls) {
            PrintStream printCallsOut = out;
            final String printCallsPrefix = "CALL: ";

            logger.addResultListener(new ResultListener() {
                @Override
                public void onResult(ApiCallDetails result) {
                    printCallsOut.println();
                    printCallsOut.println(TextUtil.indent(printCallsPrefix.length()+1, printCallsPrefix + "" + result.getGeniMethodName() +
                            " on " + result.getAuthority().getUrnString() + "" +
                            " at "+result.getConnectionConfig().getServerUrlString()));

                    String req = result.getXmlRpcRequestJsonString();
                    String res = result.getXmlRpcResponseJsonString();
                    if (req != null) printCallsOut.println(printCallsPrefix + "Request: " + req);
                    if (res != null) printCallsOut.println(printCallsPrefix + "Reply: " + res);

                    printCallsOut.println();
                    printCallsOut.println();
                    printCallsOut.flush();
                }
            });
        }
        boolean successExitStatus = true;
        try { //used for finally callLogOutput.close()
            if (amApiVersionArg != null && !amApiVersionArg.equals("2") && !amApiVersionArg.equals("3")) {
                err.println("Unsupported AM api version: \"" + amApiVersionArg + "\"");
                commandHelp(err, options, command);
                return 3;
            }

            String rspecString;
            StringRspec rspec = null;
            ModelRspec modelRspec = null;
            List<String> rspecComponentManagers = null;
            if (rspecArg != null) {
                assert rspecArg != null;
                if (rspecArg.startsWith("http")) {
                    URL url;
                    try {
                        url = new URL(rspecArg);
                    } catch (MalformedURLException e) {
                        LOG.debug("The URL specified in the --rspec argument, \"" + rspecArg + "\", is not valid", e);
                        err.println("The URL specified in the --rspec argument, \"" + rspecArg + "\", is not valid: " + e.getMessage());
                        return 3;
                    }
                    try {
                        rspecString = IOUtils.urlToString(url, Charset.defaultCharset());
                    } catch (IOException e) {
                        LOG.debug("Exception reading the URL specified in the --rspec argument, \"" + rspecArg + "\"", e);
                        err.println("There was an error reading the URL specified in the --rspec argument, \"" + rspecArg + "\": " + e.getMessage());
                        return 3;
                    }
                } else
                    try {
                        rspecString = IOUtils.fileToString(rspecArg);
                    } catch (IOException e) {
                        err.println("The RSpec file \"" + rspecArg + "\" was not found.");
                        return 3;
                    }
                rspec = new StringRspec(rspecString);

                if (rewriteRspecArg) {
                    try {
                        modelRspec = getModelRspec(modelRspec, rspec);

                        if (bindRspecUrn != null) {
                            int bindCount = 0;
                            for (RspecNode rspecNode : modelRspec.getNodes()) {
                                if (rspecNode.getComponentManagerId() == null) {
                                    rspecNode.setComponentManagerId(bindRspecUrn);
                                    bindCount++;
                                }
                            }
                            int nodeCount = modelRspec.getNodes().size();
                            if (bindCount == 0) {
                                if (!silent)
                                    err.println("WARNING: --bind-rspec was specified, but the RSpec did NOT contain any unbound nodes (0 nodes of "+nodeCount+" nodes have been bound).");
                            } else {
                                if (debug)
                                    out.println("--bind-rspec was specified: "+bindCount+" nodes of "+nodeCount+" nodes have been bound to "+bindRspecUrn);
                            }
                        }

                        RequestRspecSource rspecSource2 = new RequestRspecSource(modelRspec);
                        rspecString = rspecSource2.getRspecXmlString();
                        rspec = rspecSource2.getStringRspec();
                    } catch (AssertionError t) {
                        LOG.error("Error rewriting RSpec", t);
                        err.println("Error parsing and rewriting RSpec, will use original RSpec! Error message: " + t.getMessage());
                    } catch (Exception t) {
                        LOG.error("Error rewriting RSpec", t);
                        err.println("Error parsing and rewriting RSpec, will use original RSpec! Error message: " + t.getMessage());
                    }
                }


                rspecComponentManagers = rspec.getAllComponentManagerUrns();
                if (rspecComponentManagers.size() != 1 || debug) {
                    if (!silent)
                        out.println("Note: This RSpec has " + rspecComponentManagers.size() + " component managers: " + rspecComponentManagers);
                }
                if (rspecComponentManagers.size() == 0) {
                    if (isCreate) {
                        err.println("ERROR: The provided RSpec has no component managers. Something went wrong during parsing, or this is an empty RSpec. Either way, 'createSliver' cannot continue.");
                        return 3;
                    }
                    else
                        err.println("WARNING: The provided RSpec has no component managers. Something went wrong during parsing, or this is an empty RSpec. Either way, few commands can now meaningfully continue. Will try anyway.");
                }
            } else {
                //no RSpec provided, get rspecComponentManagers from SLIVER_INFO (or equivalent)
                //this is done below, after user info is retrieved
            }

            if (stitchingArg && rspec != null) {
                if (!rspec.isStitching(authorityListModel, authorityFinder)) {
                    err.println("The --stitching option was set, but the rspec does not specify any stitched links: Aborted CreateSliver.");
                    return 3;
                }
            }

            if (!BaseCli.checkUser(user, line, out, err, in))
                return 3;

            final SfaAuthority userAuth = user.getUserAuthority();
            final GeniUserProvider userProvider = injector.getInstance(GeniUserProvider.class);

            if (!silent) {
            }
            if (debug) {
                out.println("User " + user.getUserUrnString() + " will be used.");
                out.println("User Authority: " + user.getUserAuthority().getName());
                out.println("User Authority URN: " + user.getUserAuthority().getUrnString());
            }

            Hashtable sliceInfoHT = new Hashtable();

            boolean usesSpeaksFor = speaksForArgs != null && speaksForArgs.length > 0;
            sliceInfoHT.put("usedSpeaksfor", usesSpeaksFor);

            GeniUrn speakingForUser = null;
            List<AnyCredential> speaksForCredentials = new ArrayList<>();
            if (usesSpeaksFor) {
                LOG.debug("There are "+speaksForArgs.length+" speaksfor args: "+Arrays.asList(speaksForArgs));
                for (String speaksForArg : speaksForArgs) {
                    List<String> speaksForCredentialStrings;
                    try {
                        String speaksForMultiCredentialString = IOUtils.fileToString(speaksForArg);

                        speaksForCredentialStrings = extractCredentials(speaksForMultiCredentialString);
                    } catch (IOException e) {
                        LOG.error("Error reading speaksFor credential from file \"" + speaksForArg + "\"", e);
                        err.println("Error reading speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
                        return 3;
                    }
                    LOG.debug("File \""+speaksForArg+"\" contains "+speaksForCredentialStrings.size()+" credentials.");
                    for (String speaksForCredentialString : speaksForCredentialStrings) {
                        AnyCredential speaksForCredential;
                        GeniUrn currentSpeakingForUser = null;
                        try {
                            speaksForCredential = AnyCredential.createAutoDetect("speaksFor", speaksForCredentialString);
                        } catch (CredentialException e) {
                            LOG.error("Error parsing speaksFor credential from file \"" + speaksForArg + "\"", e);
                            err.println("Error parsing speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
                            return 3;
                        }
                        if (speaksForCredential == null) {
                            LOG.error("Error parsing speaksFor credential from file \"" + speaksForArg + "\"");
                            err.println("Error parsing speaksFor credential from file \"" + speaksForArg + "\"");
                            return 3;
                        }
                        if (!speaksForCredential.isSpeaksFor()) {
                            LOG.warn("WARNING: parsed speaksFor credential from file \"" + speaksForArg + "\" is NOT a speaksFor credential");
                            err.println("WARNING parsed speaksFor credential from file \"" + speaksForArg + "\" is NOT a speaksFor credential. (Will continue anyway)");
                        }
                        try {
                            boolean supportedCredType = false;
                            if (speaksForCredential instanceof AbacCredential) {
                                supportedCredType = true;
                                AbacCredential c = (AbacCredential) speaksForCredential;
                                if (!c.isSpeaksFor()) {
                                    LOG.error("Provided ABAC credential is not a speaksfor credential");
                                    err.println("Provided ABAC credential is not a speaksfor credential");
                                    return 3;
                                }
                                currentSpeakingForUser = c.getSpokenForUrn();
                                if (c.getSpokenForUrn() != null) {
                                    if (speakingForUser != null) {
                                        LOG.debug("Speaksfor chain search: speakingForUser="+speakingForUser);
                                        LOG.debug("Speaksfor chain search: checking "+speaksForCredentials.size()+" credentials");
                                        //there are multiple abac credentials, so it's harder to find who is spoken for
                                        //assume all credentials are ABAC, and start complex processing of chain
                                        Map<String, GeniUrn> spokenForByKeyId = new HashMap<>();
                                        List<String> speakersKeyIds = new ArrayList<>();

                                        List<AnyCredential> allCurrentlyKnownSpeaksForCredentials = new ArrayList<>(speaksForCredentials);
                                        allCurrentlyKnownSpeaksForCredentials.add(speaksForCredential);
                                        for (AnyCredential anyc : allCurrentlyKnownSpeaksForCredentials) {
                                            AbacCredential abacc = (AbacCredential) anyc;
                                            assert !spokenForByKeyId.containsKey(abacc.getHeadKeyId());
                                            assert !spokenForByKeyId.containsValue(abacc.getSpokenForUrn());
                                            assert !speakersKeyIds.contains(abacc.getTailKeyId());
                                            spokenForByKeyId.put(abacc.getHeadKeyId(), abacc.getSpokenForUrn());
                                            speakersKeyIds.add(abacc.getTailKeyId());
                                        }
                                        LOG.debug("Speaksfor chain search: spokenForByKeyId="+spokenForByKeyId);
                                        LOG.debug("Speaksfor chain search: speakersKeyIds="+speakersKeyIds);
                                        for (String speakersKeyId : speakersKeyIds)
                                            spokenForByKeyId.remove(speakersKeyId);
                                        LOG.debug("Speaksfor chain search: spokenForByKeyId="+spokenForByKeyId);
                                        if (spokenForByKeyId.size() == 1) {
                                            speakingForUser = spokenForByKeyId.values().iterator().next();
                                        } else {
                                            if (spokenForByKeyId.size() == 0) {
                                                LOG.warn("Speaksfor chain search got 0 leftover speaking_for urns");
                                                speakingForUser = null;
                                            } else {
                                                //ignore that there are multiple
                                                LOG.warn("Speaksfor chain search got "+spokenForByKeyId.size()+" leftover speaking_for urns");
                                                speakingForUser = spokenForByKeyId.values().iterator().next();
                                            }
                                        }
                                        LOG.debug("Speaksfor chain search: speakingForUser="+speakingForUser);
                                    } else {
                                        speakingForUser = c.getSpokenForUrn();
                                    }
                                }
                            }
                            if (speaksForCredential instanceof SfaCredential) {
                                if (speaksForArg.length() > 1) {
                                    LOG.error("Multiple speaksfor credentials are not supported for type SfaCredential \"" + speaksForArg + "\"");
                                    err.println("Multiple speaksfor credentials are not supported for type SfaCredential \"" + speaksForArg + "\"");
                                    return 3;
                                }
                                supportedCredType = true;
                                SfaCredential c = (SfaCredential) speaksForCredential;
                                if (c.getTargetUrn() != null)
                                    currentSpeakingForUser = new GeniUrn(c.getTargetUrn());
                                else {
                                    if (c.getTargetGid() != null) {
                                        Gid gid = new Gid(c.getTargetGid());
                                        if (gid.getSubjectUrn() != null) {
                                            currentSpeakingForUser = new GeniUrn(gid.getSubjectUrn());
                                        }
                                    }
                                }
                                speakingForUser = currentSpeakingForUser;
                            }
                            if (!supportedCredType) {
                                LOG.error("Unsupported speaksfor credential type for \"" + speaksForArg + "\": "+speaksForCredential.getType());
                                err.println("Unsupported speaksfor credential type for \"" + speaksForArg + "\": "+speaksForCredential.getType());
                                return 3;
                            }
                        } catch (GeniUrn.GeniUrnParseException e) {
                            LOG.error("Error parsing spoken for user URN from speaksFor credential from file \"" + speaksForArg + "\"", e);
                            err.println("Error parsing spoken for user URN from speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
                            return 3;
                        }
//                if (!speaksForCredential.validate()) {
//                    LOG.warn("WARNING: Provided speaksfor credential does not validate");
//                    err.println("WARNING: Provided speaksfor credential does not validate");
//                }

                        boolean check;
                        try {
                            check = speaksForCredential.check(authorityListModel);
                        } catch (CredentialException e) {
                            LOG.warn("Error trying to check credential signature: ", e);
                            check = false;
                        }
                        if (!check) {
                            LOG.warn("WARNING: Provided speaksfor credential signature check failed. (frequently caused by: you modified the credential)");
                            err.println("WARNING: Provided speaksfor credential signature check failed. (frequently caused by: you modified the credential)");
                        }

                        if (currentSpeakingForUser == null) {
                            LOG.warn("ERROR: could not determine the spoken for user in the speaksFor credential from file \"" + speaksForArg + "\". Cannot continue.");
                            err.println("ERROR: could not determine the spoken for user in the speaksFor credential from file \"" + speaksForArg + "\". Cannot continue.");
                            return 3;
                        }
                        if (speakingForUser == null) {
                            LOG.warn("ERROR: could not determine the final spoken for user. Cannot continue.");
                            err.println("ERROR: could not determine the final spoken for user. Cannot continue.");
                            return 3;
                        }
                        if (!silent)
                            out.println("Using speaksFor credential for user \"" + currentSpeakingForUser + "\"");
                        speaksForCredentials.add(speaksForCredential);
                    }
                }

                if (!silent && speaksForArgs.length > 1)
                    out.println("You are using multiple speaksfor credentials. The final user being spoken for is \"" + speakingForUser + "\"");
            }

            JFedConnectionProvider conPool = injector.getInstance(JFedConnectionProvider.class);
            AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory
                    automaticUserAndSliceApiWrapperFactory =
                    injector.getInstance(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class);
            AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
                    automaticAggregateManagerWrapperFactory =
                    injector.getInstance(AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class);

            final UserAndSliceApiWrapper sa = automaticUserAndSliceApiWrapperFactory.create();
            GeniUrn targetUserUrn = user.getUserUrn();
            sliceInfoHT.put("userUrn", targetUserUrn);

            if (usesSpeaksFor) {
                sa.setSpeaksFor(speaksForCredentials, speakingForUser);
                targetUserUrn = speakingForUser;
                if (speaksForCredentials.size() == 1)
                    sliceInfoHT.put("speaksForCredential", speaksForCredentials.get(0));
                else {
                    int i = 1;
                    for (AnyCredential speaksForCredential : speaksForCredentials)
                        sliceInfoHT.put("speaksForCredential" + (i++), speaksForCredential);
                }
                sliceInfoHT.put("speakingForUserUrn", speakingForUser);
            }
            String targetUserUrnString = targetUserUrn.toString();

            Set<GeniUrn> shareUserUrns;
            boolean shareSliceWithProjectUsers = false;
            if (shareSliceUsersArg != null) {
                String tla = targetUserUrn.getEncodedTopLevelAuthority();
                shareUserUrns = new HashSet<>();
                String[] users = shareSliceUsersArg.split(",");
                for (String opt : users) {
                    boolean validOpt = false;

                    if (opt.equals("PROJECTUSERS")) {
                        if (!silent)
                            out.println("Will share slice with all users in the slice project.");
                        shareSliceWithProjectUsers = true;
                        validOpt = true;
                    }
                    if (!validOpt && GeniUrn.valid(opt)) {
                        GeniUrn userUrn = GeniUrn.parse(opt);
                        assert opt != null;
                        shareUserUrns.add(userUrn);
                        validOpt = true;
                    }
                    if (!validOpt && opt.matches("[a-zA-Z][a-zA-Z0-9]*") && opt.length() <= 8) {
                        GeniUrn userUrn = GeniUrn.createGeniUrnFromEncodedParts(tla, "user", opt);
                        if (!silent)
                            out.println("Assuming specified user \""+opt+"\" is: "+userUrn);
                        shareUserUrns.add(userUrn);
                        validOpt = true;
                    }

                    if (!validOpt) {
                        err.println("The share-slice option you have specified (" + opt + ") is not valid a valid URN or username");
                        commandHelp(err, options, command);
                        return 3;
                    }
                }
            } else {
                shareUserUrns = Collections.emptySet();
            }
            assert shareUserUrns != null;

            List<AnyCredential> userCreds;
            try {
                userCreds = sa.getLoggedinUserCredentials(logger);
                sliceInfoHT.put("userCredentials", new Vector(userCreds));
            } catch (JFedException e) {
                return callFailed("Error retrieving user credential", e, err, callLogOutput);
            }
            if (debug)
                out.println(userCreds.size() + " user credentials retrieved.");
            List<GeniUrn> existingSlices;
            List<String> existingSlicesCompareStrings;
            try {
                existingSlices = sa.getSlicesForUser(logger, targetUserUrn);
            } catch (JFedException e) {
                return callFailed("Error retrieving slices for user", e, err, callLogOutput);
            }
            if (debug)
                out.println("User " + targetUserUrnString + " has the following slices: " + existingSlices);
            if (existingSlices != null) {
                existingSlicesCompareStrings = new ArrayList<>();
                for (GeniUrn sliceUrn : existingSlices) {
                    GeniUrn lowercaseSliceUrn = GeniUrn.createGeniUrnFromEncodedParts(
                            sliceUrn.getEncodedTopLevelAuthority().toLowerCase(),
                            sliceUrn.getEncodedResourceType(),
                            sliceUrn.getEncodedResourceName().toLowerCase()
                    );
                    existingSlicesCompareStrings.add(lowercaseSliceUrn.toString());
                }
            } else
                existingSlicesCompareStrings = null;

            if (isListUserInfo) {
                List<String> projectNames;
                try {
                    projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);
                } catch (JFedException e) {
                    return callFailed("Error retrieving projects for user", e, err, callLogOutput);
                }
                if (isJson) {
                    Hashtable outerUserHt = new Hashtable();
                    Hashtable userHt = new Hashtable();
                    userHt.put("urn", targetUserUrnString);
                    userHt.put("slices", new Vector(existingSlices));
                    userHt.put("projects", new Vector(projectNames));
                    outerUserHt.put("User Info", userHt);
//                    out.println(JSonHelper.jsonObjectToString(outerUserHt));
                    out.println(JSonHelper.jsonObjectToString(userHt));
                }
                if (isTextOutput) {
                    out.println("User Info: ");
                    out.println("   Urn: "+targetUserUrnString);
                    out.println("   Projects: "+projectNames.stream().collect(Collectors.joining(",")).toString());
                    out.println("   Slices: "+existingSlices.stream().map(u -> u.toString()).collect(Collectors.joining(",")).toString());
                }

                return 0;
            }

            List<String> projectNames = null;
            if (projectNameArg != null && projectNameArg.equals(CHOOSE_PROJECT_AUTOMATICALLY_OPTION)) {
                try {
                    projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);

                    if (debug)
                        out.println("You are a member of the following projects: " + projectNames);

                    if (projectNames.isEmpty()) {
                        err.println("WARNING: " + CHOOSE_PROJECT_AUTOMATICALLY_OPTION + " was set as project, but you do not seem to be a member of any project! Will use fallback: continue without project.");

                        //fallback: no project
                        projectNameArg = null;
                    } else {
                        //select last project
                        projectNameArg = projectNames.get(projectNames.size() - 1);
                        if (debug || !silent)
                            out.println("Automatically selected project \"" + projectNameArg + "\"");
                    }
                } catch (JFedException e) {
                    //fallback: no project
                    projectNameArg = null;

                    String message =
                            "WARNING: " + CHOOSE_PROJECT_AUTOMATICALLY_OPTION + " was set as project, but there was an error retrieving list of user projects. Will use fallback: continue without project.";
                    if (e.getXmlRpcResult() != null)
                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                    else {
                        err.println(message + ":");
                        e.printStackTrace(err);
                    }
                }
            }

            GeniUrn initialSliceUrn = sliceIdArg.startsWith("urn:") ?
                    GeniUrn.parse(sliceIdArg) :
                    GeniUrn.createGeniUrnFromEncodedParts(
                            userAuth.getNameForUrn() + (projectNameArg == null ? "" : ":" + projectNameArg),
                            "slice",
                            sliceIdArg
                    );

            GeniUrn lowercaseInitialSliceUrn = GeniUrn.createGeniUrnFromEncodedParts(
                    initialSliceUrn.getEncodedTopLevelAuthority().toLowerCase(),
                    initialSliceUrn.getEncodedResourceType(),
                    initialSliceUrn.getEncodedResourceName().toLowerCase()
            );
//            final boolean sliceExists = existingSlices.contains(initialSliceUrn);
            final boolean sliceExists = existingSlicesCompareStrings.contains(lowercaseInitialSliceUrn.toString());

            if (!silent && !(isShowSliceInfo && sliceExists)&& !(isCreateSliceOnly && !sliceExists))
                out.println("slice " + initialSliceUrn + " " + (sliceExists ? "already exists" : "does not yet exist"));

            //check if project exists
            if ((isCreate || isCreateSliceOnly) && !sliceExists && initialSliceUrn.getEncodedSubAuthName() != null) {
                try {
                    if (projectNames == null)
                        projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);
                    assert projectNames != null;

                    List<String> projectNamesLowerCase = new ArrayList<>(projectNames.size()+1);
                    for (String projectName : projectNames)
                        projectNamesLowerCase.add(projectName.toLowerCase());

                    if (!projectNamesLowerCase.contains(initialSliceUrn.getEncodedSubAuthName().toLowerCase()))
                        err.println("WARNING: You do not seem to be a member if the project (\"" + initialSliceUrn.getEncodedSubAuthName() + "\") " +
                                "in which you wish to create a slice. Projects you are member of: " + projectNames);
                    else {
                        if (!projectNames.contains(initialSliceUrn.getEncodedSubAuthName())) {
                            String userProvidedProjectName = initialSliceUrn.getEncodedSubAuthName();
                            for (String correctCaseProjectName : projectNames) {
                                if (userProvidedProjectName.equalsIgnoreCase(correctCaseProjectName) && !userProvidedProjectName.equals(correctCaseProjectName)) {
                                    GeniUrn orig = initialSliceUrn;
                                    initialSliceUrn = GeniUrn.createGeniUrnFromEncodedParts(
                                            initialSliceUrn.getEncodedTopLevelAuthority_withoutSubAuth()+":"+correctCaseProjectName,
                                            initialSliceUrn.getEncodedResourceType(),
                                            initialSliceUrn.getEncodedResourceName());
                                    LOG.info("(CaseBugs): You are a member of \"" + userProvidedProjectName + "\", " +
                                            "but that project is known with a different case: \"" + correctCaseProjectName + "\"." +
                                            "jFed will change the case to avoid any SA/AM bugs. (though it shouldn't matter.). " +
                                            "Changed slice URN from " + orig + " to " +initialSliceUrn);
                                }
                            }
                        }

                    }

                    if (debug)
                        out.println("You are a member of the following projects: " + projectNames);
                } catch (JFedException e) {
                    String message = "Error retrieving list of user projects. (this error will be ignored)";
                    if (e.getXmlRpcResult() != null)
                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                    else {
                        err.println(message + ":");
                        e.printStackTrace(err);
                    }
                }
            }

            final GeniUrn sliceUrn = initialSliceUrn;

            if (!(isCreate || isCreateSliceOnly) && !sliceExists) {
                err.println("Error: The slice " + sliceUrn + " does not exist. Cannot continue.");
                return 2;
            }

            if (fake) {
                out.println("Argument processing done. --fake is set, so no further steps are taken.");
                return 0;
            }

            List<AnyCredential> sliceCredentials;
            Date sliceExpirationDate = null;
            if (sliceExists) {
                sliceInfoHT.put("sliceUrn", sliceUrn);
                try {
                    sliceCredentials = sa.getSliceCredentials(logger, sliceUrn);
                    if (!sliceCredentials.isEmpty()) sliceExpirationDate = sliceCredentials.get(0).getExpiresDate();
                    LOG.debug("ExperimenterCli requested slice credentials, and got "+sliceCredentials.size()+" credentials");
                } catch (JFedException e) {
                    return callFailed("Error retrieving slice credential", e, err, callLogOutput);
                }

                if (rspecArg == null && !(isShowSliceInfo || isCreateSliceOnly)) {
                    //no RSpec provided, so get rspecComponentManagers from SLIVER_INFO (or equivalent)
                    try {
                        rspecComponentManagers = sa.getAggregatesForSlice(logger, sliceCredentials, sliceUrn);
                    } catch (JFedException e) {
                        return callFailed("Error retrieving slice credential", e, err, callLogOutput);
                    }
                    if (rspecComponentManagers.size() == 0) {
                        err.println("WARNING: No RSpec was provided using the '--rspec' option, and the SA did not know of any AM's that have resources. " +
                                "Few commands can now do anything. Will try anyway.");
                    }
                }

                if (isRenew || isCreate) {
                    Date currentSliceExpireDate = null;
                    for (AnyCredential credential : sliceCredentials)
                        if (credential.getExpiresDate() != null)
                            currentSliceExpireDate = credential.getExpiresDate();
                    if (currentSliceExpireDate != null && isBefore(currentSliceExpireDate, requestedExpirationDate)) {
                        try {
                            if (debug)
                                out.println("Need to Renew slice, because the slice expires at " + RFC3339Util.dateToRFC3339String(currentSliceExpireDate, true, true, true) +
                                        " while requested expiration date is " + RFC3339Util.dateToRFC3339String(requestedExpirationDate, true, true, true));
                            if (!silent)
                                out.println("Renewing slice, requested new expire date is "+RFC3339Util.dateToRFC3339String(requestedExpirationDate, true, true, true)+" ...");
                            List<AnyCredential> newSliceCredentials =
                                    sa.renewSlice(logger, sliceCredentials, requestedExpirationDate);
                            sliceExpirationDate = requestedExpirationDate; //fallback
                            if (newSliceCredentials != null) {
                                sliceCredentials = newSliceCredentials;
                                if (debug)
                                    out.println("ExperimenterCli renewed the slice and got "+sliceCredentials.size()+" credentials");
                                if (!sliceCredentials.isEmpty()) {
                                    sliceExpirationDate = sliceCredentials.get(0).getExpiresDate();
                                    if (sliceExpirationDate != null && isBefore(sliceExpirationDate, requestedExpirationDate)) {
                                        if (!silent) {
                                            err.println("WARNING: The slice was renewed, but the requested expiration date was not set. " +
                                                    "Actual new expire date: "+RFC3339Util.dateToRFC3339String(sliceExpirationDate, true, true, true));
                                        }
                                    }
                                }
                            }
                            else
                                out.println("WARNING: did not receive renewed credential");
                        } catch (JFedException e) {
                            return callFailed("Error renewing slice credential", e, err, callLogOutput);
                        }
                    } else {
                        if (currentSliceExpireDate != null) {
                            if (isRenew)
                                if (debug)
                                    out.println("No need to renew slice. It already expires at "+RFC3339Util.dateToRFC3339String(currentSliceExpireDate, true, true, true));
                        } else {
                            if (!silent) {
                                if (isRenew)
                                    err.println("WARNING: expire date of slice is not known. Will not attempt renew!");
                                else
                                    err.println("WARNING: actual expire date of created slice is not known.");
                            }
                        }
                    }
                }

                sliceInfoHT.put("sliceCredentials", new Vector(sliceCredentials));

                if (isShowSliceInfo || isCreateSliceOnly) {
                    try {
                        List<GeniUrn> sliceUsers = sa.getUsersForSlice(logger, sliceUrn, sliceCredentials);
                        sliceInfoHT.put("sliceUsers", new Vector(sliceUsers));
                    } catch (JFedException e) {
                        LOG.error("Error getting users for slice", e);
                        err.println("Error getting users for slice, will ignore.");
                    }

                    try {
                        List<String> sliceAggs = sa.getAggregatesForSlice(logger, sliceCredentials, sliceUrn);
                        sliceInfoHT.put("sliceAggregates", new Vector(sliceAggs));
                    } catch (JFedException e) {
                        LOG.error("Error getting aggregates for slice", e);
                        err.println("Error getting aggregates for slice, will ignore.");
                    }

                    try {
                        Date sliceFoundExpirationDate = sa.getSliceExpiration(logger, sliceCredentials, sliceUrn);
                        if (sliceFoundExpirationDate != null)
                            sliceInfoHT.put("sliceExpiration", RFC3339Util.dateToRFC3339String(sliceFoundExpirationDate));
                    } catch (JFedException e) {
                        LOG.error("Error getting expiration date for slice", e);
                        err.println("Error getting expiration date, will ignore.");
                    }
                }
            } else {
                assert isCreate || isCreateSliceOnly;
                if (!createSliceArg && isCreate) {
                    err.println("Error: The slice " + sliceUrn + " does not exist, and you have not specified the --create-slice option.");
                    return 2;
                }
                UserAndSliceApiWrapper.SliceInfo sliceInfo;
                try {
                    sliceInfo =
                            sa.createSlice(logger, sliceUrn.getEncodedResourceName(), requestedExpirationDate, projectNameArg);
                    if (!sliceInfo.getCredentials().isEmpty()) {
                        sliceExpirationDate = sliceInfo.getCredentials().get(0).getExpiresDate();
                    }
                } catch (JFedException e) {
                    return callFailed("Error creating slice", e, err, callLogOutput);
                }
                sliceCredentials = sliceInfo.getCredentials();
                LOG.debug("ExperimenterCli created a slice, and got "+sliceCredentials.size()+" credentials");

                sliceInfoHT.put("sliceCredentials", new Vector(sliceCredentials));
                sliceInfoHT.put("sliceUrn", sliceInfo.getUrn());
                sliceInfoHT.put("sliceExpiration", requestedExpirationDate);
            }
            if (sliceCredentials == null || sliceCredentials.isEmpty()) {
                err.println("Error: Did not receive slice credentials. (but call was successful)");
                return 2;
            }
            LOG.debug("ExperimenterCLI has "+sliceCredentials.size()+" slice credentials");



            if (shareSliceWithProjectUsers) {
                if (projectNameArg == null) {
                    if (!silent) {
                        out.println("NOTE: Slice will NOT be shared with all members of project, because no project is specified.");
                    }
                } else {
                    try {
                        List<GeniUrn> projectUsers = sa.getUsersForSubAuthority(logger, projectNameArg);
                        shareUserUrns.addAll(projectUsers);
                        if (!silent) {
                            out.println("Slice will be shared with all members of project \"" + projectNameArg + "\": " + projectUsers);
                        }
                    } catch (UnsupportedOperationException e) {
                        err.println("Cannot get project members of \"" + projectNameArg + "\" to share slice with. -> Will not share slice with project members. " +
                                "Reason: API does not support retrieving the project members.");
                    } catch (JFedException e) {
                        LOG.error("Failed to get project members of \"" + projectNameArg + "\" to share slice with. -> Will not share slice with project members.", e);
                        err.println("Failed to get project members of \"" + projectNameArg + "\" to share slice with. Will not share slice with project members. Exception: " + e.getMessage());
                        e.printStackTrace(err);
                    }
                }
            }
            if (debug) {
                out.println("Slice will be shared with: "+shareUserUrns);
            }

            if (!shareUserUrns.isEmpty()) {
                if (!silent)
                    out.println("Will share existing slice with: "+shareUserUrns);
                for (GeniUrn userUrn : shareUserUrns) {
                    try {
                        sa.shareSlice(logger, sliceUrn, sliceCredentials, userUrn);
                    } catch (JFedException e) {
                        LOG.error("Failed to share slice with \""+userUrn+"\"", e);
                        err.println("Failed to share slice with \"" + userUrn + "\": " + e.getMessage());
                        e.printStackTrace(err);
                    }
                }
            }


            GeniUrn newSliceUrn = null;
            //see FEDIBBTDEV-2071: change slice URN to match slice URN in credential
            for (AnyCredential cred : sliceCredentials) {
                if (cred instanceof SfaCredential) {
                    SfaCredential c = (SfaCredential) cred;
                    if (c.getTargetUrn() != null && c.getTargetUrn().equalsIgnoreCase(sliceUrn.toString()) && !c.getTargetUrn().equals(sliceUrn.toString())) {
                        String message = "(CaseBugs): slice URN ("+sliceUrn+") does not match URN in credential ("+c.getTargetUrn()+"). " +
                                "Will automatically fix this by changing slice URN. " +
                                "(Otherwise, some AM's might give strange errors)";
//                        if (!silent)
                        if (debug)
                            err.println("WARNING: "+message);
                        LOG.warn(message);

                        newSliceUrn = GeniUrn.parse(c.getTargetUrn());
                        if (newSliceUrn != null) {
                            break;
                        }
                        else
                            err.println("ERROR: slice URN in credential is not a valid URN. This should not happen.");
                    }
                }
                if (cred instanceof AbacCredential) {
                    AbacCredential c = (AbacCredential) cred;
                    //cannot do this for ABAC
                }
            }
            final GeniUrn usedSliceUrn = newSliceUrn == null ? sliceUrn : newSliceUrn;
            sliceInfoHT.put("sliceUrn", usedSliceUrn);

            if (isCreateSliceOnly || isShowSliceInfo) {
                if (isCreateSliceOnly && sliceExists) {
                    err.println("Error: Slice already exists.");
                }

                if (isJson) {
                    String jsonOutput = XmlRpcPrintUtil.xmlRpcObjectToString(sliceInfoHT);
                    out.println(jsonOutput);
                } else {
                    assert isTextOutput;
                    out.println("User URN: "+sliceInfoHT.get("userUrn"));
//                    out.println("User Credentials: "+sliceInfoHT.get("userCredentials"));
                    out.println("Slice URN: "+sliceInfoHT.get("sliceUrn"));
                    out.println("Slice URN: "+sliceInfoHT.get("sliceExpiration"));
                    Vector sliceCreds = (Vector) sliceInfoHT.get("sliceCredentials");
                    int i = 0;
                    if (sliceCreds != null)
                        for (Object c : sliceCreds) {
                            if (c != null && c instanceof AnyCredential) {
                                AnyCredential ac = (AnyCredential) c;
                                out.println("Slice Credential " + (++i) + ":\n"+ac.getCredentialXml()+"\n\nEND SLICE CREDENTIAL\n");
                            }
                        }
                }

                if (isCreateSliceOnly && sliceExists)
                    return 1;
                return 0;
            }

            if (isCreate && sliceRecoverInfoArg != null) {
                if (debug)
                    out.println("Writing slice recovery info file \"" + sliceRecoverInfoArg + "\"");

                try {
                    FileConfiguration conf = new PropertiesConfiguration();
                    conf.setEncoding("UTF-8");
                    conf.setProperty(KEY_EXPIRATION_TIME,
                            sliceExpirationDate == null ? "unknown"
                                    : ISO_OFFSET_DATE_TIME.format(ZonedDateTime.ofInstant(sliceExpirationDate.toInstant(), ZoneId.systemDefault())));
                    conf.setProperty(KEY_REQUEST_RSPEC, rspec.getXmlString());
                    conf.save(new File(sliceRecoverInfoArg));
                } catch (ConfigurationException e) {
                    LOG.error("Error writing slice recovery info to \"" + sliceRecoverInfoArg + "\"", e);
                    err.println("ERROR (non fatal): Exception writing slice recovery info to \"" + sliceRecoverInfoArg + "\": " + e.getMessage());
                }
            }


            if (fakeSliver) {
                out.println("Argument and slice processing done. --fake-sliver is set, so no further steps are taken.");
                return 0;
            }

            final List<AutoShareLanInfo> lanToShare = new ArrayList<>();
            if (isCreate) {
                try {
                    modelRspec = getModelRspec(modelRspec, rspec);

                    for(RspecLink link : modelRspec.getLinks()) {
                        if (link.hasAutoShareLan()) {
                            for (String shareName : link.getAutoShareLanNames()) {
                                String clientId = link.getClientId();
                                if (clientId == null) {
                                    LOG.warn("Cannot share link that has no client_id");
                                } else {
                                    if (link.getComponentManagerUrns().size() > 1) {
                                        LOG.warn("Cannot share link ("+clientId+") that has multiple component_managers");
                                    } else {
                                        if (link.getComponentManagerUrns().isEmpty()) {
                                            LOG.warn("Cannot share link ("+clientId+") that has no component_managers");
                                        } else {
                                            GeniUrn cm = link.getComponentManagerUrns().get(0);
                                            AutoShareLanInfo asli = new AutoShareLanInfo(cm, clientId, shareName);
                                            lanToShare.add(asli);
                                        }
                                    }
                                }
                            }
                        }
                    }
                } catch (RspecParseException e) {
                    LOG.error("Problem parsing RSpec to automatically share LAN. Will not automatically share LAN.", e);
                    lanToShare.clear();
                }
            }

            final List<AnsiblePlaybookInfo> ansiblePlaybooksToExecute = new ArrayList<>();
            if ((isCreate || isManifest) && !noPlaybookToExecute) {
                if (ansibleExtraPlaybooks != null) {
                    LOG.debug("Processing "+ansibleExtraPlaybooks.length+" command line extra playbooks");
                    for (String ansibleExtraPlaybook : ansibleExtraPlaybooks) {
                        String[] parts = ansibleExtraPlaybook.split(" ", 2);
                        String source = parts[0];
                        String outputFilename = parts.length > 1 ? parts[1] : null;
                        ExecuteAnsiblePlaybook executeAnsiblePlaybook = new BasicExecuteAnsiblePlaybook(source, outputFilename);
                        AnsiblePlaybookInfo info = new AnsiblePlaybookInfo(executeAnsiblePlaybook);
                        try {
                            info.loadContent(ansibleAllowPlaybookInputfile);
                            info.checkFilename(false /*mustBeBasename*/, ansibleAllowAnyPlaybookOutputFile);
                        } catch (AnsiblePlaybookInfo.AnsiblePlaybookException e) {
                            LOG.debug("Error processing ExecuteAnsiblePlaybook", e);
                            err.println("Error processing --ansible-add-playbook: " + e.getMessage());
                            return 2;
                        }
                        ansiblePlaybooksToExecute.add(info);
                    }
                }

                if (rspec != null && ansibleExecuteRspecPlaybooks) {
                    try {
                        modelRspec = getModelRspec(modelRspec, rspec);

                        for (ExecuteAnsiblePlaybook executeAnsiblePlaybook : modelRspec.getExecuteAnsiblePlaybooks()) {
                            AnsiblePlaybookInfo info = new AnsiblePlaybookInfo(executeAnsiblePlaybook);
                            try {
                                info.loadContent(ansibleAllowPlaybookInputfile);
                                info.checkFilename(false /*mustBeBasename*/, ansibleAllowAnyPlaybookOutputFile);
                            } catch (AnsiblePlaybookInfo.AnsiblePlaybookException e) {
                                LOG.debug("Error processing ExecuteAnsiblePlaybook", e);
                                err.println("Error processing " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + ": " + e.getMessage());
                                return 2;
                            }
                            ansiblePlaybooksToExecute.add(info);
                        }
                    } catch (RspecParseException e) {
                        LOG.error("Problem parsing RSpec to automatically execute ansible playbooks. Will not automatically execute ansible playbooks.", e);
                        ansiblePlaybooksToExecute.clear();
                    }
                }

                Set<String> sources = new HashSet<>();
                for (Iterator<AnsiblePlaybookInfo> it = ansiblePlaybooksToExecute.iterator(); it.hasNext();) {
                    AnsiblePlaybookInfo i = it.next();
                    if (!sources.add(i.getExecuteAnsiblePlaybook().getSource())) {
                        err.println("Warning: duplicate playbook will be ignored: " + i.getExecuteAnsiblePlaybook());
                        it.remove();
                    }
                }

                if (debug) {
                    out.println("There are "+ansiblePlaybooksToExecute.size()+" ansible playbooks to execute.");
                }
            }


            List<UserSpec> userSpecs = new ArrayList<>();
            if (isCreate) {
                if (sshKeysUserServerKeys) {
                    try {
                        List<String> userKeys = sa.getSshKeysForUser(logger, targetUserUrn);
                        addToUserSpecs(userSpecs, targetUserUrn, userKeys);
                        if (debug)
                            out.println("Added " + userKeys.size() + " keys from server.");
                    } catch (JFedException e) {
                        return callFailed("Error retrieving ssh keys for user " + targetUserUrn, e, err, callLogOutput);
                    }
                }
                if (!shareUserUrns.isEmpty() && sshKeysSharedUserKeys) {
                    for (GeniUrn shareUserUrn : shareUserUrns) {
                        try {
                            List<String> userKeys = sa.getSshKeysForUser(logger, shareUserUrn);
                            if (userKeys != null && !userKeys.isEmpty()) {
                                addToUserSpecs(userSpecs, shareUserUrn, userKeys);
                                if (debug)
                                    out.println("Added " + userKeys.size() + " keys from server for user \"" + shareUserUrn + "\".");
                            } else {
                                if (!silent)
                                    err.println("WARNING: No SSH keys found for user \"" + shareUserUrn + "\".");
                            }
                        } catch (JFedException e) {
                            LOG.error("Error retrieving ssh keys for user \"" + shareUserUrn+"\"", e);
                            err.println("Error retrieving ssh keys for user \"" + shareUserUrn+"\": "+e.getMessage());
                            e.printStackTrace(err);
                        }
                    }
                }
                if (sshKeysUserCert) {
                    if (!usesSpeaksFor) {
                        String userKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey());
                        addToUserSpecs(userSpecs, targetUserUrn, userKey);
                        if (debug)
                            out.println("Adding userspec from user certificate.");
                    } else {
                        PublicKey pubKey = null;
                        for (AnyCredential speaksForCredential : speaksForCredentials) {
                            assert speaksForCredential != null;
                            PublicKey curPubKey = null;
                            GeniUrn currentSpeakingForUser = null;
                            if (speaksForCredential instanceof AbacCredential) {
                                AbacCredential c = (AbacCredential) speaksForCredential;
                                curPubKey = c.getSpokenForPubKey();

                                if (c.getSpokenForUrn() != null)
                                    currentSpeakingForUser = c.getSpokenForUrn();
                            }
                            try {
                                if (speaksForCredential instanceof SfaCredential) {
                                    SfaCredential c = (SfaCredential) speaksForCredential;
                                    if (c.getTargetUrn() != null)
                                        currentSpeakingForUser = new GeniUrn(c.getTargetUrn());
                                    if (c.getTargetGid() != null) {
                                        Gid gid = new Gid(c.getTargetGid());
                                        if (gid.getCertificate() != null)
                                            curPubKey = gid.getCertificate().getPublicKey();
                                        if (currentSpeakingForUser == null && gid.getSubjectUrn() != null) {
                                            currentSpeakingForUser = new GeniUrn(gid.getSubjectUrn());
                                        }
                                    }
                                }
                            } catch (GeniUrn.GeniUrnParseException e) {
                                LOG.error("Ignoring speaksfor cert target URN parse exception", e);
                            }

                            if (curPubKey != null && currentSpeakingForUser != null && currentSpeakingForUser.equals(speakingForUser)) {
                                pubKey = curPubKey;
                            }
                        }
                        if (pubKey != null) {
                            String userKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(pubKey);
                            addToUserSpecs(userSpecs, targetUserUrn, userKey);
                            if (debug)
                                out.println("Adding userspec from speaksFor SfaCredential");
                        } else {
                            err.println("WARNING: Could not fetch user public key from user speaksFor certificate(s).");
                        }
                    }
                }
                if (sshKeysRspec) {
                    try {
                        modelRspec = getModelRspec(modelRspec, rspec);
                        List<UserSpec> extraUserSpec = modelRspec.getSshKeys();
                        if (debug)
                            out.println("Adding " + extraUserSpec.size() + " userspecs from RSpec");
                        for (UserSpec us : extraUserSpec) {
                            GeniUrn userUrn = GeniUrn.parse(us.getUrn());
                            addToUserSpecs(userSpecs, userUrn, us.getSshKey());
                        }
                    } catch (RspecParseException e) {
                        err.println("WARNING: Error parsing specified RSpec. The users could not be extracted from this RSpec. Note: This tool will continue, and the RSpec will still be used to create slivers.");
                    }
                }
                if (debug) {
                    out.println("The following user specification will be added to all create sliver calls:");
                    for (UserSpec us : userSpecs) {
                        out.println("  User " + us.getUrn() + " with " + us.getSshKey().size() + " keys");
                        for (String key : us.getSshKey())
                            out.println("     " + key);
                    }
                }
            }

            //create amWrappers
            List<AggregateManagerWrapper> amWrappers = new ArrayList<>();
            for (String rspecComponentManager : rspecComponentManagers) {
                GeniUrn rspecComponentManagerUrn = GeniUrn.parse(rspecComponentManager);

                SfaAuthority currentAuthority =
                        authorityFinder.findByUrn(rspecComponentManager, AuthorityFinder.Purpose.CREATE_SLIVER);
                if (currentAuthority == null) {
                    err.println("Tested authority URN is unknown: \"" + rspecComponentManager + "\"");
                    return 2;
                }
                assert currentAuthority != null;

                final AggregateManagerWrapper am;
                if (amApiVersionArg == null) {
                    am = automaticAggregateManagerWrapperFactory.create(currentAuthority);
                } else {
                    if (amApiVersionArg.equals("2"))
                        am = new AMv2Wrapper(logger, userProvider, conPool, jFedPreferences, currentAuthority);
                    else {
                        assert amApiVersionArg.equals("3");
                        am = new AMv3Wrapper(logger, userProvider, conPool, jFedPreferences, currentAuthority);
                    }
                }
                if (usesSpeaksFor) {
                    am.setSpeaksFor(speakingForUser);
                }
                amWrappers.add(am);
            }



            if (isCreate && abortIfSliversExist) {
                if (debug) {
                    out.println("--abort-if-slivers-exist option is set: will first check if slivers already exist");
                }
                for (AggregateManagerWrapper am : amWrappers) {
                    SfaAuthority currentAuthority = am.getAmAuthority();
                    String rspecComponentManager = currentAuthority.getUrnString();
                    if (!silent)
                        out.println("Contacting  " + rspecComponentManager + "...");

                    try {
                        LOG.debug("ExperimenterCLI is calling status with " + sliceCredentials.size() + " credentials");
                        StatusDetails statusDetails = am.status(usedSliceUrn, sliceCredentials);

                        if (!statusDetails.getAllSliverUrns().isEmpty()) {
                            err.println("Slivers exist at "+rspecComponentManager+" and --abort-if-slivers-exist is set: will not create any new slivers");
                            return -1;
                        }
                        if (debug) {
                            out.println("No slivers at " + rspecComponentManager+"  (got empty sliver list)");
                        }
                    } catch (JFedException e) {
                        //error is expected: normally means there are no slivers
                        if (debug) {
                            out.println("No known slivers at " + rspecComponentManager + "  (got error)");
                        }
                    }
                }
            }

            ManifestRspecMerger manifestRspecMerger = new ManifestRspecMerger(ModelRspecType.BASIC);

            Hashtable outerStatusHt = null;
            Hashtable outerStatusAmDetailHt = null;
            Date earliestSliverExpireDate = null;
            if (isStatus) {
                outerStatusHt = new Hashtable();
                outerStatusHt.put("sliceUrn", sliceUrn);
                if (sliceExpirationDate != null)
                    outerStatusHt.put("sliceExpireDate", RFC3339Util.dateToRFC3339String(sliceExpirationDate));
                outerStatusAmDetailHt = new Hashtable();
                outerStatusHt.put("AMs", outerStatusAmDetailHt);
            }

            for (AggregateManagerWrapper am : amWrappers) {
                SfaAuthority currentAuthority = am.getAmAuthority();
                String rspecComponentManager = currentAuthority.getUrnString();
                GeniUrn rspecComponentManagerUrn = currentAuthority.getUrn();
                if (!silent)
                    out.println("Contacting  " + rspecComponentManager + "...");

                if (isCreate) {
                    try {
                        String requestRspec = rspec.getXmlString();

                        String manifest;
                        if (am.hasSeperateAllocateAndProvision()) {
                            AggregateManagerWrapper.AllocateReply allocateReply =
                                    am.allocate(usedSliceUrn, sliceCredentials, requestRspec, requestedExpirationDate);

                            if (allocateReply != null && allocateReply.getSliverUrns() != null && !allocateReply.getSliverUrns().isEmpty())
                                sa.registerAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, am.getAmAuthority().getUrn(), allocateReply.getSliverUrns(), requestedExpirationDate);

                            manifest =
                                    am.provisionAndStart(usedSliceUrn, sliceCredentials, userSpecs, requestedExpirationDate);

                            if (manifest != null) {
                                Collection<GeniUrn> sliverUrns = ManifestSliverFinder.findSliversInManifest(manifest);
                                if (allocateReply != null && allocateReply.getSliverUrns() != null && !allocateReply.getSliverUrns().isEmpty())
                                    sliverUrns.removeAll(allocateReply.getSliverUrns());
                                if (!sliverUrns.isEmpty())
                                    sa.registerAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, am.getAmAuthority().getUrn(), sliverUrns, requestedExpirationDate);
                            }
                        } else {
                            manifest =
                                    am.createSliver(usedSliceUrn, sliceCredentials, requestRspec, userSpecs, requestedExpirationDate);

                            if (manifest != null) {
                                Collection<GeniUrn> sliverUrns = ManifestSliverFinder.findSliversInManifest(manifest);
                                sa.registerAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, am.getAmAuthority().getUrn(), sliverUrns, requestedExpirationDate);
                            }
                        }

                        if (!silent)
                            out.println("Sliver at  " + rspecComponentManager + " is created and initializing...");

                        manifestRspecMerger.setManifest(currentAuthority.getUrn(), new ManifestRspecSource(manifest, ModelRspecType.BASIC));
                    } catch (AggregateManagerWrapper.NotEnoughFreeResourcesException e) {
                        if (deleteOnCreateFailure)
                            deleteAll(amWrappers, silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
                        return callFailed("Not enough resources available to create sliver at " + currentAuthority.getUrnString(), e, err, callLogOutput);
                    } catch (JFedException e) {
                        if (deleteOnCreateFailure)
                            deleteAll(amWrappers, silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
                        return callFailed("Error creating sliver at " + currentAuthority.getUrnString(), e, err, callLogOutput);
                    }
                }

                if (isDelete)
                    try {
                        am.deleteSliver(usedSliceUrn, sliceCredentials);

                        if (!silent)
                            out.println("Slivers at  " + rspecComponentManager + " have been deleted.");


                        try {
                            sa.unregisterAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, rspecComponentManagerUrn, null/*sliverUrns*/);

                            if (!silent)
                                out.println("Slivers at  " + rspecComponentManager + " has been unregistered.");
                        } catch (JFedException e) {
                            String message =
                                    "Error unregistering sliver at " + rspecComponentManager + ". (this error will be ignored)";
                            if (e.getXmlRpcResult() != null)
                                err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                            else {
                                err.println(message + ":");
                                e.printStackTrace(err);
                            }
                        }
                    } catch (JFedException e) {
                        String message =
                                "Error deleting sliver at " + rspecComponentManager + ". (this error will be ignored)";
                        if (e.getXmlRpcResult() != null)
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                        else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }

                if (isStatus)
                    try {
                        LOG.debug("ExperimenterCLI is calling status with " + sliceCredentials.size() + " credentials");
                        StatusDetails statusDetails = am.status(usedSliceUrn, sliceCredentials);

                        if (!silent && isTextOutput)
                            out.println("Sliver at  " + rspecComponentManager + " has status " + statusDetails.getGlobalStatus());

                        if (isJson || isTextOutput) {
                            Date earliestAMSliverExpireDate = statusDetails.getEarliestKnownExpires();
                            if (earliestSliverExpireDate == null)
                                earliestSliverExpireDate = earliestAMSliverExpireDate;
                            else {
                                if (earliestAMSliverExpireDate != null && earliestAMSliverExpireDate.before(earliestSliverExpireDate))
                                    earliestSliverExpireDate = earliestAMSliverExpireDate;
                            }
                            Hashtable outerAMHt = new Hashtable();
                            outerAMHt.put("amUrn", am.getAmAuthority().getUrnString());
                            outerAMHt.put("amGlobalSliverStatus", statusDetails.getGlobalStatus());
                            if (earliestAMSliverExpireDate != null)
                                outerAMHt.put("earliestSliverExpireDate", RFC3339Util.dateToRFC3339String(earliestAMSliverExpireDate));
                            Vector detailStatus = new Vector();
                            for (Map.Entry<GeniUrn, StatusDetails.SliverStatus> detail : statusDetails.getStatusBySliverUrn().entrySet()) {
                                Hashtable detailHt = new Hashtable();
                                detailHt.put("urn", detail.getKey());
                                detailHt.put("type", "sliver");
                                detailHt.put("status", detail.getValue() + "");
                                detailStatus.add(detailHt);
                            }
                            for (Map.Entry<GeniUrn, StatusDetails.SliverStatus> detail : statusDetails.getStatusByComponentUrn().entrySet()) {
                                Hashtable detailHt = new Hashtable();
                                detailHt.put("urn", detail.getKey());
                                detailHt.put("type", "component");
                                detailHt.put("status", detail.getValue() + "");
                                detailStatus.add(detailHt);
                            }
                            outerAMHt.put("details", detailStatus);
                            outerStatusAmDetailHt.put(am.getAmAuthority().getUrnString(), outerAMHt);
                        }
                    } catch (JFedException e) {
                        String message =
                                "Error getting status of sliver at " + rspecComponentManager + ". (this error will be ignored)";
                        if (e.getXmlRpcResult() != null) {
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                            err.println("Request was: " + "\"" + e.getXmlRpcResult().getRequestXmlRpcString() + "\"");
                        } else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }


                if (isManifest)
                    try {
                        LOG.debug("ExperimenterCLI is fetching manifest with " + sliceCredentials.size() + " credentials");
                        String manifestString = am.describe(usedSliceUrn, sliceCredentials);

                        if (manifestString != null)
                            manifestRspecMerger.setManifest(rspecComponentManagerUrn, new ManifestRspecSource(manifestString, ModelRspecType.BASIC));

                        if (!silent && manifestString != null)
                            out.println("Manifest for " + rspecComponentManager + " received.");
                        if (manifestString == null)
                            err.println("Error fetching manifest for " + rspecComponentManager + "");
                    } catch (JFedException e) {
                        String message =
                                "Error fetching manifest at " + rspecComponentManager + ". (this error will be ignored, but the final manifest might be missing information)";
                        if (e.getXmlRpcResult() != null) {
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                            err.println("Request was: " + "\"" + e.getXmlRpcResult().getRequestXmlRpcString() + "\"");
                        } else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }

                if (isPoa) {
                    if (rspecComponentManagerUrn == null ||
                            !rspecComponentManagerUrn.getTopLevelAuthority_withoutSubAuth().equals(poaTargetSliverUrn.getTopLevelAuthority_withoutSubAuth())) {
                        if (!silent)
                            out.println("No PerformOperationalAction command on  " + rspecComponentManager + ".");
                        continue;
                    }
                    if (!silent)
                        out.println("Executing PerformOperationalAction command on " + rspecComponentManager + "...");
                    try {
                        if (poaActionArg.equalsIgnoreCase("ReloadOS")) {
                            if (!am.isReloadOSSupported()) {
                                err.println("The AM at " + rspecComponentManager + " does not support " + poaActionArg + ". Will try anyway, but the call will likely fail.");
                            }
                            LOG.debug("ExperimenterCLI is calling PerformOperationalAction with " + sliceCredentials.size() + " credentials");
                            am.reloadOS(poaTargetSliverUrn, sliceCredentials);
                        }
                        if (poaActionArg.equalsIgnoreCase("Restart")) {
                            if (!am.isRestartSupported()) {
                                err.println("The AM at " + rspecComponentManager + " does not support " + poaActionArg + ". Will try anyway, but the call will likely fail.");
                            }
                            LOG.debug("ExperimenterCLI is calling PerformOperationalAction with " + sliceCredentials.size() + " credentials");
                            am.restart(poaTargetSliverUrn, sliceCredentials);
                        }
                        if (poaActionArg.equalsIgnoreCase("ConsoleUrl")) {
                            if (!am.isOpenConsoleSupported()) {
                                err.println("The AM at " + rspecComponentManager + " does not support " + poaActionArg + ". Will try anyway, but the call will likely fail.");
                            }
                            LOG.debug("ExperimenterCLI is calling PerformOperationalAction with " + sliceCredentials.size() + " credentials");
                            URL consoleURL = am.openConsole(poaTargetSliverUrn, sliceCredentials);
                            if (consoleURL != null)
                                out.println("The console URN for " + poaTargetSliverUrn + " is " + consoleURL.toExternalForm());
                            else
                                out.println("Error: No console URN was returned for " + poaTargetSliverUrn + "");
                        }
                        if (!silent)
                            out.println("PerformOperationalAction call completed.");
                    } catch (JFedException e) {
                        String message =
                                "Error in PerformOperationalAction " + poaActionArg + " of " + poaTargetSliverUrn + ".";
                        if (e.getXmlRpcResult() != null) {
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                            err.println("Request was: " + "\"" + e.getXmlRpcResult().getRequestXmlRpcString() + "\"");
                        } else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }
                }

                if (isRenew) {
                    Date requestedSliverExpirationDate = sliceExpirationDate == null ? requestedExpirationDate : sliceExpirationDate;
                    assert requestedSliverExpirationDate != null;

                    try {
                        if (!silent) {
                            out.print("Renewing sliver at  " + rspecComponentManager + " to expire at " +
                                    RFC3339Util.dateToRFC3339String(requestedSliverExpirationDate, true, true, true));
                            if (sliceExpirationDate != null)
                                out.println(" (Slice expired at " +
                                        RFC3339Util.dateToRFC3339String(sliceExpirationDate, true, true, true) + ")");
                            else
                                out.println();
                        }

                        am.renewSlivers(usedSliceUrn, null/*no sliverUrns known*/, sliceCredentials, requestedSliverExpirationDate);

                        if (!silent)
                            out.println("Sliver at  " + rspecComponentManager + " has been renewed successfully");

                    } catch (JFedException e) {
                        successExitStatus = false;
                        String message =
                                "Error renewing sliver at " + rspecComponentManager + ". (jFed will first continue Renewing the other AMs and fail afterwards)\n" +
                                        "WARNING: Your slivers (= resources) might now expire before your slice itself expires. You can use the \"status\" command to check this.";
                        if (e.getXmlRpcResult() != null)
                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                        else {
                            err.println(message + ":");
                            e.printStackTrace(err);
                        }
                    }

                    if (sa.isRegisterAggregatesForSliceSupported())
                        try {
                            StatusDetails statusDetails = am.status(usedSliceUrn, sliceCredentials);

                            Collection<GeniUrn> sliverUrns = statusDetails.getAllSliverUrns();
                            Date firstExpires = statusDetails.getEarliestKnownExpires();
                            Date lastExpires = statusDetails.getLatestKnownExpires();
                            Date updateExpireDate = requestedSliverExpirationDate;
                            if (firstExpires != null && lastExpires != null) {
                                updateExpireDate = lastExpires; //better register them too long than too short
                                if (!firstExpires.equals(lastExpires)) {
                                    if (!silent)
                                        err.println("WARNING: The slivers at "+rspecComponentManager+" do not expire at the same time. " +
                                                "first expire time = "+firstExpires+" last expire time = "+lastExpires+"" +
                                                " -> Will update all slivers as expiring at "+updateExpireDate);
                                }
                            }

                            sa.updateSliverExpirationDateInfoForSlice(logger, sliceCredentials, usedSliceUrn, sliverUrns, updateExpireDate);
                        } catch (Exception e) {
                            String message =
                                    "Error registering renewed sliver(s) at " + rspecComponentManager + ". This is not critical, and will be ignored. This will not influence the exit code.";
                            if (e instanceof JFedException) {
                                JFedException je = (JFedException) e;
                                if (je.getXmlRpcResult() != null)
                                    err.println(message + ": " + "\"" + je.getXmlRpcResult().getResultXmlRpcString() + "\"");
                                else {
                                    err.println(message + ":");
                                    e.printStackTrace(err);
                                }
                            } else {
                                err.println(message + ":");
                                e.printStackTrace(err);
                            }
                        }
                }
            }

            if (isManifest || isCreate) {
                if (debug || (!silent && isManifest))
                    out.println("Writing combined manifest to file \"" + manifestArg + "\"");
                String combinedManifestXmlString = manifestRspecMerger.getMergedManifest().getRspecXmlString();
                IOUtils.stringToFile(new File(manifestArg), combinedManifestXmlString);

                if (ansibleDir != null) {
                    if (!ansibleDir.exists()) {
                        boolean created = ansibleDir.mkdir();
                        if (!ansibleDir.exists()) {
                            err.println("ERROR: Could not create ansible dir: \""+ansibleDirName+"\"");
                        }
                    }
                    if (ansibleDir.exists()) {
                        AnsibleFileWriter writer = null;
                        if (user.getPrivateKeyFile() != null) {
                            try {
                                String privKey = IOUtils.fileToString(user.getPrivateKeyFile());
                                if (!KeyUtil.isPemPrivateKeyEncrypted(privKey)) {
                                    writer = AnsibleFileWriter.createWithLinkedPrivateKey(
                                            new BasicStringRspec(combinedManifestXmlString),
                                            user.getPrivateKeyFile(),
                                            user.getPublicKey(),
                                            user.getUserUrnString(),
                                            null/*sshProxyInfo*/,
                                            null /*preferredUser*/);
                                }
                            } catch (Exception e) {
                                LOG.error("Exception trying to determine if private key file is encrypted. Safe fallback: Will assume it is.", e);
                                writer = null;
                            }
                        }
                        if (writer == null) {
                            writer = AnsibleFileWriter.createWithCopiedPrivateKey(
                                    new BasicStringRspec(combinedManifestXmlString),
                                    user.getPrivateKey(),
                                    user.getPublicKey(),
                                    user.getUserUrnString(),
                                    null, null /*preferredUser*/);
                        }
                        writer.writeFilesToDir(ansibleDir);
                    }
                }
            }

            if (isStatus) {
                if (earliestSliverExpireDate != null)
                    outerStatusHt.put("earliestSliverExpireDate", RFC3339Util.dateToRFC3339String(earliestSliverExpireDate));
                if (isJson && outerStatusHt != null) {
                    out.println(JSonHelper.jsonObjectToString(outerStatusHt));
                }
                if (isTextOutput && outerStatusHt != null) {
                    out.println("Status Info:");
                    out.println("   Slice URN: "+outerStatusHt.get("sliceUrn"));
                    out.println("   Slice expire date: "+outerStatusHt.get("sliceExpireDate"));
                    if (outerStatusHt.get("earliestSliverExpireDate") != null)
                        out.println("   Earliest sliver expire date: "+outerStatusHt.get("earliestSliverExpireDate"));
                    else
                        out.println("   Earliest sliver expire date: no slivers");
                    Hashtable amsHt = (Hashtable) outerStatusHt.get("AMs");
                    Set<Map.Entry> entries = amsHt.entrySet();
                    for (Map.Entry e : entries) {
                        String key = (String) e.getKey();
                        Hashtable amHt = (Hashtable)e.getValue();

                        String amUrn = amHt.get("amUrn")+"";
                        String amGlobalSliverStatus = amHt.get("amGlobalSliverStatus")+"";
                        String earliestAMSliverExpireDateStr = amHt.get("earliestSliverExpireDate") == null ? null : amHt.get("earliestSliverExpireDate")+"";

                        out.println("   AM: "+amUrn+" "+amGlobalSliverStatus+(earliestAMSliverExpireDateStr == null ? "" : " "+earliestAMSliverExpireDateStr));

                        Vector details= (Vector) amHt.get("details");
                        for (Object detail : details) {
                            Hashtable detailHt = (Hashtable) detail;
                            String urn = detailHt.get("urn")+"";
                            String type = (String)detailHt.get("type");
                            String status = detailHt.get("status")+"";
                            out.println("      "+type+": "+urn+" "+status);
                        }
                    }
                }
            }

            if (isCreate) {
                if (waitUntilReady) {
                    if (!silent)
                        if (rspecComponentManagers.size() == 1)
                            out.println("Will now wait until the sliver is ready...");
                        else
                            out.println("Will now wait until the " + rspecComponentManagers.size() + " slivers are ready...");

                    for (AggregateManagerWrapper am : amWrappers) {
                        SfaAuthority currentAuthority = am.getAmAuthority();
                        String rspecComponentManager = currentAuthority.getUrnString();

                        StatusDetails statusDetails = null;
                        while (statusDetails == null || !statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.READY)) {
                            if (debug)
                                out.println("Waiting 5 seconds before checking status...");
                            try {
                                Thread.sleep(5 * 1000l);
                            } catch (InterruptedException e) {
                                LOG.debug("InterruptedException while sleeping", e);
                            }

                            if (!silent)
                                out.println("Contacting " + rspecComponentManager + " to check status...");

                            try {
                                statusDetails = am.status(usedSliceUrn, sliceCredentials);
                            } catch (JFedException e) {
                                if (deleteOnBecomeReadyFailure)
                                    deleteAll(amWrappers, silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
                                return callFailed("Error getting sliver status at " + currentAuthority.getUrnString(), e, err, callLogOutput);
                            }

                            if (!silent)
                                out.println("Status of sliver at  " + rspecComponentManager + " is " + statusDetails.getGlobalStatus());

                            if (statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.FAIL) ||
                                    statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.UNALLOCATED)) {
                                if (deleteOnBecomeReadyFailure)
                                    deleteAll(amWrappers, silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
                                return callFailed("ERROR: Sliver status at " + currentAuthority.getUrnString() + " is " +
                                        statusDetails.getGlobalStatus() + " -> " + statusDetails, null, err, callLogOutput);
                            }
                        }
                    }

                    if (!lanToShare.isEmpty()) {
                        LOG.debug("Will try to share LANs");

                        if (!silent)
                            out.println((lanToShare.size())+" LAN(s) need to be shared. Will wait 10 seconds before trying.");
                        try {
                            Thread.sleep(10);
                        } catch (InterruptedException e) {
                            //ignore
                        }

                        for (AggregateManagerWrapper am : amWrappers) {
                            SfaAuthority currentAuthority = am.getAmAuthority();

                            for (AutoShareLanInfo info : lanToShare) {
                                if (currentAuthority.getUrn().equals(info.componentManagerUrn)) {
                                    LOG.debug("Will try to share LAN "+info.lanClientId+" as "+info.sharedLanName+" on "+info.componentManagerUrn);
                                    if (am.isShareLanSupported()) {
                                        try {
                                            am.shareLan(usedSliceUrn, sliceCredentials, info.lanClientId, info.sharedLanName);

                                            if (!silent) {
                                                out.println("Shared LAN " + info.lanClientId + " as " + info.sharedLanName + " on " + info.componentManagerUrn);
                                            }
                                        } catch (JFedException e) {
                                            LOG.error("Failed to share LAN", e);
                                            err.println("Failed to share LAN " + info.lanClientId + " as " + info.sharedLanName + " on " + info.componentManagerUrn + ": "+e.getMessage());
                                            e.printStackTrace();
                                        }
                                    } else {
                                        err.println("Cannot share LAN "+info.lanClientId+" as "+info.sharedLanName+" on "+info.componentManagerUrn+": AM does not support share LAN");
                                    }
                                }
                            }

                        }
                    }

                    if (!silent)
                        if (rspecComponentManagers.size() == 1)
                            out.println("The sliver is ready.");
                        else
                            out.println("The " + rspecComponentManagers.size() + " slivers are ready.");
                }
            }

            if (!ansiblePlaybooksToExecute.isEmpty()) {
                LOG.debug("Will execute "+ansiblePlaybooksToExecute.size()+" Ansible Playbooks");

                if (!silent)
                    out.println((ansiblePlaybooksToExecute.size())+" ansible playbooks will be executed. Will wait 5 seconds before beginning.");
                try {
                    Thread.sleep(5);
                } catch (InterruptedException e) {
                    //ignore
                }

                int playbookIndex = 0;
                for (AnsiblePlaybookInfo info : ansiblePlaybooksToExecute) {
                    File outputFile;
                    if (info.getExecuteAnsiblePlaybook().getOutputFilename() == null)
                        outputFile = new File(ansibleDir, "playbook-"+playbookIndex+"-output.txt");
                    else
                        outputFile = ansibleAllowAnyPlaybookOutputFile ? new File(info.getOutputFilename()) : new File(ansibleDir, info.getOutputBasename());
                    File playbookFile = new File(ansibleDir, "playbook-"+playbookIndex+".yml");
                    try {
                        IOUtils.stringToFile(playbookFile, info.getScriptContent());

                        if (!silent)
                            out.println("Executing ansible playbook \""+playbookFile.getName()+"\"...");
                        callAnsible(ansiblePlayBookExe, playbookFile, ansibleDir, ansibleDebug, outputFile);
                    } catch (Exception e) {
                        successExitStatus = false;
                        err.println("Exception while executing ansible playbook (will still try other ansible playbooks, if any, but exit status will indicate failure). Exception:");
                        e.printStackTrace();
                    }

                    playbookIndex++;
                }
            }
        } finally {
            callLogOutput.close();

            if (printCalls) {
                Thread.yield();
                try {
                    Thread.sleep(500l);
                } catch (InterruptedException e) {
                    LOG.error("Unexpected InterruptedException", e);
                }
            }
        }

        if (debug)
            out.println("All done. Exiting.");

        if (successExitStatus)
            return 0;
        else
            return 2;
    }

    private static void callAnsible(String ansiblePlaybookExe, File ansiblePlaybook, File ansibleDir, boolean ansibleDebug, File outputFile) throws IOException {
        PrintWriter pw = new PrintWriter(new FileWriter(outputFile));
        ArrayList<String> command = new ArrayList<String>();
        assert ansiblePlaybookExe != null;
        String[] ansiblePlaybookExeParts = ansiblePlaybookExe.split(" ");
        for (String ansiblePlaybookExePart : ansiblePlaybookExeParts)
            command.add(ansiblePlaybookExePart);
        command.add(ansiblePlaybook.getAbsolutePath());
        if (ansibleDebug)
            command.add("-vvvv");
        ProcessBuilder pb = new ProcessBuilder(command);
        pb.directory(ansibleDir);
        pb.redirectErrorStream(); //combine stderr and stout, so we can read both at once in one thread. Makes everything easy.

        Process p = pb.start();
        BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));

//        String output = "";
        String line;
        while ((line = input.readLine()) != null) {
//            output += line + System.lineSeparator();
            pw.println(line);
        }
        //note: input might close before terminal stops.
        input.close();

        try {
            p.waitFor();
        } catch (InterruptedException e) {
            //ignore
        }

        pw.close();
    }


    private static void deleteAll(List<AggregateManagerWrapper> amWrappers, boolean silent,
                                  PrintStream out, PrintStream err, GeniUrn usedSliceUrn,
                                  List<AnyCredential> sliceCredentials, Logger logger,
                                  UserAndSliceApiWrapper sa) {
        if (!silent)
            err.println("Failure: will delete all slivers, on "+amWrappers+" AMs.");

        for (AggregateManagerWrapper am : amWrappers) {
            SfaAuthority currentAuthority = am.getAmAuthority();
            String rspecComponentManager = currentAuthority.getUrnString();
            GeniUrn rspecComponentManagerUrn = currentAuthority.getUrn();
            if (!silent)
                out.println("Contacting  " + rspecComponentManager + "...");

            try {
                am.deleteSliver(usedSliceUrn, sliceCredentials);

                if (!silent)
                    out.println("Slivers at  " + rspecComponentManager + " have been deleted.");

                try {
                    sa.unregisterAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, rspecComponentManagerUrn, null/*sliverUrns*/);

                    if (!silent)
                        out.println("Slivers at  " + rspecComponentManager + " has been unregistered.");
                } catch (JFedException e) {
                    String message =
                            "Error unregistering sliver at " + rspecComponentManager + ". (this error will be ignored)";
                    if (e.getXmlRpcResult() != null)
                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                    else {
                        err.println(message + ":");
                        e.printStackTrace(err);
                    }
                }
            } catch (JFedException e) {
                String message =
                        "Error deleting sliver at " + rspecComponentManager + ". (this error will be ignored)";
                if (e.getXmlRpcResult() != null)
                    err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
                else {
                    err.println(message + ":");
                    e.printStackTrace(err);
                }
            }
        }
    }

    private static int callFailed(String message, JFedException e, PrintStream err, CallLogOutput callLogOutput) {
        Thread.yield();
        Thread.yield();
        callLogOutput.flush();
        try {
            Thread.sleep(100);
        } catch (InterruptedException e1) {
            //ignored
        }
        Thread.yield();
        callLogOutput.flush();

        if (e == null)
            err.println(message);
        else {
            if (e.getXmlRpcResult() != null)
                err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
            else {
                err.println(message + ":");
                e.printStackTrace(err);
            }
        }
        return 2;
    }

    private static void addToUserSpecs(List<UserSpec> userSpecs, GeniUrn userUrn, String sshKey) {
        assert userUrn != null;
        addToUserSpecs(userSpecs, userUrn, Collections.singletonList(sshKey));
    }
    private static void addToUserSpecs(List<UserSpec> userSpecs, String userUrnString, String sshKey) {
        addToUserSpecs(userSpecs, userUrnString, Collections.singletonList(sshKey));
    }
    private static void addToUserSpecs(List<UserSpec> userSpecs, GeniUrn userUrn, Collection<String> sshKeys) {
        assert userUrn != null;
        addToUserSpecs(userSpecs, userUrn.toString(), sshKeys);
    }
    private static void addToUserSpecs(List<UserSpec> userSpecs, String userUrnString, Collection<String> sshKeys) {
        if (sshKeys == null || sshKeys.isEmpty())
            return;
        for (UserSpec us : userSpecs) {
            if (us.getUrn().equals(userUrnString)) {
                //TODO (low priority) make this work for SSH keys that differ only in comment, but not in actual key
                Set<String> newSskKeys = new HashSet<>(sshKeys);
                newSskKeys.removeAll(us.getSshKey());
                if (!newSskKeys.isEmpty())
                    us.getSshKey().addAll(newSskKeys);
                return;
            }
        }

        UserSpec newUs = new UserSpec(userUrnString, sshKeys);
        userSpecs.add(newUs);
    }

    private static class CallLogOutput {
        public FileOutputStream fos;
        public XMLEventWriter xtw;
        public XMLEventFactory eventFactory;
        public final PrintStream err;

        public CallLogOutput(PrintStream err) {
            this.err = err;
        }

        public void onFailure() {
            close();
            fos = null;
            xtw = null;
            eventFactory = null;
        }

        public void flush() {
            try {
                xtw.flush();
                fos.flush();
            } catch (Exception e) {
                LOG.warn("Unexpected error flusing", e);
            }
        }

        public boolean isDisabled() {
            return fos == null || xtw == null || eventFactory == null;
        }

        public void close() {
            if (xtw != null && eventFactory != null) {
                try {
                    xtw.add(eventFactory.createEndElement(new QName("calls"), null));
                    xtw.add(eventFactory.createEndDocument());
                    xtw.flush();
                } catch (XMLStreamException e) {
                    LOG.error("Exception while closing call log output", e);
                    err.println("Exception while closing call log output: " + e.getMessage());
                    e.printStackTrace();
                }
            }
            if (fos != null) {
                try {
                    fos.close();
                } catch (IOException e) {
                    LOG.error("Exception while closing call log output file", e);
                    err.println("Exception while closing call log output file: " + e.getMessage());
                }
            }
        }
    }

    public static List<String> extractCredentials(String multiCredentialString) {
        List<String> res = new ArrayList<>();

        //There can be multiple speaksfor credentials in one file.
        //We don't parse XML to detect them, we use some simple regex matching tricks to catch all(?) cases

        //NameSpaceStartChar	   ::=   	[A-Z] | "_" | [a-z] | [#xC0-#xD6] | [#xD8-#xF6] | [#xF8-#x2FF] | [#x370-#x37D] | [#x37F-#x1FFF] | [#x200C-#x200D] | [#x2070-#x218F] | [#x2C00-#x2FEF] | [#x3001-#xD7FF] | [#xF900-#xFDCF] | [#xFDF0-#xFFFD] | [#x10000-#xEFFFF]
        //NameSpaceChar	   ::=   	NameStartChar | "-" | "." | [0-9] | #xB7 | [#x0300-#x036F] | [#x203F-#x2040]

        //<signed-credential or <[NameSpaceStartChar][NameSpaceChar]*:signed-credential
        //</signed-credential.*> or <[NameSpaceStartChar][NameSpaceChar]*:signed-credential.*>

        //TODO: add special chars

        String namespaceStartChar = "A-Za-z_\\xC0-\\xD6\\xD8-\\xF6\\xF8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD"; //java doesn't seem to like \x{10000}-\x{EFFFF}
        String namespaceChar = namespaceStartChar+".0-9\\xB7\\u0300-\\u036F\\u203F-\\u2040-";
        Pattern startCredentialPattern = Pattern.compile("<(["+namespaceStartChar+"]["+namespaceChar+"]*:)?signed-credential");
        Pattern endCredentialPattern = Pattern.compile("</(["+namespaceStartChar+"]["+namespaceChar+"]*:)?signed-credential[^>]*>");

        Matcher startMatcher = startCredentialPattern.matcher(multiCredentialString);
        Matcher endMatcher = endCredentialPattern.matcher(multiCredentialString);

        int i = 0;
        while (startMatcher.find(i) && endMatcher.find(startMatcher.end())) {
            String credentialString = multiCredentialString.substring(startMatcher.start(), endMatcher.end());
            res.add(credentialString);
            i = endMatcher.end();
        }

        return res;
    }

    private static ModelRspec getModelRspec(ModelRspec existingModel, StringRspec rspec) throws RspecParseException {
        if (existingModel != null)
            return existingModel;

        RequestRspecSource rspecSource = new RequestRspecSource(rspec, ModelRspecType.BASIC);
        ModelRspec modelRspec = rspecSource.getModelRspec();
        return modelRspec;

//        return new StaxRspecParser(rspec.getXmlString(), BasicRspecFactory.getInstance()).call();
    }

    private static class AutoShareLanInfo {
        private final GeniUrn componentManagerUrn;
        private final String lanClientId;
        private final String sharedLanName;

        public AutoShareLanInfo(GeniUrn componentManagerUrn, String lanClientId, String sharedLanName) {
            this.componentManagerUrn = componentManagerUrn;
            this.lanClientId = lanClientId;
            this.sharedLanName = sharedLanName;
        }
    }

    private static class AnsiblePlaybookInfo {
        public static class AnsiblePlaybookException extends Exception {
            public AnsiblePlaybookException(String message) {
                super(message);
            }

            public AnsiblePlaybookException(String message, Throwable cause) {
                super(message, cause);
            }
        }

        private final ExecuteAnsiblePlaybook executeAnsiblePlaybook;
        private String outputFilename;
        private String outputBasename;
        private String scriptContent;

        public AnsiblePlaybookInfo(ExecuteAnsiblePlaybook executeAnsiblePlaybook) {
            this.executeAnsiblePlaybook = executeAnsiblePlaybook;
            this.outputFilename = executeAnsiblePlaybook.getOutputFilename();
        }

        public void checkFilename(boolean mustBeBasename, boolean allowAnyOutputFile) throws AnsiblePlaybookException {
            if (outputFilename == null)
                return;

            boolean isBasename = outputFilename.indexOf(File.separatorChar) == -1;

            if (mustBeBasename && !isBasename)
                throw new AnsiblePlaybookException("Invalid Output filename: "+
                        RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart()+" "+
                        RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart()+" \""+
                        outputFilename+"\" is not a base filename.");

            if (isBasename) {
                outputBasename = outputFilename;
                return;
            }

            File outputFile = new File(outputFilename);
            if (allowAnyOutputFile) {
                if (!outputFile.getParentFile().exists())
                    throw new AnsiblePlaybookException("Invalid Output filename: " +
                            RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + " " +
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + " \"" +
                            outputFilename + "\" parent directory does not exist.");

                if (outputFile.exists() && !outputFile.isFile())
                    throw new AnsiblePlaybookException("Invalid Output filename: " +
                            RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + " " +
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_OUTPUT_FILE.getLocalPart() + " \"" +
                            outputFilename + "\" exists but is not a normal file.");
            } else {
                //else don't check, because we won't use it anyway.

                //overwrite outputFilename to be sure the full path is not used
                outputFilename = outputFile.getName();
            }

            outputBasename = outputFile.getName();
        }

        public void loadContent(boolean allowFile) throws AnsiblePlaybookException {
            if (executeAnsiblePlaybook.getSource() == null)
                throw new AnsiblePlaybookException("Invalid RSpec: "+
                        RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart()+" has no "+
                        RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart()+" attribute");

            if (executeAnsiblePlaybook.getSource().startsWith("http")) {
                URL url;
                try {
                    url = new URL(executeAnsiblePlaybook.getSource());
                } catch (MalformedURLException e) {
                    throw new AnsiblePlaybookException("Invalid "+
                            RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart()+" "+
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart()+
                            " url: \""+
                            executeAnsiblePlaybook.getSource()+
                            "\"", e);
                }
                try {
                    this.scriptContent = IOUtils.urlToString(url, Charset.defaultCharset());
                } catch (IOException e) {
                    throw new AnsiblePlaybookException("Error reading URL in "+
                            RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart()+" "+
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart()+
                            " attribute: \""+
                            executeAnsiblePlaybook.getSource()+
                            "\": "+e.getMessage(), e);
                }
            } else {
                if (!allowFile)
                    throw new AnsiblePlaybookException(""+
                            RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart()+" "+
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart()+
                            " is \""+
                            executeAnsiblePlaybook.getSource()+
                            "\" but the --ansible-allow-playbook-inputfile option is not set, " +
                            "so only URLs are allowed.");

                File localFile = new File(executeAnsiblePlaybook.getSource());

                if (!localFile.exists())
                    throw new AnsiblePlaybookException("Bad "+
                            RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart()+" "+
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart()+
                            " filename: \""+
                            executeAnsiblePlaybook.getSource()+
                            "\" does not exist");

                try {
                    this.scriptContent = IOUtils.fileToString(localFile);
                } catch (IOException e) {
                    throw new AnsiblePlaybookException("Error reading "+
                            RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart()+" "+
                            RspecXmlConstants.Q_ATT_JFED_EXECUTE_ANSIBLE_PLAYBOOK_SOURCE.getLocalPart()+
                            " file: \""+
                            executeAnsiblePlaybook.getSource()+
                            "\": "+e.getMessage(), e);
                }
            }
        }

        public ExecuteAnsiblePlaybook getExecuteAnsiblePlaybook() {
            return executeAnsiblePlaybook;
        }

        public String getOutputFilename() {
            return outputFilename;
        }

        public String getScriptContent() {
            return scriptContent;
        }

        public String getOutputBasename() {
            return outputBasename;
        }
    }

    private static boolean isBefore(Date a, Date b) {
        Date aa = new Date(a.getTime() + 999l);
        return aa.before(b);
    }
}
