package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experiment.CapabilityChecker;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.call_gui.TasksWindow;
import be.iminds.ilabt.jfed.experimenter_gui.ext_requests.JFedEnvironment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewControllerFactory;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.jfed.gts_highlevel.jobs.GtsJobFactory;
import javafx.application.HostServices;
import javafx.collections.ObservableMap;
import javafx.stage.Window;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;

/**
 * User: twalcari
 * Date: 12/2/2014
 * Time: 16:37
 */
@Singleton
public class DialogsFactoryImpl implements DialogsFactory {

    private final TasksWindow tasksWindow;

    private final GeniUserProvider geniUserProvider;
    private final SfaModel sfaModel;
    private final HighLevelTaskFactory hltf;
    private final ObservableMap<GeniUrn, ExperimentViewController> experimentViewControllers;
    private final ExperimentViewControllerFactory experimentViewControllerFactory;
    private final JFedGuiPreferences jFedPreferences;
    private final AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
            automaticAggregateManagerWrapperFactory;
    private final AuthorityListModel authorityListModel;
    private final AuthorityFinder authorityFinder;
    private final JFedEnvironment environment;
    private final Provider<HostServices> hostServicesProvider;
    private final GtsJobFactory gtsJobFactory;
    private final Provider<CapabilityChecker> capabilityCheckerProvider;

    @Inject
    public DialogsFactoryImpl(TasksWindow tasksWindow, GeniUserProvider geniUserProvider, SfaModel sfaModel, HighLevelTaskFactory hltf,
                              ObservableMap<GeniUrn, ExperimentViewController> experimentViewControllers, ExperimentViewControllerFactory experimentViewControllerFactory,
                              JFedGuiPreferences jFedPreferences,
                              AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory,
                              AuthorityListModel authorityListModel, AuthorityFinder authorityFinder, JFedEnvironment environment, Provider<HostServices> hostServicesProvider, GtsJobFactory gtsJobFactory, Provider<CapabilityChecker> capabilityCheckerProvider) {
        this.tasksWindow = tasksWindow;
        this.geniUserProvider = geniUserProvider;
        this.sfaModel = sfaModel;
        this.hltf = hltf;
        this.experimentViewControllers = experimentViewControllers;
        this.experimentViewControllerFactory = experimentViewControllerFactory;
        this.jFedPreferences = jFedPreferences;
        this.automaticAggregateManagerWrapperFactory = automaticAggregateManagerWrapperFactory;
        this.authorityListModel = authorityListModel;
        this.authorityFinder = authorityFinder;
        this.environment = environment;
        this.hostServicesProvider = hostServicesProvider;
        this.gtsJobFactory = gtsJobFactory;
        this.capabilityCheckerProvider = capabilityCheckerProvider;
    }

    @Override
    public void showCreateSliceDialog(Window parentWindow, RequestRspecSource requestRspecSource) {
        new CreateSliceDialog(parentWindow, requestRspecSource, hltf, geniUserProvider, sfaModel, experimentViewControllers,
                experimentViewControllerFactory, authorityListModel, authorityFinder, environment).showDialog();
    }

    @Override
    public void showRebootDialog(Window parentWindow, ExperimentController experimentController) {
        new RebootDialog(parentWindow, experimentController, automaticAggregateManagerWrapperFactory).showDialog();
    }

    @Override
    public void showEditSshKeysDialog(Window parentWindow, ExperimentController experimentController) {
        new EditSshKeysDialog(parentWindow, experimentController, geniUserProvider,
                sfaModel, jFedPreferences, capabilityCheckerProvider.get()).showDialog();
    }

    @Override
    public void showShareSliceDialog(Window parentWindow, ExperimentController experimentController) {
        new ShareSliceDialog(parentWindow, experimentController,
                geniUserProvider, sfaModel, hltf).showDialog();

    }

    @Override
    public void showCreateReservationDialog(Window parentWindow, RequestRspecSource requestRspecSource) {
        new CreateReservationDialog(parentWindow, requestRspecSource, hltf, geniUserProvider, sfaModel,
                experimentViewControllers, experimentViewControllerFactory,
                hostServicesProvider, authorityFinder).showDialog();
    }

    @Override
    public void showRenewSliceDialog(Window parentWindow, ExperimentController experimentController) {
        new RenewSliceDialog(parentWindow, experimentController,  authorityFinder).showDialog();
    }

    @Override
    public void showTerminateSliceDialog(Window parentWindow, ExperimentController experimentController) {
        new TerminateSliceDialog(parentWindow, experimentController/*, gtsJobFactory, authorityFinder*/).showDialog();
    }

}
