package be.iminds.ilabt.jfed.experimenter_gui.editor;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.actions.EditorActionsFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.views.EditorViewsFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ExperimentEditorTabHost;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;

import javax.inject.Inject;
import javax.inject.Singleton;

/**
 * User: twalcari
 * Date: 12/8/2014
 * Time: 15:58
 */
@Singleton
public class ExperimentEditorTabFactory {

    private final ExperimentEditorTabHost experimentEditorTabHost;

    private final JFedConfiguration jFedConfiguration;
    private final AuthorityList authorityList;
    private final EditorViewsFactory editorViewsFactory;
    private final EditorActionsFactory editorActionsFactory;


    @Inject
    public ExperimentEditorTabFactory(ExperimentEditorTabHost experimentEditorTabHost,
                                      AuthorityList authorityList, EditorViewsFactory editorViewsFactory, EditorActionsFactory editorActionsFactory,
                                      JFedConfiguration jFedConfiguration) {
        this.authorityList = authorityList;
        this.editorViewsFactory = editorViewsFactory;
        this.editorActionsFactory = editorActionsFactory;
        this.experimentEditorTabHost = experimentEditorTabHost;
        this.jFedConfiguration = jFedConfiguration;
    }


    public ExperimentEditorTab createExperimentEditorTab() {
        EditableRspec editableRspec = new EditableRspec(getNextNewExperimentDefinitionName());

        return new ExperimentEditorTab(editableRspec, authorityList, editorViewsFactory, editorActionsFactory, jFedConfiguration);
    }


    public ExperimentEditorTab createExperimentEditorTab(EditableRspec editableRspec) {
        return new ExperimentEditorTab(editableRspec, authorityList, editorViewsFactory, editorActionsFactory, jFedConfiguration);
    }


    /**
     * @return A new unique experiment definition name
     */
    private String getNextNewExperimentDefinitionName() {
        String baseName = "Untitled";

        if (!experimentDefinitionNameExists(baseName))
            return baseName;

        int counter = 1;
        String name = baseName + counter;
        while (experimentDefinitionNameExists(name)) {
            counter++;
            name = baseName + counter;
        }
        return name;
    }


    /**
     * Tests whether a given experiment definition is already open in the GUI
     */
    private boolean experimentDefinitionNameExists(String name) {
        return experimentEditorTabHost.getExperimentEditorTabs()
                .stream()
                .map(ExperimentEditorTab::getExperimentEditor)
                .map(ExperimentEditor::getEditableRspec)
                .map(EditableRspec::getName)
                .anyMatch(name::equals);
    }

}
