/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.controller;


import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.OntologyAdaptivePropertiesDialog;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.ResultMapDialog;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.model.QueryFilter;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.model.query.SparqlService;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.view.FilterCriteriaView;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.view.MapFilterView;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.view.RangeFilterView;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.view.ValueFilterView;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.FlatIcon;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.binding.Bindings;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.FlowPane;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;
import javafx.util.StringConverter;
import org.apache.jena.query.QueryException;
import org.apache.jena.query.QueryFactory;
import org.apache.jena.query.Syntax;
import org.controlsfx.dialog.ExceptionDialog;
import org.controlsfx.glyphfont.FontAwesome;

import java.util.*;
import java.util.stream.Collectors;

import static be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.model.query.SparqlService.NODE_QUERY_START;
import static be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.model.query.SparqlService.SPARQL_PREFIXES_DEFENITIONS;
import static javafx.scene.control.Alert.AlertType.ERROR;
import static javafx.scene.control.Alert.AlertType.WARNING;

/**
 * @author Patrick
 */
public class NodePropertyOverview extends BorderPane {

    public static final StringConverter<SparqlService.NodeInfo>
            NODE_INFO_STRING_CONVERTER = new StringConverter<SparqlService.NodeInfo>() {
        @Override
        public String toString(SparqlService.NodeInfo object) {
            return "[" + object.getComponentManagerUrn().getTopLevelAuthority() + "] " + object.getComponentUrn().getResourceName();
        }

        @Override
        public SparqlService.NodeInfo fromString(String string) {
            throw new UnsupportedOperationException();
        }
    };
    @FXML
    private MenuButton mnuCriteria;

    @FXML
    private TextField nameTextField;

    @FXML
    private FlowPane flowPane;

    @FXML
    private BorderPane basicPane;

    @FXML
    private BorderPane advancedPane;

    @FXML
    private VBox advancedContentPane;

    @FXML
    private Button searchButton;

    @FXML
    private ComboBox<SparqlService.NodeInfo> nodeList;

    @FXML
    private Button switchToAdvancedButton;
    @FXML
    private Button switchToBasicButton;

    @FXML
    private ToggleButton mapButton;

    @FXML
    private Button cancelButton;

    @FXML
    private Button saveButton;

    @FXML
    private GridPane searchGrid;

    private TextArea advancedQueryTextArea;

    private BooleanProperty advanced = new SimpleBooleanProperty(false);

    private OntologyAdaptivePropertiesDialog mainApp;

    private final SparqlService service = new SparqlService();

    @FXML
    private ProgressIndicator indicator;

    private static final Comparator<GeniUrn> URN_COMPARATOR = (o1, o2) -> o1.toString().compareTo(o2.toString());


    @FXML
    private void initialize() {
        searchButton.disableProperty().bind(service.runningProperty());
        mnuCriteria.disableProperty().bind(service.runningProperty());
    }

    private Map<String, List<String>> crit;
    private Map<String, String> reversedCrit;

    private List<QueryFilter> queryFilters;
    private ToggleGroup toggleGroup;
    private RspecNode node;
    private ModelRspec modelRspec;

    private final ResultMapDialog resultMapDialog;



    public NodePropertyOverview() {

        FXMLUtil.injectFXML(this);

        nameTextField.setTooltip(
                TooltipBuilder.create()
                        .text("Only alphanumerical, maxLat. " + getMaximumNodeNameLength() + " characters")
                        .build()
        );

        nodeList.setConverter(NODE_INFO_STRING_CONVERTER);

        //create advanced pane
        TextArea beforeQueryTextArea = new TextArea();
        beforeQueryTextArea.setEditable(false);
        beforeQueryTextArea.setStyle("-fx-background-color: #cccccc");
        beforeQueryTextArea.setText(SPARQL_PREFIXES_DEFENITIONS + NODE_QUERY_START);
        beforeQueryTextArea.setPrefRowCount(16);

        advancedQueryTextArea = new TextArea();
        advancedQueryTextArea.setWrapText(true);
        advancedQueryTextArea.setPromptText("Enter SPARQL-Query");
        advancedQueryTextArea.textProperty().addListener((observable) -> {
            if (advancedQueryTextArea.getText().isEmpty()) {
                advancedQueryTextArea.setStyle("-fx-text-fill: black;");
            } else if (isAdvancedQueryValid()) {
                advancedQueryTextArea.setStyle("-fx-text-fill: green;");
            } else {
                advancedQueryTextArea.setStyle("-fx-text-fill: red;");
            }
        });
        advancedQueryTextArea.setPrefRowCount(10);

        TextArea afterQueryTextArea = new TextArea();
        afterQueryTextArea.setEditable(false);
        afterQueryTextArea.setStyle("-fx-background-color: #cccccc");
        afterQueryTextArea.setText(SparqlService.NODE_QUERY_END);
        afterQueryTextArea.setPrefRowCount(1);

        advancedContentPane.getChildren().addAll(beforeQueryTextArea, advancedQueryTextArea, afterQueryTextArea);

        basicPane.visibleProperty().bind(advanced.not());
        advancedPane.visibleProperty().bind(advanced);

        searchButton.disableProperty().bind(service.runningProperty());
        indicator.visibleProperty().bind(service.runningProperty());

        resultMapDialog = new ResultMapDialog();

        resultMapDialog.resultProperty().addListener(observable -> {
            nodeList.getSelectionModel().select(resultMapDialog.getResult());
        });
        resultMapDialog.getDialogStage().setOnHidden(event -> mapButton.setSelected(false));

        mapButton.disableProperty().bind(Bindings.isEmpty(nodeList.getItems()));

        service.setOnSucceeded(event -> {
            final List<SparqlService.NodeInfo> result = new ArrayList<SparqlService.NodeInfo>(service.getValue());

            Collections.sort(result, Comparator.comparing(SparqlService.NodeInfo::getComponentManagerUrn, URN_COMPARATOR).thenComparing(SparqlService.NodeInfo::getComponentUrn, URN_COMPARATOR));
            nodeList.getItems().setAll(result);

            Optional<SparqlService.NodeInfo> currentNodeInfo =
                    result.stream().filter(ni -> ni.getComponentManagerUrn().equals(node.getComponentManagerId()) && ni.getComponentUrn().equals(node.getComponentId())).findFirst();

            if (currentNodeInfo.isPresent()) {
                nodeList.getSelectionModel().select(currentNodeInfo.get());
            }

            resultMapDialog.setNodes(result);

        });

        service.setOnFailed(event -> {
            ExceptionDialog errorAlert = new ExceptionDialog(service.getException());
            errorAlert.initOwner(mainApp.getScene().getWindow());
            errorAlert.setContentText(service.getException().getMessage());
            errorAlert.setHeaderText("Could not execute query!");
            errorAlert.setTitle("Query error");
            errorAlert.showAndWait();
        });

    }


    public boolean isAdvancedQueryValid() {
        try {
            QueryFactory.create(SPARQL_PREFIXES_DEFENITIONS + NODE_QUERY_START + advancedQueryTextArea.getText() + "}", Syntax.syntaxSPARQL);
            return true;
        } catch (QueryException ex) {
            return false;
        }

    }

    public class FilterCriteriaMenuItem extends CheckMenuItem {
        private final FilterCriteriaView filterCriteriaView;

        public FilterCriteriaMenuItem(String text, FilterCriteriaView filterCriteriaView) {
            super(text);
            this.filterCriteriaView = filterCriteriaView;

            setOnAction(event -> {
                if (isSelected()) {
                    addFilterCriteriaView(filterCriteriaView);
                } else {
                    removeFilterCriteriaView(filterCriteriaView);
                }
            });
        }

        public FilterCriteriaMenuItem(String text, Node graphic, FilterCriteriaView filterCriteriaView) {
            this(text, filterCriteriaView);
            setGraphic(graphic);
            filterCriteriaView.setGraphic(graphic);
        }
    }


    public void setMainApp(OntologyAdaptivePropertiesDialog mainApp) {
        this.mainApp = mainApp;

//        crit = mainApp.getCriteria();
//        reversedCrit = mainApp.getReversedCriteria();

        {
            FilterCriteriaMenuItem cpuFcmi =
                    new FilterCriteriaMenuItem("CPU Speed", GlyphUtils.createGlyph(FlatIcon.Glyph.CPU),
                            new RangeFilterView("CPU Speed", "Mhz") {
                                @Override
                                public String createFilter() {
                                    List<String> filters = new ArrayList<>();

                                    if (min.get().length() > 0) {
                                        filters.add("?cpu_value1 >= '" + min.get() + "'^^<http://www.w3.org/2001/XMLSchema#float>");
                                    }

                                    if (max.get().length() > 0) {
                                        filters.add("?cpu_value1 <= '" + max.get() + "'^^<http://www.w3.org/2001/XMLSchema#float>");
                                    }

                                    return "?node omn:hasComponent ?cpu.\n" +
                                            "?cpu rdf:type omncomp:CPU.  \n" +
                                            "?cpu dbpedia:fastest ?cpu_value1.\n" +
                                            "FILTER(" + filters.stream().collect(Collectors.joining(" && ")) + ").";
                                }
                            });
            mnuCriteria.getItems().add(cpuFcmi);
        }
        {
            FilterCriteriaMenuItem diskSpaceFcmi =
                    new FilterCriteriaMenuItem("Disk Space", GlyphUtils.createGlyph(FlatIcon.Glyph.HARDDRIVE),
                            new RangeFilterView("Disk Space", "Gb") {
                                @Override
                                public String createFilter() {
                                    List<String> filters = new ArrayList<>();

                                    if (min.get().length() > 0) {
                                        filters.add("?storage_value1 >= '" + min.get() + "'^^<http://www.w3.org/2001/XMLSchema#float>");
                                    }

                                    if (max.get().length() > 0) {
                                        filters.add("?storage_value1 <= '" + max.get() + "'^^<http://www.w3.org/2001/XMLSchema#float>");
                                    }

                                    return "?node omn:hasComponent  ?storage.\n" +
                                            "?storage rdf:type omncomp:StorageComponent.\n" +
                                            "?storage dbpedia:storage ?storage_value1." +
                                            "FILTER(" + filters.stream().collect(Collectors.joining(" && ")) + ").";
                                }
                            });
            mnuCriteria.getItems().add(diskSpaceFcmi);
        }
        {
            FilterCriteriaMenuItem ramFcmi =
                    new FilterCriteriaMenuItem("RAM", GlyphUtils.createGlyph(FlatIcon.Glyph.RAM),
                            new RangeFilterView("RAM", "Mb") {
                                @Override
                                public String createFilter() {
                                    List<String> filters = new ArrayList<>();

                                    if (min.get().length() > 0) {
                                        filters.add("?ram_value1 >= '" + min.get() + "'^^<http://www.w3.org/2001/XMLSchema#float>");
                                    }

                                    if (max.get().length() > 0) {
                                        filters.add("?ram_value1 <= '" + max.get() + "'^^<http://www.w3.org/2001/XMLSchema#float>");
                                    }

                                    return "?node omn:hasComponent  ?mem.\n" +
                                            "?mem rdf:type omncomp:MemoryComponent.\n" +
                                            "?mem dbpedia:memory ?ram_value1." +
                                            "FILTER(" + filters.stream().collect(Collectors.joining(" && ")) + ").";
                                }
                            });
            mnuCriteria.getItems().add(ramFcmi);
        }

        {
            FilterCriteriaMenuItem componentNameFcmi =
                    new FilterCriteriaMenuItem("Name", GlyphUtils.createGlyph(FontAwesome.Glyph.COMMENT_ALT),
                            new ValueFilterView("Component Name", "Name") {
                                @Override
                                public String createFilter() {
                                    return "?node omnlc:hasComponentName \"" + value.get() + "\"";
                                }
                            });
            mnuCriteria.getItems().add(componentNameFcmi);
        }

        {
            FilterCriteriaMenuItem locationFcmi =
                    new FilterCriteriaMenuItem("Location", GlyphUtils.createGlyph(FontAwesome.Glyph.MAP_MARKER),
                            new MapFilterView() {
                                @Override
                                public String createFilter() {
                                    List<String> filters = new ArrayList<>();

                                    if (minLat.get() != null)
                                        filters.add("xsd:float(?lat) >= " + minLat.get());
                                    if (maxLat.get() != null)
                                        filters.add("xsd:float(?lat) <= " + maxLat.get());
                                    if (minLon.get() != null)
                                        filters.add("xsd:float(?lon) >= " + minLon.get());
                                    if (maxLon.get() != null)
                                        filters.add("xsd:float(?lon) <= " + maxLon.get());

                                    return "FILTER(" + filters.stream().collect(Collectors.joining(" && ")) + ").";

                                }
                            });
            mnuCriteria.getItems().add(locationFcmi);
        }


//
//        for (Map.Entry<String, List<String>> entry : crit.entrySet()) {
//            String key = entry.getKey();
//            CheckMenuItem cmi = new CheckMenuItem(key);
//            cmi.setOnAction(new EventHandler<ActionEvent>() {
//                @Override
//                public void handle(ActionEvent event) {
//                    if (event.getSource() instanceof CheckMenuItem) {
//                        CheckMenuItem cmi1 = (CheckMenuItem) event.getSource();
//                        if (cmi1.isSelected()) {
//                            addCustomControl(cmi1.getText());
//                        } else {
//                            int index = findCustomControl(cmi1.getText());
//                            if (index >= 0) {
//                                flowPane.getChildren().remove(index);
//                                queryFilters.remove(index);
//                                if ("Longitude".equals(cmi1.getText()) || "Latitude".equals(cmi1.getText())) {
//                                    NodePropertyOverview.this.mainApp.getWebMap().clearRectangleDrawing();
//                                    index = findCustomControl("Longitude");
//                                    if (index >= 0) {
//                                        CustomControlView view = (CustomControlView) flowPane.getChildren().get(index);
//                                        view.performSave();
//                                    }
//                                    index = findCustomControl("Latitude");
//                                    if (index >= 0) {
//                                        CustomControlView view = (CustomControlView) flowPane.getChildren().get(index);
//                                        view.performSave();
//                                    }
//                                }
//                            }
//                        }
//                    }
//                }
//            });
//            mnuCriteria.getItems().add(cmi);
//        }
//        mnuCriteria.showingProperty().addListener((ov, t, t1) -> {
//            if (mnuCriteria.isShowing()) {
//                CustomControlView ccv = (CustomControlView) toggleGroup.getSelectedToggle();
//                if (ccv != null) {
//                    ccv.performSave();
//                }
//            }
//        });
//
//        nodeList.showingProperty().addListener((ov, t, t1) -> {
//            if (nodeList.isShowing()) {
//                CustomControlView ccv = (CustomControlView) toggleGroup.getSelectedToggle();
//                if (ccv != null) {
//                    ccv.performSave();
//                }
//            }
//        });

        searchButton.setOnAction(event -> {

            boolean error = false;

            if (!advanced.get()) {
                final String query = getBasicQueryString();
                if (query != null)
                    service.setNodeRestrictions(query);
                else
                    error = true;
            } else {
                service.setNodeRestrictions(advancedQueryTextArea.getText());
            }

            if (error)
                return;

            service.restart();

        });

        switchToAdvancedButton.setOnAction(event -> {
            String query = getBasicQueryString();
            if (query != null) {
                advancedQueryTextArea.setText(query);
                advanced.set(true);
            }
        });

        switchToBasicButton.setOnAction(event -> {
            Optional<ButtonType> result =
                    new Alert(WARNING, "Changes made will be lost! Are you sure you want to continue?", ButtonType.YES, ButtonType.NO, ButtonType.CANCEL).showAndWait();

            if (result.filter(buttonType -> buttonType == ButtonType.YES).isPresent()) {
                advanced.set(false);
            }


        });


        mapButton.setOnAction(event -> {
            if (mapButton.isSelected()) {
                resultMapDialog.showDialog();
            } else {
                resultMapDialog.hideDialog();
            }
        });

        cancelButton.setOnAction(event -> {
            Stage stage = (Stage) cancelButton.getScene().getWindow();
            stage.close();
        });

        saveButton.setOnAction(event -> {

            if (!isValidNodeName(nameTextField.getText())) {
                final Alert alert =
                        new Alert(ERROR, "A node name may only contain alphanumerical characters, and is maximum " + getMaximumNodeNameLength() + " characters long.");
                alert.setTitle("The entered node name is invalid.");
                alert.showAndWait();
                return;
            }
            if (nodeList.getSelectionModel().isEmpty()) {
                Alert alert = new Alert(ERROR, "Please select a node from the list");
                alert.setTitle("No node selected");
                alert.showAndWait();

                return;
            }

            node.setClientId(nameTextField.getText());
            final SparqlService.NodeInfo selectedNodeInfo = nodeList.getSelectionModel().getSelectedItem();
            node.setComponentId(selectedNodeInfo.getComponentUrn());
            node.setComponentManagerId(selectedNodeInfo.getComponentManagerUrn());

            //now close the window
            Stage stage = (Stage) saveButton.getScene().getWindow();
            stage.close();

        });

        queryFilters = new ArrayList<>();
        toggleGroup = new ToggleGroup();

        // Enter the query loop. A user is expected to enter class
        // expression on the command line.
    }

    private String getBasicQueryString() {
        StringBuilder queryBuilder = new StringBuilder();
        for (QueryFilter queryFilter : queryFilters) {
            if (!queryFilter.isValid()) {
                new Alert(ERROR, queryFilter.getName() + "does not have a valid value!").showAndWait();
                return null;
            } else {
                queryBuilder.append(queryFilter.createFilter());
                queryBuilder.append("\n");
            }
        }

        return queryBuilder.toString();
    }


    public void addFilterCriteriaView(FilterCriteriaView view) {
        queryFilters.add(view);
        flowPane.getChildren().add(view);
    }

    public void removeFilterCriteriaView(FilterCriteriaView view) {
        queryFilters.remove(view);
        flowPane.getChildren().remove(view);
    }


//    public CustomControlController addCustomControl(String title) {
//        CustomControlView view = new CustomControlView(title, crit.get(title).get(1), crit.get(title).get(2));
//        CustomControl model = new CustomControl();
//        model.setType(crit.get(title).get(1));
//        model.setName(crit.get(title).get(0));
//        CustomControlController controller = new CustomControlController(view, model);
//        controller.setTestbedInfo(mainApp.getReversedTestbedInfo());
//        controller.setFakeToRealUrn(mainApp.getFakeToRealUrn());
//        String query = "";
//        switch (crit.get(title).get(0)) {
//            case "cpu":
//                query = "ProcessorComponent";
//                break;
//            case "os":
//                query = "OS";
//                break;
//            case "sliverType":
//                query = "SliverType";
//                break;
//            case "testbed":
//                query = "Testbed";
//                break;
//        }
//        /*if (!"".equals(query)) {
//            List<String> values = dlQueryPrinter.askQuery(query.trim());
//            if ("Testbed".equals(query)) {
//                List<String> testbeds = new ArrayList<>();
//                for (String value : values) {
//                    if (mainApp.getTestbedInfo().get(value) != null) {
//                        testbeds.add(mainApp.getTestbedInfo().get(value));
//                    }
//                }
//                values = testbeds;
//            }
//            Collections.sort(values);
//            controller.populateListValues(values);
//        }   */
//        queryFilters.add(controller);
//        flowPane.getChildren().add(view);
//        view.setOwner(mainApp.getScene().getWindow());
//        view.setToggleGroup(toggleGroup);
//        view.addSaveHandler(new EventHandler<ActionEvent>() {
//            @Override
//            public void handle(ActionEvent event) {
//                CustomControlView ccv = (CustomControlView) toggleGroup.getSelectedToggle();
//                if (ccv != null) {
//                    ccv.setValue(ccv.getValueField().getText());
//                    ccv.setMin(ccv.getMinField().getText());
//                    ccv.setMax(ccv.getMaxField().getText());
//                    ccv.setSelected(false);
//                    ccv.hideStage();
//                }
//                int index = findCustomControl("Longitude");
//                double lat1, lat2, lon1, lon2;
//                if (index >= 0) {
//                    try {
//                        mainApp.getWebMap().clearRectangleDrawing();
//                        CustomControlView view1 = (CustomControlView) getFlowPane().getChildren().get(index);
//                        try {
//                            lon1 = Double.parseDouble(view1.getMinField().getText());
//                        } catch (Exception e) {
//                            lon1 = -180;
//                        }
//                        try {
//                            lon2 = Double.parseDouble(view1.getMaxField().getText());
//                        } catch (Exception e) {
//                            lon2 = 180;
//                        }
//
//                        index = findCustomControl("Latitude");
//                        if (index >= 0) {
//                            view1 = (CustomControlView) getFlowPane().getChildren().get(index);
//                            try {
//                                lat1 = Double.parseDouble(view1.getMinField().getText());
//                            } catch (Exception e) {
//                                lat1 = -90;
//                            }
//                            try {
//                                lat2 = Double.parseDouble(view1.getMaxField().getText());
//                            } catch (Exception e) {
//                                lat2 = 90;
//                            }
//                            mainApp.getWebMap().addRectangle(lat1, lat2, lon1, lon2);
//                        } else {
//                            mainApp.getWebMap().addRectangle(-90, 90, lon1, lon2);
//                        }
//                    } catch (Exception e) {
//                    }
//                } else {
//                    mainApp.getWebMap().clearRectangleDrawing();
//                    index = findCustomControl("Latitude");
//                    if (index >= 0) {
//                        try {
//                            CustomControlView view1 = (CustomControlView) getFlowPane().getChildren().get(index);
//                            try {
//                                lat1 = Double.parseDouble(view1.getMinField().getText());
//                            } catch (Exception e) {
//                                lat1 = -90;
//                            }
//                            try {
//                                lat2 = Double.parseDouble(view1.getMaxField().getText());
//                            } catch (Exception e) {
//                                lat2 = 90;
//                            }
//                            mainApp.getWebMap().addRectangle(lat1, lat2, -180, 180);
//                        } catch (Exception e) {
//
//                        }
//                    }
//                }
//            }
//        });
//        int i = getIndexMenuItem(title);
//        if (i >= 0) {
//            ((CheckMenuItem) mnuCriteria.getItems().get(i)).setSelected(true);
//        }
//        return controller;
//    }
//
//    EventHandler sceneEh = new EventHandler<ActionEvent>() {
//        @Override
//        public void handle(ActionEvent event) {
//            CustomControlView ccv = (CustomControlView) toggleGroup.getSelectedToggle();
//            if (ccv != null) {
//                ccv.performSave();
//            }
//        }
//
//    };

    public FlowPane getFlowPane() {
        return flowPane;

    }

    public ToggleGroup getToggleGroup() {
        return toggleGroup;
    }


    private void switchToBasic(String query) {/*
        query = query.replaceAll(" and ", "\n");
        query = query.replaceAll("\\(", "");
        query = query.replaceAll("\\)", "");
        if (!"".equals(query)) {
            queryFilters.clear();
            for (MenuItem cmi : mnuCriteria.getItems()) {
                ((CheckMenuItem) cmi).setSelected(false);
            }
            String crits = "";
            for (String s : reversedCrit.keySet()) {
                crits = crits.concat(s + "|");
            }
            crits = crits.substring(0, crits.length() - 1);
            Matcher matcher = Pattern.compile("(" + crits + ") " + "(value|some) (.*)").matcher(query);
            while (matcher.find()) {
                CustomControlController controller = addCustomControl(reversedCrit.get(matcher.group(1)));
                if ("value".equals(matcher.group(2))) {
                    Matcher valueMatcher = Pattern.compile("\"(.*)\"\\^\\^xsd:(.*)").matcher(matcher.group(3));
                    if (valueMatcher.find()) {
                        String value = valueMatcher.group(1);
                        String type = valueMatcher.group(2);
                        if (!controller.getModel().getType().equals(type)) {
                            //todo: show error about wrong type
                        } else {
                            controller.getView().getValueField().setText(value);
                        }
                    } else {
                        String value = matcher.group(3);
                        value = value.replaceAll("\\+", "-");
                        value = value.replaceAll(":", "-");
                        value = value.replaceAll("ibbtple-", "");
                        if (mainApp.getTestbedInfo().get(value) != null) {
                            value = mainApp.getTestbedInfo().get(value);
                        }
                        controller.getView().getValueField().setText(value);
                        controller.getView().getListValues().getSelectionModel().select(value);
                    }
                } else if ("some".equals(matcher.group(2))) {
                    Matcher valueMatcher = Pattern.compile("(.*)\\[(.*)\\]").matcher(matcher.group(3));
                    if (valueMatcher.find()) {
                        String type = valueMatcher.group(1);
                        if (!controller.getModel().getType().equals(type)) {
                            //todo: show error about wrong type
                        } else {
                            String value = valueMatcher.group(2);
                            value = value.replace(",", " ");
                            Matcher rangeMatcher = Pattern.compile(">\\s+([^\\s]*)").matcher(value);
                            if (rangeMatcher.find()) {
                                double d = Double.parseDouble(rangeMatcher.group(1));
                                d++;
                                controller.getView().getMinField().setText(String.valueOf(d));
                            }
                            rangeMatcher = Pattern.compile(">=\\s+([^\\s]*)").matcher(value);
                            if (rangeMatcher.find()) {
                                double d = Double.parseDouble(rangeMatcher.group(1));
                                controller.getView().getMinField().setText(String.valueOf(d));
                            }
                            rangeMatcher = Pattern.compile("<\\s+([^\\s]*)").matcher(value);
                            if (rangeMatcher.find()) {
                                double d = Double.parseDouble(rangeMatcher.group(1));
                                d--;
                                controller.getView().getMaxField().setText(String.valueOf(d));
                            }
                            rangeMatcher = Pattern.compile("<=\\s+([^\\s]*)").matcher(value);
                            if (rangeMatcher.find()) {
                                double d = Double.parseDouble(rangeMatcher.group(1));
                                controller.getView().getMaxField().setText(String.valueOf(d));
                            }
                        }
                    }
                }
                controller.getView().saveSettings();
            }
        } else {
            for (CustomControlController cc : queryFilters) {
                flowPane.getChildren().add(cc.getView());
            }
        }     */
    }

    public int getIndexMenuItem(String s) {
        int i = 0;
        for (MenuItem mi : mnuCriteria.getItems()) {
            if (s.equals(mi.getText())) {
                return i;
            }
            i++;
        }
        return -1;
    }

    public void unselectMenuItem(int index) {
        CheckMenuItem cmi = (CheckMenuItem) mnuCriteria.getItems().get(index);
        cmi.setSelected(false);
    }

    private boolean isValidNodeName(String text) {
        if (text.length() > getMaximumNodeNameLength()) {
            return false;
        }

        for (char c : text.toCharArray()) {
            boolean validChar = (c >= 'a' && c <= 'z')
                    || (c >= 'A' && c <= 'Z')
                    || (c >= '0' && c <= '9');

            if (!validChar) {
                return false;
            }
        }

        return true;
    }


    public void performHideMap() {
        mapButton.fire();
    }

    public void setNameTextFieldText(String s) {
        nameTextField.setText(s);
    }

    private int getMaximumNodeNameLength() {
        return 10;
    }

    public ComboBox getNodeList() {
        return nodeList;
    }

    public TextField getNameTextField() {
        return nameTextField;
    }

    public Button getSaveButton() {
        return saveButton;
    }

    public Button getCancelButton() {
        return cancelButton;
    }

    public List<QueryFilter> getQueryFilters() {
        return queryFilters;
    }

    public boolean isAdvanced() {
        return advanced.get();
    }

    public void setAdvancedQuery(String s) {
        String query = advancedQueryTextArea.getText();
        query = query.concat(s);
        advancedQueryTextArea.setText(query);
    }

    public void replaceAdvancedQuery(String s) {
        advancedQueryTextArea.setText(s);
    }

    public String getAdvancedQuery() {
        return advancedQueryTextArea.getText();
    }

    public RspecNode getNode() {
        return node;
    }

    public void setNode(RspecNode node) {
        this.node = node;
    }


    public void setModel(ModelRspec modelRspec) {
        this.modelRspec = modelRspec;
    }
}
