package be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.model.query;

import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.concurrent.Service;
import javafx.concurrent.Task;
import org.apache.jena.query.*;
import org.apache.jena.rdf.model.Literal;
import org.apache.jena.sparql.engine.http.QueryEngineHTTP;

import java.util.HashSet;
import java.util.Set;

/**
 * User: twalcari
 * Date: 4/5/2016
 * Time: 12:07
 */
public class SparqlService extends Service<Set<SparqlService.NodeInfo>> {

    public static final String SPARQL_ENDPOINT = "http://lod.fed4fire.eu/sparql";

    public static final String SPARQL_PREFIXES_DEFENITIONS =
            "PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>" +
                    "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>\n" +
                    "PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>\n" +
                    "PREFIX omn: <http://open-multinet.info/ontology/omn#>\n" +
                    "PREFIX omnfed: <http://open-multinet.info/ontology/omn-federation#>\n" +
                    "PREFIX omnres: <http://open-multinet.info/ontology/omn-resource#>\n" +
                    "PREFIX omnlc: <http://open-multinet.info/ontology/omn-lifecycle#>\n" +
                    "PREFIX omnwireless: <http://open-multinet.info/ontology/omn-domain-wireless#>\n" +
                    "PREFIX omncomp: <http://open-multinet.info/ontology/omn-component#>\n" +
                    "PREFIX dbpedia: <http://dbpedia.org/property/>\n" +
                    "\n";

    public static final String NODE_QUERY_START = "SELECT DISTINCT ?node ?cm ?lat ?lon WHERE {\n" +
            "  ?node omnlc:managedBy ?cm.\n" +
            "  ?node rdf:type omnres:Node.\n" +
            "OPTIONAL {  ?node <http://www.w3.org/2003/01/geo/wgs84_pos#long> ?lon.\n" +
            "  ?node <http://www.w3.org/2003/01/geo/wgs84_pos#lat> ?lat. }.";

    public static final String NODE_QUERY_END = "}";

    private final StringProperty nodeRestrictions = new SimpleStringProperty();

    public String getNodeRestrictions() {
        return nodeRestrictions.get();
    }

    public StringProperty nodeRestrictionsProperty() {
        return nodeRestrictions;
    }

    public void setNodeRestrictions(String nodeRestrictions) {
        this.nodeRestrictions.set(nodeRestrictions);
    }

    @Override
    protected Task<Set<NodeInfo>> createTask() {
        final String _nodeRestrictions = getNodeRestrictions();


        return new Task<Set<NodeInfo>>() {
            @Override
            protected Set<NodeInfo> call() throws Exception {
                Query query = createQuery(_nodeRestrictions);

                QueryEngineHTTP httpQuery = new QueryEngineHTTP(SPARQL_ENDPOINT, query);

                Set<NodeInfo> resultSet = new HashSet<>();

                // execute a Select query
                ResultSet results = httpQuery.execSelect();
                while (results.hasNext()) {
                    QuerySolution solution = results.next();

                    final Literal latLiteral = solution.getLiteral("lat");
                    final Literal lonLiteral = solution.getLiteral("lon");
                    resultSet.add(new NodeInfo(GeniUrn.parse(solution.getResource("cm").asResource().getURI()),
                            GeniUrn.parse(solution.getResource("node").asResource().getURI()),
                            latLiteral != null ? latLiteral.getDouble() : null,
                            lonLiteral != null ? lonLiteral.getDouble() : null));
                }

                return resultSet;
            }
        };
    }

    private Query createQuery(String _nodeRestrictions) throws QueryException {
        return QueryFactory.create(SPARQL_PREFIXES_DEFENITIONS + NODE_QUERY_START + _nodeRestrictions + "}", Syntax.syntaxSPARQL);
    }

    public boolean isValidQuery() {
        try {
            createQuery(nodeRestrictions.get());
            return true;
        } catch (QueryException ex) {
            return false;
        }

    }

    public static class NodeInfo {
        public final GeniUrn componentManagerUrn;
        public final GeniUrn componentUrn;

        private final Double latitude, longitude;

        public NodeInfo(GeniUrn componentManagerUrn, GeniUrn componentUrn, Double latitude, Double longitude) {
            this.componentManagerUrn = componentManagerUrn;
            this.componentUrn = componentUrn;
            this.latitude = latitude;
            this.longitude = longitude;
        }

        public GeniUrn getComponentManagerUrn() {
            return componentManagerUrn;
        }

        public GeniUrn getComponentUrn() {
            return componentUrn;
        }

        public Double getLatitude() {
            return latitude;
        }

        public Double getLongitude() {
            return longitude;
        }
    }
}
