package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.NodeDescription;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.bonfire.BonfireNodePropertiesTab;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManager;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManagerFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.ComponentManagerInfo;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.MapDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.health.MaintenanceMessage;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.TasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.AvailabilityMeter;
import be.iminds.ilabt.jfed.experimenter_gui.ui.HealthStatusPane;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.util.BonfireLicenseUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.SliverType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecLink;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import javafx.beans.property.*;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.geometry.Insets;
import javafx.scene.control.*;
import javafx.scene.layout.*;
import javafx.scene.paint.Color;
import javafx.stage.Stage;
import org.controlsfx.glyphfont.FontAwesome;
import org.controlsfx.glyphfont.Glyph;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.net.MalformedURLException;
import java.net.URL;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 4/10/2014
 * Time: 11:55
 */
public class AdaptivePropertiesDialog extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(AdaptivePropertiesDialog.class);

    private static final Comparator<? super MaintenanceMessage> MAINTENANCE_MESSAGE_COMPARATOR =
            Comparator.nullsLast(Comparator.comparing(MaintenanceMessage::getStartTime))
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(MaintenanceMessage::getEndTime)))
                    .thenComparing(MaintenanceMessage::getMessage);

    private final JFedConfiguration config;
    private final AuthorityList authorityList;

    private final FXRspecNode node;
    private final ModelRspecEditor modelRspecEditor;

    private final List<AdaptivePropertiesTab> availableTabs;
    private final CMIManager cmiManager;


    @FXML
    private HealthStatusPane healthStatusPane;
    @FXML
    private VBox propertiesBox;

    @FXML
    private Pane diskImageBox;
    @FXML
    private Button addDiskImageButton;

    @FXML
    private TextField nameTextField;

    @FXML
    private Button saveButton;
    @FXML
    private Button cancelButton;
    @FXML
    private ComboBox<ComponentManagerInfo> componentManagerInfoComboBox;
    @FXML
    private Button testbedsMapButton;

    @FXML
    private Label testbedHealthInfoLabel;

    @FXML
    private TextField diskImageTextField;

    @FXML
    private TabPane tabPane;

    private final Map<ComponentManagerInfo, ComponentManagerPropertiesPane> cmiPropertiesPanes = new HashMap<>();
    private final ObjectProperty<ComponentManagerPropertiesPane> currentCmip = new SimpleObjectProperty<>(null);
    private final BooleanProperty isSpecificNodeChosen = new SimpleBooleanProperty();
    private final ReadOnlyObjectProperty<ComponentManagerInfo> currentCmi;
    private final CMIManagerFactory cmiManagerFactory;
    private final MapDialogFactory mapDialogFactory;
    private final HighLevelTaskFactory hltf;
    private final TasksFactory tasksFactory;
    private final TaskService taskService;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;

    private class CMICell extends ListCell<ComponentManagerInfo> {


        protected final HBox container = new HBox(2);

        protected final Label nameLabel = new Label("");
        protected final Glyph maintenanceIcon =
                GlyphUtils.createGlyph(FontAwesome.Glyph.WRENCH, Color.RED);
        protected final Glyph reputationGlyph = GlyphUtils.createGlyph(FontAwesome.Glyph.STAR, Color.GRAY);
        protected final Glyph healthStatusGlyph = GlyphUtils.createGlyph(FontAwesome.Glyph.HEART, Color.GRAY);

        private final boolean shouldUseVMCalculationMethod;
        @Nullable
        protected final AvailabilityMeter availabilityMeter;

        {
            Pane fillerPane = new Pane();
            HBox.setHgrow(fillerPane, Priority.ALWAYS);

            container.setPadding(new Insets(0, 5, 2, 0));
            container.getChildren().addAll(nameLabel, fillerPane, maintenanceIcon, reputationGlyph, healthStatusGlyph);

            boolean shouldUseVMCalculationMethod = false;
            if (node.getNodeDescriptionId() != null) {
                final NodeDescription nodeDescription =
                        AdaptivePropertiesDialog.this.config.getNodeDescription(node.getNodeDescriptionId());
                if (nodeDescription != null) {
                    if (NodeDescription.NodeType.RAW_PC.equals(nodeDescription.getType())) {
                        availabilityMeter = new AvailabilityMeter();
                        container.getChildren().add(availabilityMeter);
                    } else if (NodeDescription.NodeType.VM.equals(nodeDescription.getType())) {
                        availabilityMeter = new AvailabilityMeter();
                        shouldUseVMCalculationMethod = true;
                        availabilityMeter.setUseVmCalculationMethod(true);
                        container.getChildren().add(availabilityMeter);
                    } else {
                        availabilityMeter = null;
                    }
                } else {
                    availabilityMeter = null;
                }

            } else {
                availabilityMeter = null;
            }

            setGraphic(container);

            this.shouldUseVMCalculationMethod = shouldUseVMCalculationMethod;
        }


        @Override
        protected void updateItem(ComponentManagerInfo item, boolean empty) {
            super.updateItem(item, empty);
            if (!empty) {
                nameLabel.setText(StringConverters.CMI_STRING_CONVERTER.toString(item));

                final JFedConfiguration.TestbedDescription testbedDescription =
                        AdaptivePropertiesDialog.this.config.getTestbedDescription(item.getAuthority().getUrn());


                if (testbedDescription != null && testbedDescription.getHealthStatus() != null) {
                    if (testbedDescription.getHealthStatus().getHealth() != null) {
                        final int health = testbedDescription.getHealthStatus().getHealth();

                        if (health == 100) {
                            healthStatusGlyph.setTextFill(Color.GREEN);
                        } else if (health > 66) {
                            healthStatusGlyph.setTextFill(Color.YELLOWGREEN);
                        } else if (health > 33) {
                            healthStatusGlyph.setTextFill(Color.ORANGE);
                        } else {
                            healthStatusGlyph.setTextFill(Color.RED);
                        }
                        Tooltip.install(healthStatusGlyph, new Tooltip(String.format("%s has a health rating of %d%%",
                                StringConverters.CMI_STRING_CONVERTER.toString(item),
                                health)));

                        healthStatusGlyph.setVisible(true);
                    } else {
                        healthStatusGlyph.setVisible(false);
                    }

                    if (testbedDescription.getHealthStatus().getReputation() != null) {
                        final int reputation = testbedDescription.getHealthStatus().getReputation();

                        if (reputation > 95) {
                            reputationGlyph.setTextFill(Color.GREEN);
                        } else if (reputation > 66) {
                            reputationGlyph.setTextFill(Color.YELLOWGREEN);
                        } else if (reputation > 33) {
                            reputationGlyph.setTextFill(Color.ORANGE);
                        } else {
                            reputationGlyph.setTextFill(Color.RED);
                        }
                        Tooltip.install(reputationGlyph, new Tooltip(String.format("%s has a reputation rating of %d%%",
                                StringConverters.CMI_STRING_CONVERTER.toString(item),
                                reputation)));


                        reputationGlyph.setVisible(true);
                    } else {
                        reputationGlyph.setVisible(false);
                    }

                    if (testbedDescription.getHealthStatus().getMaintenance() != null) {
                        List<MaintenanceMessage> activeMaintenances =
                                testbedDescription.getHealthStatus().getMaintenance().stream()
                                        .filter(mm -> mm.getStartTime() == null || mm.getStartTime().isBefore(ZonedDateTime.now()))
                                        .filter(mm -> mm.getEndTime() == null || mm.getEndTime().isAfter(ZonedDateTime.now()))
                                        .sorted(MAINTENANCE_MESSAGE_COMPARATOR)
                                        .collect(Collectors.toList());

                        List<MaintenanceMessage> futureMaintenances =
                                testbedDescription.getHealthStatus().getMaintenance().stream()
                                        .filter(mm -> mm.getStartTime() != null && mm.getStartTime().isAfter(ZonedDateTime.now()))
                                        .sorted(MAINTENANCE_MESSAGE_COMPARATOR)
                                        .collect(Collectors.toList());

                        if (!activeMaintenances.isEmpty()) {
                            Tooltip.install(maintenanceIcon, new Tooltip(MaintenanceMessage.formatMaintenanceMessages(activeMaintenances)));
                            maintenanceIcon.setTextFill(Color.RED);
                            maintenanceIcon.setVisible(true);
                        } else if(!futureMaintenances.isEmpty()) {
                            Tooltip.install(maintenanceIcon, new Tooltip(MaintenanceMessage.formatMaintenanceMessages(futureMaintenances)));
                            maintenanceIcon.setTextFill(Color.ORANGE);
                            maintenanceIcon.setVisible(true);
                        }else{
                            maintenanceIcon.setVisible(false);
                        }
                    } else {
                        maintenanceIcon.setVisible(false);
                    }

                    if (availabilityMeter != null) {
                        if (shouldUseVMCalculationMethod
                                && testbedDescription.getHealthStatus().getVMsAvailable() != null
                                && testbedDescription.getHealthStatus().getVMsTotal() != null
                                && testbedDescription.getHealthStatus().getVMsTotal() != 0) {
                            availabilityMeter.setUseVmCalculationMethod(true);

                            availabilityMeter.setAvailable(testbedDescription.getHealthStatus().getVMsAvailable());
                            availabilityMeter.setTotal(testbedDescription.getHealthStatus().getVMsTotal());
                            availabilityMeter.setVisible(true);
                            availabilityMeter.setUndetermined(false);

                            Tooltip.install(availabilityMeter,
                                    new Tooltip(testbedDescription.getHealthStatus().getVMsAvailable() + " of "
                                            + testbedDescription.getHealthStatus().getVMsTotal() + " VM available"));
                        } else {
                            availabilityMeter.setUseVmCalculationMethod(false);
                            if (testbedDescription.getHealthStatus().getRawPCsAvailable() != null
                                    && testbedDescription.getHealthStatus().getRawPCsTotal() != null
                                    && testbedDescription.getHealthStatus().getRawPCsTotal() != 0) {
                                availabilityMeter.setAvailable(testbedDescription.getHealthStatus().getRawPCsAvailable());
                                availabilityMeter.setTotal(testbedDescription.getHealthStatus().getRawPCsTotal());
                                availabilityMeter.setVisible(true);
                                availabilityMeter.setUndetermined(false);

                                Tooltip.install(availabilityMeter,
                                        new Tooltip(testbedDescription.getHealthStatus().getRawPCsAvailable() + " of "
                                                + testbedDescription.getHealthStatus().getRawPCsTotal() + " bare metal available"));

                                availabilityMeter.setVisible(true);
                            }else if(testbedDescription.getHealthStatus().getAnyAvailable() != null
                                    && testbedDescription.getHealthStatus().getAnyTotal() != null
                                    && testbedDescription.getHealthStatus().getAnyTotal() != 0) {
                                availabilityMeter.setAvailable(testbedDescription.getHealthStatus().getAnyAvailable());
                                availabilityMeter.setTotal(testbedDescription.getHealthStatus().getAnyTotal());
                                availabilityMeter.setVisible(true);
                                availabilityMeter.setUndetermined(false);

                                Tooltip.install(availabilityMeter,
                                        new Tooltip(testbedDescription.getHealthStatus().getAnyAvailable() + " of "
                                                + testbedDescription.getHealthStatus().getAnyTotal() + " resources available"));
                            }    else{
                                availabilityMeter.setAvailable(0);
                                availabilityMeter.setTotal(0);
                                availabilityMeter.setVisible(true);
                                availabilityMeter.setUndetermined(true);
                            }
                        }
                    }


                } else {
                    healthStatusGlyph.setVisible(false);
                     reputationGlyph.setVisible(false);
                    maintenanceIcon.setVisible(false);
                    if (availabilityMeter != null)
                        availabilityMeter.setVisible(false);
                }


            } else {
                nameLabel.setText("");
                healthStatusGlyph.setVisible(false);
                reputationGlyph.setVisible(false);
                maintenanceIcon.setVisible(false);
                if (availabilityMeter != null)
                    availabilityMeter.setVisible(false);
            }
        }

    }

    AdaptivePropertiesDialog(ModelRspecEditor modelRspecEditor, final FXRspecNode node, JFedConfiguration config,
                             AuthorityList authorityList, CMIManagerFactory cmiManagerFactory, MapDialogFactory mapDialogFactory,
                             HighLevelTaskFactory hltf, TasksFactory tasksFactory, TaskService taskService,
                             AggregateManagerWrapperFactory aggregateManagerWrapperFactory, BonfireLicenseUtil bonfireLicenseUtil) {
        this.config = config;
        this.authorityList = authorityList;
        this.cmiManagerFactory = cmiManagerFactory;
        this.mapDialogFactory = mapDialogFactory;
        this.hltf = hltf;
        this.tasksFactory = tasksFactory;
        this.taskService = taskService;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        assert (modelRspecEditor != null);
        assert (node != null);

        this.modelRspecEditor = modelRspecEditor;
        this.node = node;

        this.cmiManager = cmiManagerFactory.getCMIManager(node);

        //initialize the Scene
        FXMLUtil.injectFXML(this);

        //initialize fields of main tab
        nameTextField.setTooltip(
                new Tooltip("Only alphanumerical, maxLat. " + ModelRspecEditor.MAX_NODE_NAME_LENGTH + " characters"));


        //set initial values to these of the received node
        nameTextField.setText(node.getClientId());

        if (node.getNodeDescriptionId() != null) {
            NodeDescription nodeDescription = config.getNodeDescription(node.getNodeDescriptionId());
            if (nodeDescription != null)
                healthStatusPane.setType(nodeDescription.getType());
        }

        //populate tabPane
        availableTabs = new ArrayList<>();

        BootScriptsTab bootScriptsTab = new BootScriptsTab(this);
        bootScriptsTab.load(node);
        availableTabs.add(bootScriptsTab);

        ExoGeniConfigTab exoGeniConfigTab = new ExoGeniConfigTab(this);
        exoGeniConfigTab.load(node);
        availableTabs.add(exoGeniConfigTab);

        RoutableIPTab routableIPTab = new RoutableIPTab(this);
        routableIPTab.load(node);
        availableTabs.add(routableIPTab);

        XenTab xenTab = new XenTab(this);
        xenTab.load(node);
        availableTabs.add(xenTab);

        //add tab for each interface. Can only be done after nameTextField has been properly initialized!
        for (FXRspecInterface iface : node.getInterfaces()) {
            tabPane.getTabs().add(new InterfacePropertiesTab(modelRspecEditor, node, iface, nameTextField.textProperty()));
        }


        componentManagerInfoComboBox.setItems(cmiManager.getComponentManagerInfos());
        componentManagerInfoComboBox.setConverter(StringConverters.CMI_STRING_CONVERTER);
        componentManagerInfoComboBox.setCellFactory(param -> new CMICell());
        componentManagerInfoComboBox.setButtonCell(new CMICell() {

            {
                nameLabel.setTextFill(Color.BLACK);
                container.setPadding(new Insets(0, 0, 0, 0));
            }

        });

        this.currentCmip.addListener(new ChangeListener<ComponentManagerPropertiesPane>() {
            @Override
            public void changed(ObservableValue<? extends ComponentManagerPropertiesPane> observable, ComponentManagerPropertiesPane oldValue, ComponentManagerPropertiesPane newValue) {
                isSpecificNodeChosen.unbind();
                if (newValue != null) {
                    isSpecificNodeChosen.bind(newValue.isSpecificNodeChosen());
                } else
                    isSpecificNodeChosen.setValue(false);
            }
        });


        this.currentCmi = componentManagerInfoComboBox.getSelectionModel().selectedItemProperty();

        this.currentCmi.addListener(new ChangeListener<ComponentManagerInfo>() {
            @Override
            public void changed(ObservableValue<? extends ComponentManagerInfo> observableValue,
                                ComponentManagerInfo oldValue, ComponentManagerInfo newValue) {

                boolean activeNodesOnly = true;

                if (oldValue != null) {
                    final JFedConfiguration.TestbedDescription testbedDescription =
                            AdaptivePropertiesDialog.this.config.getTestbedDescription(oldValue.getAuthority().getUrn());

                    if (testbedDescription != null) {

                        List<JFedConfiguration.SliverType> sliverTypes = testbedDescription.getSliverTypes();
                        if (!sliverTypes.isEmpty() &&
                                sliverTypes.get(0).getDiskImageName() != null &&
                                sliverTypes.get(0).getDiskImageName().equals(diskImageTextField.getText())) {
                            diskImageTextField.setText("");
                        }
                    } else {
                        LOG.warn("No testbed description available for " + oldValue.getAuthority().getUrn());
                    }
                    activeNodesOnly = cmiPropertiesPanes.get(oldValue).isActiveNodesOnly();
                }

                if (newValue != null) {
                    if (!cmiPropertiesPanes.containsKey(newValue)) {
                        cmiPropertiesPanes.put(currentCmi.get(), new ComponentManagerPropertiesPane(newValue, node,
                                activeNodesOnly, cmiManager, config, mapDialogFactory, bonfireLicenseUtil));
                    }
                    propertiesBox.getChildren().add(cmiPropertiesPanes.get(newValue));
                    currentCmip.set(cmiPropertiesPanes.get(newValue));


                    JFedConfiguration.TestbedDescription testbedDescription =
                            AdaptivePropertiesDialog.this.config.getTestbedDescription(newValue.getAuthority().getUrn());

                    if (testbedDescription != null) {
                        List<JFedConfiguration.SliverType> sliverTypes = testbedDescription.getSliverTypes();

                        //mobile12, zotacH1
                        if (!sliverTypes.isEmpty() &&
                                sliverTypes.get(0).getDiskImageName() != null &&
                                diskImageTextField.getText().isEmpty()) {
                            diskImageTextField.setText(sliverTypes.get(0).getDiskImageName());
                        }

                        healthStatusPane.setHealthStatus(testbedDescription.getHealthStatus());

                    } else {
                        LOG.warn("No testbed description available for " + newValue.getAuthority().getUrn());
                        healthStatusPane.setHealthStatus(null);
                    }
                } else
                    currentCmip.set(null);

                if (oldValue != null) {
                    propertiesBox.getChildren().remove(cmiPropertiesPanes.get(oldValue));
                }

                if (newValue != null) {
                    //only show disk image button for instageni and emulab-testbeds
                    SfaAuthority authority = newValue.getAuthority();
                    boolean showDiskImageDialog = authority.isType("instageni") || authority.isType("emulab")
                            || AdaptivePropertiesDialog.this.config.getSupportsListDiskImages(authority.getUrn());
                    addDiskImageButton.setVisible(showDiskImageDialog);
                }

                determineTabsVisibility(node);
            }
        });


        ComponentManagerInfo componentManagerInfo =
                node.getComponentManagerId() != null ?
                        cmiManager.getComponentManagerInfo(node.getComponentManagerId()) :
                        null;

        if (componentManagerInfo == null) {
            //no component manager: select nothing
            componentManagerInfoComboBox.getSelectionModel().clearSelection();
        } else {
            //select the current testbed in the list
            componentManagerInfoComboBox.getSelectionModel().select(componentManagerInfo);
        }

        if (!node.getSliverTypes().isEmpty()) {
            List<DiskImage> diskImages = node.getSliverTypes().get(0).getDiskImages();

            if (!diskImages.isEmpty())
                diskImageTextField.setText(diskImages.get(0).getName() != null ?
                        diskImages.get(0).getName() :
                        diskImages.get(0).getUrl());
        }

        diskImageBox.managedProperty().bind(diskImageBox.visibleProperty());
        addDiskImageButton.managedProperty().bind(addDiskImageButton.visibleProperty());

        //only show maps-button if running on a 64-bit JVM (see FEDIBBTDEV-1344)
        String arch = System.getProperty("os.arch");

        if (arch.equals("x86")) {
            testbedsMapButton.setVisible(false);
        }

        //do this when currentCmi is initialised and set to current CMI
        BonfireNodePropertiesTab bonfireNodePropertiesTab = new BonfireNodePropertiesTab(this);
        bonfireNodePropertiesTab.load(node);
        availableTabs.add(bonfireNodePropertiesTab);

        determineTabsVisibility(node);
    }

    @FXML
    private void showTestbedsMap() {
        ComponentManagerInfo selectedCmi =
                mapDialogFactory.showTestbedMapDialog(getScene().getWindow(), cmiManager);
        if (selectedCmi != null)
            componentManagerInfoComboBox.getSelectionModel().select(selectedCmi);
    }

    /**
     * This function shows and hides tabs as needed. It is called by evenListeners on the testbeds- and node-comboboxes.
     */
    private void determineTabsVisibility(FXRspecNode rspecNode) {
        int insertPosition = 1;
        for (AdaptivePropertiesTab apt : availableTabs) {
            if (apt.shouldBeVisible(rspecNode)) {
                LOG.debug("Tab {} should be made visible.", apt.getClass().getSimpleName());
                if (tabPane.getTabs().contains(apt)) {
                    insertPosition++;
                } else {
                    tabPane.getTabs().add(insertPosition, apt);
                }
            } else {
                LOG.debug("Tab {} should be made invisible.", apt.getClass().getSimpleName());
                if (tabPane.getTabs().contains(apt)) {
                    tabPane.getTabs().remove(apt);
                }
            }
        }
    }

    @FXML
    @SuppressWarnings("unchecked")
    public void onSaveAction(ActionEvent actionEvent) {
        nameTextField.setText(nameTextField.getText().trim());

        if (!ModelRspecEditor.isValidNodeName(nameTextField.getText())) {
            JFDialogs.create()
                    .owner(this)
                    .message("A node name may only contain alphanumerical characters, and is maximum " + ModelRspecEditor.MAX_NODE_NAME_LENGTH + " characters long.")
                    .masthead("The entered node name is invalid.")
                    .showError();
            return;
        }

        if (currentCmi.get() == null) {
            JFDialogs.create()
                    .owner(this)
                    .message("You must select an authority for this node.")
                    .masthead("No authority selected")
                    .title("No authority selected")
                    .showError();
            return;
        }


        if (!diskImageTextField.getText().isEmpty()) {
            //first remove any extra spaces that have slipped in
            diskImageTextField.setText(diskImageTextField.getText().trim().replace(" ", ""));

            String diskImageValue = diskImageTextField.getText();

            //now check if it is a valid URN or HTTP(S)-link

            if (diskImageValue.startsWith("urn")) {
                GeniUrn diskImageUrn = GeniUrn.parse(diskImageValue);

                if (diskImageUrn == null || !"image".equals(diskImageUrn.getEncodedResourceType())) {
                    JFDialogs.create()
                            .owner(this)
                            .message("The provided disk image URN is invalid. Please correct the entered value.")
                            .masthead("Invalid disk image URN")
                            .showError();

                    LOG.warn("User entered invalid disk image URN '{}'", diskImageValue);

                    return;
                }

            } else {
                //we expect a valid URL instead
                try {
                    new URL(diskImageValue);
                } catch (MalformedURLException e) {
                    JFDialogs.create()
                            .owner(this)
                            .message("The provided disk image URL is invalid. Please correct the entered value.")
                            .masthead("Invalid disk image URL")
                            .showError();


                    LOG.warn("User entered invalid disk image URL '{}'", diskImageValue);

                    return;
                }
            }
        }

        ComponentManagerPropertiesPane componentManagerPropertiesPane = cmiPropertiesPanes.get(currentCmi.get());

        if (!componentManagerPropertiesPane.validate())
            return;


        for (AdaptivePropertiesTab visibleTab : Iterables.filter(tabPane.getTabs(), AdaptivePropertiesTab.class)) {
            String validation = visibleTab.validateConfiguration();

            if (validation != null) {
                JFDialogs.create()
                        .owner(this)
                        .message(validation)
                        .masthead("Node configuration could not be saved")
                        .showError();
                return;
            }

        }


        for (PropertiesTab visibleTab : Iterables.filter(tabPane.getTabs(), PropertiesTab.class)) {
            String validation = visibleTab.validateConfiguration();

            if (validation != null) {
                JFDialogs.create()
                        .owner(this)
                        .message(validation)
                        .masthead("Node configuration could not be saved")
                        .showError();
                return;
            }

        }


        FXRspecNode currentNode = this.node;

        if (!nameTextField.getText().equals(currentNode.getClientId())) {
            modelRspecEditor.updateNodeClientId(currentNode, nameTextField.getText());
        }

        ComponentManagerInfo cmi = componentManagerInfoComboBox.getSelectionModel().getSelectedItem();

        boolean componentManagerChanged = !cmi.getAuthority().getUrn().equals(currentNode.getComponentManagerId());
        currentNode.setComponentManagerId(cmi.getAuthority().getUrn());

        //force SliverType to an existing type on the testbed
        //NOTE: this is a good default, that gets overwritten when the ExoGeni tab is present and is saved later in this step
        if ((currentNode.getSliverTypeName() == null || !cmi.hasSliverType(currentNode.getSliverTypeName()))
                && cmi.getDefaultSliverType() != null) {
            SliverType chosenNewType = cmi.getDefaultSliverType();
            if (chosenNewType != null) {

                if (!cmi.getSliverTypes().isEmpty() && !cmi.hasSliverType(chosenNewType.getName())) {
                    LOG.warn("Default sliver_type for " + cmi.getAuthority().getUrn() + " is \"" + chosenNewType.getName() + "\", but that sliver_type is not in ListResources. Picking an existing one instead.");
                    chosenNewType = null;
                } else {
                    LOG.warn("Default sliver_type for " + cmi.getAuthority().getUrn() + " is \"" + chosenNewType.getName() + "\", but that sliver_type is not in ListResources. However, no alternative is found in ListResources...");
                }
            }

            if (chosenNewType == null && !cmi.getSliverTypes().isEmpty())
                chosenNewType = cmi.getSliverTypes().get(0); //choose first (= quite randomly)

            if (chosenNewType != null)
                currentNode.setSliverTypeName(chosenNewType.getName());
        }

        currentNode.setComponentId(componentManagerPropertiesPane.getSelectedNode() != null &&
                componentManagerPropertiesPane.getSelectedNode().getComponentId() != null ?
                componentManagerPropertiesPane.getSelectedNode().getComponentId()
                : null);

        if (componentManagerPropertiesPane.getSelectedHardwareType() != null)
            currentNode.getHardwareTypes().setAll(componentManagerPropertiesPane.getSelectedHardwareType());


        if (componentManagerChanged) {
            final boolean supportsStitching =
                    cmi.getAuthority().hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_STITCHING);
            List<FXRspecLink> toRemove = new ArrayList<>();

            if (!cmi.isLinkSupport() && node.getInterfaces().size() > 0) {
                JFDialogs.create()
                        .owner(this)
                        .message(String.format("'%s' does not support dedicated links. Existing links to '%s' are being removed.",
                                cmi.getAuthority().getName(),
                                node.getClientId()))
                        .masthead("Illegal links detected")
                        .title("Illegal links detected")
                        .showWarning();

                toRemove.addAll(Lists.transform(node.getInterfaces(),
                        rspecInterface -> rspecInterface.isLinkBound() ? rspecInterface.getLink() : null
                ));
            } else {
                for (FXRspecInterface iface : currentNode.getInterfaces()) {
                    if (!iface.isLinkBound())
                        continue;

                    if (iface.getLink().isStitched(authorityList.getAuthorityListModel())) {

                        if (!isValidStitchedLink(iface.getLink())) {

                            Alert alert =
                                    new Alert(Alert.AlertType.WARNING,
                                            String.format("By changing the authority of this node, the link '%s' has become a stitched link.\n" +
                                                            "However, one or more authorities do not support stitched links. It is therefore removed.",
                                                    iface.getLink().getClientId()),
                                            ButtonType.OK);
                            alert.setHeaderText("Invalid stitched link");
                            alert.setTitle("Invalid stitched link");
                            alert.show();

                            toRemove.add(iface.getLink());
                        } else if (JFDialogs.create()
                                .owner(this)
                                .message(
                                        "Link '" + iface.getLink().getClientId() + "' is now a stitched link.\n\n" +
                                                "Stitched links between testbeds are currently VERY experimental and not supported." +
                                                " Do you want to remove this link?")
                                .masthead("Experimental functionality detected.")
                                .title("Experimental functionality detected.")
                                .buttonTypes(ButtonType.YES, ButtonType.NO)
                                .showWarning().filter(r -> r == ButtonType.YES).isPresent()) {


                            toRemove.add(iface.getLink());

                        } else {

                            modelRspecEditor.convertLinkToStitchedLink(iface.getLink());
                        }
                    } else {
                        //the link used to be stitched, but isn't anymore
                        JFDialogs.create()
                                .owner(this)
                                .message("Link '" + iface.getLink().getClientId() + "' is now a normal link." +
                                        " Some settings are reverted to default.")
                                .masthead("Link type change detected")
                                .title("Link type change detected")
                                .showInformation();

                        modelRspecEditor.convertLinkToNormalLink(iface.getLink());

                    }
                }
            }

            toRemove.forEach(modelRspecEditor.getModelRspec()::deleteLink);
        }

        if (!currentNode.getSliverTypes().isEmpty()) {
            SliverType sliverType = currentNode.getSliverTypes().get(0);

            if (sliverType != null) {
                assert diskImageTextField != null;
                if (diskImageTextField.getText() != null) {
                    String diskImageString = diskImageTextField.getText().trim();

                    if (!diskImageString.isEmpty()) {
                        currentNode.setSliverTypeDiskImage(diskImageString);
                    } else {
                        currentNode.setSliverTypeDiskImage(null);
                    }
                } else {
                    currentNode.setSliverTypeDiskImage(null);
                }
            }
        } else {
            LOG.error("Could not find sliver-type on currentNode: cannot save disk image");
        }

        //check if we need to change the node exclusivity
        Boolean exclusive =
                config.getNodeExclusivity(currentNode.getComponentManagerId(), currentNode.getSliverTypeName());

        if (exclusive != null) {
            currentNode.setExclusive(exclusive);
        }

        final FXRspecNode resultNode = currentNode;

        tabPane.getTabs().stream()
                .filter(tab -> tab instanceof AdaptivePropertiesTab)
                .map(tab -> (AdaptivePropertiesTab) tab).forEach(tab -> {
            assert (tab.shouldBeVisible(resultNode)) :
                    tab.getClass().getName() + "indicates it shouldn't be visible at this time!";

            //we assume here that when the tab is visible, it is also able to save it's information
            // to the new FXRspecNode-class!
            // THIS IS POTENTIALLY UNSAFE!
            tab.save(resultNode);
        });


        tabPane.getTabs().stream()
                .filter(tab -> tab instanceof PropertiesTab)
                .map(tab -> (PropertiesTab) tab)
                .forEach(PropertiesTab::save);


        NodeDescription nodeDescription = config.detectJFedNodeType(currentNode);
        if (nodeDescription != null)
            currentNode.setNodeDescriptionId(nodeDescription.getId());


        Stage stage = (Stage) saveButton.getScene().getWindow();
        stage.close();
    }

    String getCurrentSliverTypeName() {

        //get the CMI for the authority that is currently visible
        ComponentManagerInfo cmi = componentManagerInfoComboBox.getSelectionModel().getSelectedItem();

        if (cmi == null)
            return null;

        if (node.getSliverTypeName() != null && cmi.hasSliverType(node.getSliverTypeName())) {
            return node.getSliverTypeName();
        }
        //fallback onto the default sliver-type defined in the CMI
        SliverType fallbackSliverType = cmi.getDefaultSliverType();

        if (fallbackSliverType != null && cmi.hasSliverType(fallbackSliverType.getName())) {
            return fallbackSliverType.getName();
        }
        LOG.warn("Default sliver_type for {} is '{}', but that sliver_type is not in ListResources. " +
                        "Picking an existing one instead.",
                cmi.getAuthority().getHrn(), fallbackSliverType);

        if (!cmi.getSliverTypes().isEmpty())
            return cmi.getSliverTypes().get(0).getName(); //choose first (= quite randomly)
        else
            return null;
    }


    private boolean isValidStitchedLink(FXRspecLink link) {
        for (FXRspecInterface iface : link.getInterfaces()) {
            if (!iface.isNodeBound())
                continue;


            SfaAuthority nodeAuth =
                    authorityList.finder().findByUrn(iface.getNode().getComponentManagerId(), AuthorityFinder.Purpose.STITCHING);

            if (!nodeAuth.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_STITCHING))
                return false;
        }

        return true;
    }


    @FXML
    public void onCancelAction(ActionEvent actionEvent) {
        Stage stage = (Stage) cancelButton.getScene().getWindow();
        stage.close();
    }

    @FXML
    private void onRequestDiskImageAction() {

        String newDiskImage =
                new SelectDiskImageDialog(this.getScene().getWindow(), componentManagerInfoComboBox.getValue(),
                        cmiManager, hltf, tasksFactory, taskService, aggregateManagerWrapperFactory).showDialog();

        if (newDiskImage != null) {
            diskImageTextField.setText(newDiskImage);
        }
    }

    /**
     * @return may return null if nothing is selected
     */
    public ComponentManagerInfo getSelectedComponentManagerInfo() {
        return currentCmi.get();
    }

    public FXRspecNode getSelectedComponentId() {
        return cmiPropertiesPanes.containsKey(currentCmi.get()) ?
                cmiPropertiesPanes.get(currentCmi.get()).getSelectedNode() : null;
    }

    public ReadOnlyObjectProperty<ComponentManagerInfo> selectedComponentManagerProperty() {
        return currentCmi;
    }

    public ReadOnlyBooleanProperty isSpecificNodeChosenProperty() {
        return isSpecificNodeChosen;
    }
}
