package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManager;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.ComponentManagerInfo;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.DeleteDiskImageTask;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.TasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.ListDiskImagesTask;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.DiskImage;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.application.Platform;
import javafx.beans.InvalidationListener;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.collections.transformation.FilteredList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Optional;

/**
 * User: twalcari
 * Date: 9/11/2014
 * Time: 11:57
 */
public class SelectDiskImageDialog extends VBox {
    private static final Logger LOG = LoggerFactory.getLogger(SelectDiskImageDialog.class);

    /*
         Items for public disk images
     */
    @FXML
    private RadioButton publicDiskImagesRadioButton;
    @FXML
    private ComboBox<DiskImage> publicDiskImagesComboBox;

    @FXML
    private VBox publicFetchOverlay;
    @FXML
    private ProgressBar publicAdvertisementFetchProgressBar;

    private ComponentManagerInfo currentPublicDiskImagesCMI = null;

    /*
        Items for private disk images
     */
    @FXML
    private RadioButton privateDiskImagesRadioButton;
    @FXML
    private ComboBox<AggregateManagerWrapper.ImageInfo> privateDiskImagesComboBox;

    @FXML
    private VBox privateFetchOverlay;
    @FXML
    private ProgressBar privateAdvertisementFetchProgressBar;

    @FXML
    private Button deletePrivateDiskImageButton;

    /*
    Advanced
     */
    @FXML
    private CheckBox advancedCheckBox;
    @FXML
    private VBox advancedVBox;
    @FXML
    private ComboBox<ComponentManagerInfo> sourceTestbedComboBox;

    private final BooleanProperty isLoadingUserImages = new SimpleBooleanProperty(false);
    private ComponentManagerInfo currentPrivateDiskImagesCMI = null;


    private final Dialog<String> dialog;

    private final ComponentManagerInfo nodeComponentManagerInfo;
    private final CMIManager globalCmiManager;
    private final HighLevelTaskFactory hltf;
    private final TasksFactory tasksFactory;
    private final TaskService taskService;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;

    public SelectDiskImageDialog(Window parentWindow, ComponentManagerInfo nodeComponentManagerInfo, CMIManager globalCmiManager,
                                 HighLevelTaskFactory hltf, TasksFactory tasksFactory, TaskService taskService,
                                 AggregateManagerWrapperFactory aggregateManagerWrapperFactory) {
        this.nodeComponentManagerInfo = nodeComponentManagerInfo;
        this.globalCmiManager = globalCmiManager;
        this.hltf = hltf;
        this.tasksFactory = tasksFactory;
        this.taskService = taskService;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;

        //initialize the Scene
        FXMLUtil.injectFXML(this);

        //initialize advanced options first (as it contains the source testbed-combobox)
        //advancedVBox.managedProperty().bind(advancedVBox.visibleProperty());
        advancedVBox.visibleProperty().bind(advancedCheckBox.selectedProperty());
        deletePrivateDiskImageButton.visibleProperty().bind(advancedCheckBox.selectedProperty());

        sourceTestbedComboBox.setItems(new FilteredList<>(this.globalCmiManager.getComponentManagerInfos(),
                cmi -> cmi.getAuthority().hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_EXT_USER_DISK_IMAGE_OPS)));
        sourceTestbedComboBox.getSelectionModel().select(nodeComponentManagerInfo);
        sourceTestbedComboBox.setConverter(StringConverters.CMI_STRING_CONVERTER);

        this.sourceTestbedComboBox.getSelectionModel().selectedItemProperty().addListener(observable -> {
            //check if private disk images are available
            if (!areUserDiskImagesAvailable()) {
                publicDiskImagesRadioButton.setSelected(true);
            }
            privateDiskImagesRadioButton.setDisable(!areUserDiskImagesAvailable());

            //update the correct combobox
            if (publicDiskImagesRadioButton.isSelected()) {
                updatePublicDiskImagesComponentManagerInfo();
            } else {
                updatePrivateDiskImagesComponentManagerInfo(false /* no forced refresh */);
            }
        });

        publicDiskImagesComboBox.disableProperty().bind(publicDiskImagesRadioButton.selectedProperty().not());
        publicDiskImagesComboBox.setConverter(StringConverters.DISK_IMAGE_STRING_CONVERTER);

        //public is selected by default, so update that list
        updatePublicDiskImagesComponentManagerInfo();

        privateDiskImagesComboBox.disableProperty().bind(privateDiskImagesRadioButton.selectedProperty().not());
        privateDiskImagesComboBox.setConverter(StringConverters.IMAGEINFO_STRING_CONVERTER);
        privateFetchOverlay.visibleProperty().bind(isLoadingUserImages);

        privateDiskImagesRadioButton.setDisable(!areUserDiskImagesAvailable());

        deletePrivateDiskImageButton.disableProperty().bind(privateDiskImagesComboBox.getSelectionModel().selectedItemProperty().isNull());
        deletePrivateDiskImageButton.setOnAction(this::deleteCurrentPrivateDiskImage);


        privateDiskImagesRadioButton.selectedProperty().addListener(observable -> {
            updatePrivateDiskImagesComponentManagerInfo(false);
        });

        updatePublicDiskImagesList();


        this.dialog = new Dialog<>();
        dialog.initOwner(parentWindow);
        dialog.setTitle("Select Disk Image");
        dialog.getDialogPane().setContent(this);
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.HDD_ALT, Color.BLACK));

        ButtonType selectDiskImageButtonType = new ButtonType("Select Disk Image", ButtonBar.ButtonData.OK_DONE);
        dialog.getDialogPane().getButtonTypes().setAll(selectDiskImageButtonType, ButtonType.CANCEL);

        dialog.setResultConverter(param -> {
            if (param == selectDiskImageButtonType){
                            return getDiskImageString();
            }   else
                return null;
        });
    }

    private void deleteCurrentPrivateDiskImage(ActionEvent actionEvent) {

        AggregateManagerWrapper.ImageInfo imageInfo = privateDiskImagesComboBox.getSelectionModel().getSelectedItem();

        Optional<ButtonType> answer =
                JFDialogs.create()
                        .title("Delete disk image")
                        .message(String.format("Are you sure you want to permanently delete disk image '%s' from %s?",
                                StringConverters.IMAGEINFO_STRING_CONVERTER.toString(imageInfo),
                                currentPrivateDiskImagesCMI.getAuthority().getHrn()))
                        .showConfirm();

        if (answer.isPresent() && answer.get() == ButtonType.YES) {
            DeleteDiskImageTask deleteDiskImageTask =
                    tasksFactory.createDeleteDiskImageTask(currentPrivateDiskImagesCMI.getAuthority(), imageInfo);
            deleteDiskImageTask.setOnSucceeded(event -> updatePrivateDiskImagesComponentManagerInfo(true));

            taskService.submitTask(deleteDiskImageTask);


        }
    }


    private final InvalidationListener publicRefreshTaskInvalidationListener = observable -> {
        if (currentPublicDiskImagesCMI.getAllResources().getCurrentRefreshAdvertisementTask() == null) {
            publicAdvertisementFetchProgressBar.progressProperty().unbind();

            updatePublicDiskImagesList();
        } else {
            publicAdvertisementFetchProgressBar.progressProperty().bind(
                    currentPublicDiskImagesCMI.getAllResources().getCurrentRefreshAdvertisementTask().progressProperty());
        }
    };


    private void updatePublicDiskImagesComponentManagerInfo() {
        final ComponentManagerInfo cmi = sourceTestbedComboBox.getSelectionModel().getSelectedItem();
        if (publicDiskImagesRadioButton.isSelected() && currentPublicDiskImagesCMI != cmi) {

            if (currentPublicDiskImagesCMI != null) {
                currentPublicDiskImagesCMI.getDiskImages().removeListener(publicRefreshTaskInvalidationListener);
            }

            currentPublicDiskImagesCMI = cmi;

            publicFetchOverlay.visibleProperty().unbind();
            publicFetchOverlay.visibleProperty().bind(currentPublicDiskImagesCMI.getAllResources().currentRefreshAdvertisementTaskProperty().isNotNull());

            currentPublicDiskImagesCMI.getAllResources().currentRefreshAdvertisementTaskProperty().addListener(publicRefreshTaskInvalidationListener);
            publicRefreshTaskInvalidationListener.invalidated(null); //force invalidation
        }


        if (!currentPublicDiskImagesCMI.getAllResources().isComponentIdsAvailable() && !currentPublicDiskImagesCMI.getAvailableResources().isComponentIdsAvailable())
            currentPublicDiskImagesCMI.getAllResources().refreshAdvertisement();
        else if (currentPublicDiskImagesCMI.getDiskImages().isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message(String.format("'%s' did not provide any available disk images", currentPublicDiskImagesCMI.getAuthority().getHrn()))
                    .showWarning();
        }
    }


    private void updatePublicDiskImagesList() {
        publicDiskImagesComboBox.getSelectionModel().clearSelection();
        publicDiskImagesComboBox.getItems().clear();
        publicDiskImagesComboBox.getItems().addAll(sourceTestbedComboBox.getSelectionModel().getSelectedItem().getDiskImages());
    }


    private void updatePrivateDiskImagesComponentManagerInfo(boolean forceRefresh) {
        final ComponentManagerInfo cmi = sourceTestbedComboBox.getSelectionModel().getSelectedItem();

        if (forceRefresh ||
                privateDiskImagesRadioButton.isSelected() && currentPrivateDiskImagesCMI != cmi && !isLoadingUserImages.get()) {
            LOG.debug("Getting private disk images from {}", cmi.getAuthority().getHrn());
            isLoadingUserImages.set(true);

            final ListDiskImagesTask listDiskImagesTask =
                    hltf.listDiskImages(cmi.getAuthority());
            hltf.submitTask(listDiskImagesTask, (task, te, state) ->
                    Platform.runLater(() -> {
                        privateDiskImagesComboBox.getItems().clear();
                        privateDiskImagesComboBox.getItems().setAll(listDiskImagesTask.getDiskImageInfos());

                        if (!privateDiskImagesComboBox.getItems().isEmpty()) {
                            privateDiskImagesComboBox.getSelectionModel().selectFirst();
                        }

                        LOG.debug("Updating private images list with {} items from authority {}",
                                listDiskImagesTask.getDiskImageInfos().size(),
                                cmi.getAuthority().getHrn());

                        currentPrivateDiskImagesCMI = cmi;
                        isLoadingUserImages.setValue(false);
                    }));

        }
    }

    public String showDialog() {
        return dialog.showAndWait().orElse(null);
    }

    private String getDiskImageString() {
        /*
            When the disk image resides on the same testbed as the node, we can use the URN.
            Otherwise, we have to use the URL.

             If we would use the URL all the time, this would incur a performance hit.
             Emulab-testbeds will always copy a URL-provided image locally, even if it is already present locally!
        */

        if (publicDiskImagesRadioButton.isSelected() && !publicDiskImagesComboBox.getSelectionModel().isEmpty()) {
            DiskImage selectedDiskImage = publicDiskImagesComboBox.getSelectionModel().getSelectedItem();

            if (nodeComponentManagerInfo == currentPublicDiskImagesCMI) {
                return selectedDiskImage.getName();
            } else {
                return selectedDiskImage.getUrl();
            }
        } else if (privateDiskImagesRadioButton.isSelected() && !privateDiskImagesComboBox.getSelectionModel().isEmpty()) {
            AggregateManagerWrapper.ImageInfo imageInfo =
                    privateDiskImagesComboBox.getSelectionModel().getSelectedItem();


            if (nodeComponentManagerInfo == currentPrivateDiskImagesCMI) {
                return imageInfo.getUrn();
            } else {
                return imageInfo.getUrl();
            }
        } else {
            return null;
        }
    }

    private boolean areUserDiskImagesAvailable() {
        return aggregateManagerWrapperFactory.getAggregateManagerWrapper(sourceTestbedComboBox.getSelectionModel().getSelectedItem().getAuthority())
                .areUserDiskImageFunctionsSupported();
    }
}
