package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.call_gui.TasksWindow;
import be.iminds.ilabt.jfed.experimenter_gui.controller.view.ControllerNodeFactory;
import be.iminds.ilabt.jfed.experimenter_gui.controller.view.ControllerUI;
import be.iminds.ilabt.jfed.experimenter_gui.slice.errors.ErrorsView;
import be.iminds.ilabt.jfed.experimenter_gui.slice.raw.RawSliceView;
import be.iminds.ilabt.jfed.experimenter_gui.util.*;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsCache;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;

/**
 * User: twalcari
 * Date: 1/30/2015
 * Time: 11:50
 */
@Singleton
public class ExperimentViewsFactory {

    private final Provider<TasksWindow> tasksWindowProvider;
    private final Provider<ApiCallDetailsCache> apiCallDetailsCacheProvider;
    private final Provider<SSHTerminalUtil> sshTerminalUtilProvider;
    private final Provider<JFedGuiPreferences> jFedPreferencesProvider;
    private final Provider<AuthorityListModel> authorityListModelProvider;
    private final Provider<SliceExperimentCanvasFactory> sliceExperimentCanvasFactoryProvider;
    private final Provider<NodePropertiesDialogFactory> nodePropertiesDialogFactoryProvider;
    private final Provider<ImageUtil> imageUtilProvider;
    private final Provider<GeniUserProvider> geniUserProviderProvider;
    private final Provider<ProxyPreferencesManager> proxyPreferencesManagerProvider;
    private final Provider<ProxyServiceUtil> proxyServiceUtilProvider;
    private final Provider<ControllerNodeFactory> controllerNodeFactoryProvider;
    private final Provider<BrowserUtil> browserUtilProvider;
    private final Provider<BonfireLicenseUtil> bonfireLicenseUtilProvider;


    @Inject
    public ExperimentViewsFactory(Provider<ApiCallDetailsCache> apiCallDetailsCacheProvider,
                                  Provider<TasksWindow> tasksWindowProvider,
                                  Provider<SSHTerminalUtil> sshTerminalUtilProvider,
                                  Provider<JFedGuiPreferences> jFedPreferencesProvider,
                                  Provider<AuthorityListModel> authorityListModelProvider,
                                  Provider<SliceExperimentCanvasFactory> sliceExperimentCanvasFactoryProvider,
                                  Provider<NodePropertiesDialogFactory> nodePropertiesDialogFactoryProvider,
                                  Provider<ImageUtil> imageUtilProvider,
                                  Provider<GeniUserProvider> geniUserProviderProvider,
                                  Provider<ProxyPreferencesManager> proxyPreferencesManagerProvider,
                                  Provider<ProxyServiceUtil> proxyServiceUtilProvider,
                                  Provider<ControllerNodeFactory> controllerNodeFactoryProvider,
                                  Provider<BrowserUtil> browserUtilProvider,
                                  Provider<BonfireLicenseUtil> bonfireLicenseUtilProvider) {
        this.apiCallDetailsCacheProvider = apiCallDetailsCacheProvider;
        this.tasksWindowProvider = tasksWindowProvider;
        this.sshTerminalUtilProvider = sshTerminalUtilProvider;
        this.jFedPreferencesProvider = jFedPreferencesProvider;
        this.authorityListModelProvider = authorityListModelProvider;
        this.sliceExperimentCanvasFactoryProvider = sliceExperimentCanvasFactoryProvider;
        this.nodePropertiesDialogFactoryProvider = nodePropertiesDialogFactoryProvider;
        this.imageUtilProvider = imageUtilProvider;
        this.geniUserProviderProvider = geniUserProviderProvider;
        this.proxyPreferencesManagerProvider = proxyPreferencesManagerProvider;
        this.proxyServiceUtilProvider = proxyServiceUtilProvider;
        this.controllerNodeFactoryProvider = controllerNodeFactoryProvider;
        this.browserUtilProvider = browserUtilProvider;
        this.bonfireLicenseUtilProvider = bonfireLicenseUtilProvider;
    }

    public ErrorsView createErrorsView() {
        return new ErrorsView(tasksWindowProvider.get(), apiCallDetailsCacheProvider.get(), browserUtilProvider.get(),
                bonfireLicenseUtilProvider.get());
    }

    public RawSliceView createRawSliceView(Experiment experiment, ManifestRspecSource canvasRspecSource) {
        return new RawSliceView(experiment, canvasRspecSource, sshTerminalUtilProvider.get(), browserUtilProvider.get(),
                jFedPreferencesProvider.get(), authorityListModelProvider.get());
    }

    public ModelSliceView createModelSliceView(ExperimentController experimentController, ManifestRspecSource canvasRspecSource) {
        return new ModelSliceView(experimentController, canvasRspecSource, authorityListModelProvider.get(),
                sliceExperimentCanvasFactoryProvider.get(), browserUtilProvider.get());
    }

    public ControllerUI createControllerUi(Experiment experiment, ManifestRspecSource canvasRspecSource, DetailsTabHost detailsTabHost) {
        return new ControllerUI(experiment, canvasRspecSource, detailsTabHost, nodePropertiesDialogFactoryProvider.get(),
                imageUtilProvider.get(), geniUserProviderProvider.get(), proxyPreferencesManagerProvider.get(),
                proxyServiceUtilProvider.get(), controllerNodeFactoryProvider.get(), jFedPreferencesProvider.get());
    }
}
