package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.ReputationService;
import be.iminds.ilabt.jfed.lowlevel.api.RestApi;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.JavaFXLogger;

import java.util.Date;
import java.util.List;

/**
 * User: twalcari
 * Date: 1/11/2016
 * Time: 11:38
 */
public class ListPendingReputationQuestionnairesTask extends ReputationTask {

    private final Slice slice;

    private List<ReputationService.SliceReputation> pendingQuestionnaires;

    public ListPendingReputationQuestionnairesTask(Slice slice, JavaFXLogger logger, JFedPreferences jFedPreferences,
                                                   AuthorityList authorityList, JFedConnectionProvider connectionProvider,
                                                   GeniUserProvider geniUserProvider) {
        super("Requesting pending reputation questionnaires",
                logger, jFedPreferences, authorityList,
                connectionProvider, geniUserProvider);

        this.slice = slice;
    }

    @Override
    protected void doReputationTask(TaskExecution taskExecution, ReputationService reputationService) throws JFedException, InterruptedException {
        RestApi.RestReply<List<ReputationService.SliceReputation>> reply =
                reputationService.registerExperiment(createConnection(),
                        geniUserProvider.getLoggedInGeniUser().getUserUrn(),
                        slice.getUrn(),
                        slice.getCreationDate() != null ? Date.from(slice.getCreationDate()) : null,
                        slice.getExpirationDate() != null ? Date.from(slice.getExpirationDate()) : null);

        if (reply.getHttpCallDetails().getResultHttpStatusCode() != 200) {
            throw new JFedHighLevelException("Error while registering experiment to reputation service");
        }

        pendingQuestionnaires = reply.getValue();
    }

    public List<ReputationService.SliceReputation> getPendingQuestionnaires() {
        return pendingQuestionnaires;
    }
}
