package be.iminds.ilabt.jfed.lowlevel.stitching;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.AggregateManager2;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.*;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.JFedPasswordManager;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import org.slf4j.LoggerFactory;

import java.net.URISyntaxException;
import java.util.List;

/**
 * SynchronousStitcher does synchonouse calls needed to make the stiched topology ready.
 * Calls are not executed in parallel, and there is no error recovery.
 * <p/>
 * There is also no "Status" call functionality (yet)
 * <p/>
 * This is mostly useful as an example to make a Stitcher that does parallel calls with failure handling and Status calls.
 */
@Deprecated
public class SynchronousStitcher {
    private static org.slf4j.Logger LOG = LoggerFactory.getLogger(SynchronousStitcher.class);

    private final Logger logger;
    private final JFedPreferences jFedPreferences;
    private final StitchingDirector director;
    private HandleUntrustedCallback handleUntrustedCallback = null;
    private final SfaConnectionPool conPool;

    /**
     *
     * */
    public SynchronousStitcher(Logger logger, JFedPreferences jFedPreferences, StitchingDirector director, ConnectionBuilderFactory connectionBuilderFactory, JFedPasswordManager jFedPasswordManager, JFedTrustStore jFedTrustStore) {
        if (true)
            throw new RuntimeException("this class needs to be updated, to support DELETE StitchActions"); //see StitchingPingTest for example on how to stitch non-parallel

        this.logger = logger;
        this.director = director;
        this.jFedPreferences = jFedPreferences;
        this.conPool = new SfaConnectionPool(jFedTrustStore, jFedPreferences, connectionBuilderFactory, jFedPasswordManager);
    }

    public void setInsecure() {
        handleUntrustedCallback =
                new BasicConnectionBuilder.INSECURE_TRUSTALL_HandleUntrustedCallback();
    }

    public void setHandleUntrustedCallback(HandleUntrustedCallback handleUntrustedCallback) {
        this.handleUntrustedCallback = handleUntrustedCallback;
    }


    /**
     * Assumes the StitchingDirector has ready hops. (= it has received the computation result)
     * <p/>
     * does all CreateSliver calls needed to create the stitched topology
     */
    public void stitchCreate(List<AnyCredential> credentialList, String sliceUrn, GeniUser user, List<UserSpec> users) {
        System.out.println("Calling createSliver on all aggregates");

        List<StitchingDirector.StitchAction> readyHops = director.getStitchActions();
        while (director.areAnyActionsLeft()) {
            for (StitchingDirector.StitchAction hop : readyHops) {
                SfaAuthority auth = hop.getAuthority();
                String requestRspec = hop.getRequestRspec();

                System.out.println("\n\n***************************************************************************************");
                System.out.println("   CreateSliver call for " + auth.getUrnString());

                //do createsliver call
                AggregateManager2 am2 = new AggregateManager2(logger, jFedPreferences);
//                String manifestRspec = null;
                try {
                    //TODO handle errors (adding retry etc)
                    String url = auth.getUrl(ServerType.GeniServerRole.AM, 2).toURI().toString();
                    SfaConnection con = null;
                    if (url.startsWith("http://")) {
                        LOG.warn("Connection URL is http instead of https! " +
                                "This is unsecure, so this connection protocol will never used. " +
                                "Will try using https instead, maybe that works.");
                        url = url.replaceFirst("http", "https");
                    }
                    //TODO: set handleUntrustedCallback
                    con = (SfaConnection) conPool.getConnectionByAuthority(user, auth, new ServerType(ServerType.GeniServerRole.AM, 2));
                    assert con != null : "URL has unsupported protocol: " + url;

                    System.out.println("Calling CreateSliver...");
                    AbstractGeniAggregateManager.AggregateManagerReply<String> reply = am2.createSliver(
                            con, credentialList, sliceUrn, requestRspec, users,
                            null);
                    System.out.println("Call returned " + reply != null ? reply.getGeniResponseCode() : "null");

                    StitchingDirector.AllocateVerdict verdict = director.processAllocateResult(hop, reply);

                    if (verdict != StitchingDirector.AllocateVerdict.GIVEUP && verdict != StitchingDirector.AllocateVerdict.CALLER_DELAYED_RETRY)
                        LOG.debug("Overview of hop data after CreateSliver call: \n" + director.getSuggestedAvailableOverviewsString());

                    assert verdict != StitchingDirector.AllocateVerdict.GIVEUP && verdict != StitchingDirector.AllocateVerdict.CALLER_DELAYED_RETRY;
//                    assert reply.getGeniResponseCode().isSuccess() : "Call not successful: "+reply.getGeniResponseCode();
//                    manifestRspec = reply.getValue();
                } catch (JFedException e) {
                    LOG.error("ABORTING: GeniException in CreateSliver call: ", e);
                    return;
                } catch (URISyntaxException e) {
                    LOG.error("ABORTING: URISyntaxException in CreateSliver call: ", e);
                    return;
                }
//                assert manifestRspec != null;

//                director.processManifest(hop, manifestRspec);
            }

            readyHops = director.getStitchActions();
        }

        System.out.println("Calls for stitching done");
    }

    public void deleteAll(List<AnyCredential> credentialList, String sliceUrn, GeniUser user) {
        AggregateManager2 am2 = new AggregateManager2(logger, jFedPreferences);

        for (SfaAuthority authorityInfo : director.getActivelyInvolvedAuthorities()) {
            //do delete call
            try {
                System.out.println("Calling delete on " + authorityInfo.getUrnString());

                //TODO handle errors
                String url = authorityInfo.getUrl(ServerType.GeniServerRole.AM, 2).toURI().toString();
                SfaConnection con = null;
                con = (SfaConnection) conPool.getConnectionByAuthority(user, authorityInfo, new ServerType(ServerType.GeniServerRole.AM, 2));
//                if (url.startsWith("http://")) {
//                    LOG.warn("Connection URL is http instead of https! This is unsecure, so connection will not be tried. I will try using https instead, maybe that works.");
//                    url = url.replaceFirst("http", "https");
//                }
//                if (url.startsWith("https://"))
//                    con = new SfaSslWithClientAuthConnection(
//                            url,
//                            authorityInfo,
//                            user,
//                            null/*proxyInfo*/, //TODO fill in ProxyInfo when needed
//                            handleUntrustedCallback);
//                if (url.startsWith("http://"))
//                    con = new SfaPlainHttpConnection(
//                            url, authorityInfo,
//                            null/*proxyInfo*/ //TODO fill in ProxyInfo when needed
//                    );
                assert con != null : "URL has unsupported protocol: " + url;

                AbstractGeniAggregateManager.AggregateManagerReply<Boolean> reply =
                        am2.deleteSliver(con, credentialList, sliceUrn, null);

                System.out.println("Called delete on " + authorityInfo.getUrnString() + " results: " + reply.getValue());
            } catch (JFedException e) {
                e.printStackTrace();
            } catch (URISyntaxException e) {
                e.printStackTrace();
            }
        }
    }
}
