package be.iminds.ilabt.jfed.ui.javafx.scanner;

import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.authority.JFedAuthorityList;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.authority.StoredAuthorityList;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.TextArea;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import javafx.stage.Window;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.io.File;
import java.util.Arrays;
import java.util.Optional;
import java.util.regex.Pattern;

/**
 * ScanResultPanelController
 */
public class ScanResultPanelController {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ScanResultPanelController.class);

    @FXML
    private TextArea xmlTextArea;
    @FXML
    private Button addButton;

    private final JavaFXLogger logger;
    private final GeniUserProvider geniUserProvider;
    private final JFedPreferences jFedPreferences;

    private Stage stage;
    private SfaAuthority authority;
    private boolean scanAgain;

    @Inject
    public ScanResultPanelController(JavaFXLogger logger, GeniUserProvider geniUserProvider, JFedPreferences jFedPreferences) {
        this.logger = logger;
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
        this.scanAgain = false;
    }

    @FXML
    private void initialize() { }

    public void showSnippet() {
        assert xmlTextArea != null;
        if (authority != null)
            xmlTextArea.setText(authority.toXmlString(true/*omitDeclaration*/, true/*nonamespace*/));
        else
            xmlTextArea.setText("No authority");

        addButton.setVisible(false);
    }
    public void showStandalone() {
        assert xmlTextArea != null;
        if (authority != null)
            xmlTextArea.setText(StoredAuthorityList.toXml(Arrays.asList(authority), false/*omitDeclaration*/, false/*nonamespace*/));
        else
            xmlTextArea.setText("No authority");

        addButton.setVisible(true);
    }

    public void save() {
        FileChooser fc = new FileChooser();
        //set the default open dir
        if (System.getProperty("user.home") != null)
            fc.setInitialDirectory(new File(System.getProperty("user.home")));
        fc.setInitialFileName("new_authority.xml");
        File file = fc.showSaveDialog(null);

        if (file == null)
            return;

        IOUtils.stringToFile(file, xmlTextArea.getText());
    }

    private Window getOwner() {
        return addButton.getScene().getWindow();
    }

    public void add() {
        File dir = JFedAuthorityList.getExtraTestbedsDir();
        String name = authority.getNameForUrn().replaceAll(Pattern.quote("."), "_");
        File file = new File(dir, name+".xml");

        if (file.exists()) {
            Optional<ButtonType> reply = JFDialogs.create()
                    .owner(getOwner())
                    .message("Overwrite existing \"" + file.getAbsolutePath() + "\"?")
                    .masthead("File exists")
                    .showConfirm();

            if (!reply.isPresent() && reply.get() != ButtonType.YES) {
                return;
            }
        }

        try {
            if (!dir.exists())
                dir.mkdir();

            IOUtils.stringToFile(file, xmlTextArea.getText());
            System.out.println("Wrote authority to jFed: " + file.getAbsolutePath());
            LOG.debug("Wrote authority to jFed: " + file.getAbsolutePath());
            JFDialogs.create()
                    .owner(getOwner())
                    .message("The Authority XML has been successfully stored to the jFed \"extra authorities dir\".\n" +
                            "Filename: "+file.getAbsolutePath())
                    .masthead("Successfully stored info")
                    .showInformation();
        } catch (Exception | AssertionError t) {
            LOG.error("Error writing \""+file.getAbsolutePath()+"\"", t);
            JFDialogs.create()
                    .owner(getOwner())
                    .message("Error writing \"" + file.getAbsolutePath() + "\": "+t.getMessage())
                    .masthead("Error writing file")
                    .showError();
        }
    }

    public void scan() {
        scanAgain = true;
        if (stage != null)
            stage.close();
    }
    public void close() {
        scanAgain = false;
        if (stage != null)
            stage.close();
    }

    public void setAuthority(SfaAuthority authority) {
        this.authority = authority;
    }

    public void setStage(Stage stage) {
        this.stage = stage;
    }

    public boolean isScanAgain() {
        return scanAgain;
    }
}
